#include <iostream>
#include <algorithm>
#include <stdint.h>
#include "distree_t.h"
#include "DError.h"
#include "mathtool.h"
#include "utility.h"
#include "hpgv_utiltiming.h"
#include "parallel_test.h"
#include "distree_vertex_t.h"
#include "feature_geometry_t.h"
#include "feature_isoblock_t.h"
#include "feature_global_leaves_t.h"

using namespace std;
using namespace lily;

distree_node_t::distree_node_t( void )
    :isLeaf(1),level(-1),parent(-1),key(0)
    ,location(REMOTE),bHas_dirty_proc_3x(false)
{
    for (int i = 0 ; i < 8 ; i ++){
        m_children[i] = -1;
        vertex_indices[i] = -1;
    }
}

distree_node_t::~distree_node_t()
{
}

BBox distree_node_t::get_triple_bbox( void )
{
    BBox cur_nodel_triple = m_bound;
    cur_nodel_triple.Expand(cur_nodel_triple.MaximumExtentf());
    return cur_nodel_triple;
}

distree_t::distree_t(int peid, feature_type_t feature_type, MPI_Comm comm,
    int max_depth, float resolution/*=1.0f*/, float epsilon/*=1.0*/)
    :m_myid(peid),m_feature_type(feature_type)
    ,m_features(NULL),m_cur_active_features(NULL)
    ,m_global_leaf_block_array(NULL), m_isoblock_array(NULL)
    ,root_index(0),max_depth_id(max_depth)
    ,min_extend(resolution),m_epsilon(epsilon),m_communicator(comm)
    ,m_has_features(false),m_leaf_count(0)
{
    set_leaf_node_size(vec3f(1.0f,1.0f,1.0f));
}

distree_t::~distree_t(void)
{
    for (int i = 0; i < m_distance_volumes.size() ; i++)
    {
        delete m_distance_volumes[i];
        m_distance_volumes[i] = NULL;
    }
    m_distance_volumes.clear();
}

bitmap_t * distree_t::generate_bitmap(int groupsize, color_t mycolor, MPI_Comm mpicomm)
{
    uint8_t  *lmap   = NULL;
    uint8_t  *gmap   = NULL;
    bitmap_t *bitmap = NULL;
    int     mapsize = (int)(ceil(groupsize / 8.0f));
    uint8_t mycode = 0;
    int nbits = 0;
    size_t myelement, mybit, i;

    lmap = (uint8_t *)calloc(mapsize, sizeof(uint8_t));
    D_ASSERT_P(m_myid, lmap != 0, "Out of memory", DERROR_ERR_MEM);

    gmap = (uint8_t *)calloc(mapsize, sizeof(uint8_t));
    D_ASSERT_P(m_myid, gmap != 0, "Out of memory", DERROR_ERR_MEM);
    if (mycolor == BLACK) {
        myelement = m_myid / 8;
        mybit  = m_myid % 8;
        mycode = 1 << (/*7 -*/ mybit);
        lmap[myelement] = mycode;
    }

    MPI_Allreduce(lmap, gmap, mapsize, MPI_BYTE, MPI_BOR, mpicomm);

    free(lmap);

#ifdef DEBUG_0
    fprintf(stderr, "id %d ", m_myid);
    for (i = 0; i < mapsize; i++) {
        fprintf(stderr, " %s ",  byte_to_binary(gmap[i]));
    }
    fprintf(stderr, "\n");
    fflush(stderr);
#endif

    bitmap = (bitmap_t *)calloc(1, sizeof(bitmap_t));
    D_ASSERT_P(m_myid, bitmap != 0, "Out of memory", DERROR_ERR_MEM);

    for (i = 0; i < mapsize ; i++){
        nbits += bit_count_table[gmap[i]];
    }
    bitmap->data = (uint8_t*)gmap;
    bitmap->size = mapsize;
    bitmap->nbits= nbits;

    return bitmap;
}
int distree_t::compute_parent_key(int child_key,int child_depth_id) const
{
    int nShift = ((max_depth_id-child_depth_id+1)*3);
    return (child_key >> nShift ) << nShift;
}
int distree_t::compute_parent_key(const distree_node_t &child_node) const
{
    int nShift = ((max_depth_id-child_node.level+1)*3);
    return (child_node.key >> nShift ) << nShift;
}

int distree_t::compute_global_child_key(const distree_node_t &parent_node,
                                    int lockey) const
{
    int parent_key = parent_node.key;
    int parent_depth_id = parent_node.level;
    int child_key
        = parent_key | (lockey << ((max_depth_id - parent_depth_id - 1 )*3));
    return  child_key;
}

int distree_t::compute_local_child_key(const distree_node_t &child_node) const
{
    int global_key = child_node.key;
    int child_depth_id = child_node.level;
    return ((global_key >> ((max_depth_id - child_depth_id)*3)) & 0x00000007);
}

int distree_t::compute_octree_vertex_key(distree_vertex_t &vtx) const
{
    int i_depth = 0;
    int key=0;
    int locKey=0;
    float threshold  = min_extend*0.5f;
    float px = vtx.coords.x(), py = vtx.coords.y(), pz = vtx.coords.z();
    const vec3f center_offset[8] = {
        vec3f(-1,-1,-1), vec3f(1,-1, -1), vec3f(-1, 1,-1), vec3f(1, 1, -1),
        vec3f(-1,-1, 1), vec3f(1,-1,  1), vec3f(-1, 1, 1), vec3f(1, 1,  1)
    };
    float radius = m_global_domain_size.max_extend1i()*0.5f;
    vec3f c = m_center;

    for (i_depth = 1; i_depth <= max_depth_id+1 && radius >= threshold; i_depth++)
    {
        locKey = ( ( pz > c[2]) << 2 ) +	
                 ( ( py > c[1]) << 1 ) +	
                 ( ( px > c[0]) << 0 ) ;	
        key |= locKey << ((max_depth_id+1 - i_depth)*3); 
        radius *= 0.5f;
        c += center_offset[locKey] * radius;
    }
    return key;
}

void distree_t::compute_from_key_to_coords(int key, vec3f &coords) const
{
    const vec3f center_offset[8] = {
        vec3f(-1,-1,-1), vec3f(1,-1, -1), vec3f(-1, 1,-1), vec3f(1, 1, -1),
        vec3f(-1,-1, 1), vec3f(1,-1,  1), vec3f(-1, 1, 1), vec3f(1, 1,  1)
    };
    float radius = m_global_domain_size.max_extend1i()*0.5f;
    vec3f c(m_center);
    int   locKey=0;
    int i_depth = 0;
    for (i_depth = 1; i_depth <= max_depth_id/*+1*/ && radius >= min_extend; i_depth++) {
         locKey = key >> ((max_depth_id+1 - i_depth)*3); 
         locKey &= 0x07;
        radius *= 0.5;
        c[0] = c[0] + center_offset[locKey][0]*radius;
        c[1] = c[1] + center_offset[locKey][1]*radius;
        c[2] = c[2] + center_offset[locKey][2]*radius;
    }
    locKey = key >> ((max_depth_id+1 - i_depth)*3); 
    locKey &= 0x07;
    c += center_offset[locKey]*radius;
    coords = c;
}

void distree_t::create_global_distree(
    const vec3i &global_domain_size,
    const vec3f &global_feature_size,
    const vec3i &np_xyz, int groupsize)
{
    m_np_xyz = np_xyz;
    m_groupsize = groupsize;
    m_global_feature_size = global_feature_size;
    m_center = vec3f(0,0,0);
    float global_extend        = (float)global_domain_size.max_extend1i();
    m_global_domain_size       = vec3i((int)global_extend);
    m_global_domain_bound.pMin = m_center - global_extend*0.5f;
    m_global_domain_bound.pMax = m_center + global_extend*0.5f;

    vec3i pxyz   = vec3i::index1D_to_index3D((int)m_myid, np_xyz);
    m_global_feature_bound.pMin= m_center - global_feature_size*0.5f;
    m_global_feature_bound.pMax= m_center + global_feature_size*0.5f;

    HPGV_TIMING_END(parallel_test::MY_ALL_SLOVE_TIME);
    HPGV_TIMING_END(parallel_test::CREATE_GLOBAL_TREE_TIME);
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(parallel_test::T_g_init);
    HPGV_TIMING_COUNT(parallel_test::T_g_init);
    m_isoblock_array->initial_distree_feature_array(*this);
    HPGV_TIMING_END(parallel_test::T_g_init);

    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(parallel_test::MY_ALL_SLOVE_TIME);
    HPGV_TIMING_BEGIN(parallel_test::CREATE_GLOBAL_TREE_TIME);
    distree_node_t node;
    m_node_array.push_back(node,true);
    m_node_array[0].level  = 0;
    m_node_array[0].parent = -1;
    m_node_array[0].isLeaf = 1;
    m_node_array[0].isoblock_3x_indices.resize(m_isoblock_array->size_feature(), true);
    m_node_array[0].isoblock_overlap_indices.resize(m_isoblock_array->size_feature(), true);
    m_node_array[0].m_bound = m_global_domain_bound;
    for (int i=0; i< m_node_array[0].isoblock_3x_indices.size() ; i++){
        m_node_array[0].isoblock_3x_indices[i] = i;
        m_node_array[0].isoblock_overlap_indices[i] = i;
    }
    
    m_node_array[0].processor_indices.resize(groupsize, true);
    m_node_array[0].processor_3x_indices.resize(groupsize, true);
    for (int i=0; i< m_node_array[0].processor_indices.size() ; i++){
        m_node_array[0].processor_indices[i] = i;
        m_node_array[0].processor_3x_indices[i] = i;
    }

    int leafnum=1;
    size_t start=0, end=m_node_array.size();
    int parent_depth_id = 0;
    while (true){
        for (size_t parent_idx = start ; parent_idx < end ; parent_idx++ ){
            distree_node_t &parent_node = m_node_array[parent_idx];
            parent_depth_id = parent_node.level;
            if (parent_node.isoblock_3x_indices.size() > 0){
                bool bSplitted = create_children(parent_idx, m_isoblock_array, true);
                distree_node_t &parent_node = m_node_array[parent_idx];
                if (bSplitted){
                    parent_node.isLeaf = 0;
                    leafnum += 7;
                }
            }
        }
        start = end;
        end = m_node_array.size();

        if ((float)leafnum/(float)groupsize >= m_epsilon
            || parent_depth_id > max_depth_id
            || end==start){
                break;
        }
    }
    distree_node_t& lastNode = m_node_array.back();
    HPGV_TIMING_COUNT(parallel_test::C_global_tree_h);
    HPGV_TIMING_INCREASE(parallel_test::C_global_tree_h,lastNode.level);
    HPGV_TIMING_INCREASE(parallel_test::C_global_tree_size,m_node_array.size());
    char msg[128];
    sprintf(msg, "****global tree height = %d****\n", lastNode.level);
    D_P_MSG(m_myid, 0, msg);
    HPGV_TIMING_BEGIN(parallel_test::ASSIGN_LOCAL_REGION);
    HPGV_TIMING_COUNT(parallel_test::ASSIGN_LOCAL_REGION);

    assign_leaf_node();

    HPGV_TIMING_END(parallel_test::ASSIGN_LOCAL_REGION);
}

bool distree_t::create_children(size_t parent_node_index, IFeatures *features, bool is_profiling/*=false*/)
{
    distree_node_t &parent_node = m_node_array[parent_node_index];
    int parent_depth_id = parent_node.level;
    float child_extend = parent_node.m_bound.MaximumExtentf() * 0.5f;

    const vec3f pt_min_offset[8]= {
        vec3f(0, 0, 0), vec3f(1, 0, 0), vec3f(0, 1, 0), vec3f(1, 1, 0),
        vec3f(0, 0, 1), vec3f(1, 0, 1), vec3f(0, 1, 1), vec3f(1, 1, 1)
    };
    if (child_extend < min_extend
        || parent_depth_id >= max_depth_id){
        return false;
    }
    if (!is_profiling){
        for (int i = 0; i < 8 ; i++){
            m_node_array.push_back(distree_node_t());
            size_t cur_child_index = m_node_array.size()-1;
            distree_node_t&
                current_child =	m_node_array[cur_child_index];
            current_child.level = parent_depth_id + 1;
            distree_node_t &parent_node = m_node_array[parent_node_index];
            current_child.m_bound.pMin = parent_node.m_bound.pMin + pt_min_offset[i]*child_extend;
            current_child.m_bound.pMax = current_child.m_bound.pMin + child_extend;
            current_child.location = parent_node.location;

            current_child.parent = (int)parent_node_index;
            current_child.key = compute_global_child_key(parent_node, i);
            parent_node.m_children[i] = (int)cur_child_index;
            current_child.isLeaf = features->collect_feature(cur_child_index,
                                                            parent_node_index,
                                                            *this,is_profiling);
        }
    }else{
        for (int i = 0; i < 8 ; i++){
            m_node_array.push_back(distree_node_t(),is_profiling);
            size_t cur_child_index = m_node_array.size()-1;
            distree_node_t&
                current_child =	m_node_array[cur_child_index];
            current_child.level = parent_depth_id + 1;
            distree_node_t &parent_node = m_node_array[parent_node_index];
            current_child.m_bound.pMin = parent_node.m_bound.pMin + pt_min_offset[i]*child_extend;
            current_child.m_bound.pMax = current_child.m_bound.pMin + child_extend;
            current_child.location = parent_node.location;

            current_child.parent = (int)parent_node_index;
            HPGV_TIMING_BEGIN(parallel_test::T_g_key);
            current_child.key = compute_global_child_key(parent_node, i);
            HPGV_TIMING_END(parallel_test::T_g_key);
            parent_node.m_children[i] = (int)cur_child_index;
            HPGV_TIMING_BEGIN(parallel_test::T_g_collect);
            current_child.isLeaf = features->collect_feature(cur_child_index,
                                                            parent_node_index,
                                                            *this,is_profiling);
            HPGV_TIMING_END(parallel_test::T_g_collect);
        }
    }

    return true;
}

void distree_t::assign_leaf_node()
{
    
    bool bHasWorkLoad = false;
    m_leaf_count = 0;
    svector<int> leaf_node_index;
    feature_isoblock_t* isoblocks
        = dynamic_cast<feature_isoblock_t*>(m_isoblock_array);
    svector<int> &my_isoblock_index = m_processor_array[m_myid].m_isoblock_indices;
    
    for (int i=0 ; i < m_node_array.size() ; i++){
        distree_node_t &cur_node = m_node_array[i];
        if (cur_node.isLeaf)
        {
            m_leaf_count++;
            leaf_node_index.push_back(i);
            for (int ib=0 ; ib < my_isoblock_index.size() ; ib++)
            {
                distree_isoblock_t& cur_isoblock = (*isoblocks)[my_isoblock_index[ib]];
                if (cur_node.m_bound.Overlap(cur_isoblock.m_bound,0.001f)
                    && cur_isoblock.has_feature)
                {
                    bHasWorkLoad = true;
                    m_assigned_local_leaf_indices.push_back(i);
                }
            }
        }
    }

    m_assigned_local_leaf_indices.remove_duplicates(int_compare);
    
    svector<uint8_t> hasWorkloadFlags;
    hasWorkloadFlags.resize(m_groupsize);

    MPI_Allgather(&(bHasWorkLoad),      1, MPI_UNSIGNED_CHAR,
                  &hasWorkloadFlags[0], 1, MPI_UNSIGNED_CHAR,
                  m_communicator);

    bool allHasWorkLoad = true;
    for (int i = 0; allHasWorkLoad && i < hasWorkloadFlags.size() ; i++){
        allHasWorkLoad = allHasWorkLoad & (bool)hasWorkloadFlags[i];
    }
    svector<int> all_assigned_dirty_leaf_idx;
    svector<int> count_array, displ_array;
    all_assigned_dirty_leaf_idx.mpi_allgatherv(m_assigned_local_leaf_indices,
                                          count_array, displ_array, MPI_INT);

    all_assigned_dirty_leaf_idx.remove_duplicates(int_compare);
    int total_node_count = m_node_array.size();

    svector<int> empty_leaf_indices;
    empty_leaf_indices.substract(leaf_node_index, all_assigned_dirty_leaf_idx, int_compare);
    int nNoWorkProcs = 0;
    if (allHasWorkLoad){
        D_P_MSG(m_myid, 0,"*********all has workload***********\n");
        
        size_t num_empty_leaf_nodes = empty_leaf_indices.size();
        int istart = BLOCK_LOW( m_myid, m_groupsize, num_empty_leaf_nodes);
        int iend   = BLOCK_HIGH(m_myid, m_groupsize, num_empty_leaf_nodes);
        char msg[256];
        sprintf(msg, "get %d distributed no-assigned leaf nodes[%d~%d]."
             ,iend-istart+1,istart, iend);
        D_P_MSG(m_myid, m_myid, msg);
        for (int i = istart; i <= iend ; i++){
            m_assigned_local_leaf_indices.push_back(empty_leaf_indices[i]);
        }
    }else
    {
        D_P_MSG(m_myid, 0,"***********second case***********\n");
        if (m_assigned_local_leaf_indices.size() <= 0){
            for (int i = 0; i < hasWorkloadFlags.size() ; i++){
                nNoWorkProcs = hasWorkloadFlags[i] ? nNoWorkProcs : nNoWorkProcs+1;
            }
            
            size_t num_empty_leaf_nodes = empty_leaf_indices.size();
            int noWorkProcId=0;
            for (int i = 0; i < m_myid && i < hasWorkloadFlags.size() ; i++){
                if (! hasWorkloadFlags[i]){
                    noWorkProcId++;
                }
            }
            if (num_empty_leaf_nodes>=nNoWorkProcs)
            {
                int istart = BLOCK_LOW( noWorkProcId, nNoWorkProcs, num_empty_leaf_nodes);
                int iend   = BLOCK_HIGH(noWorkProcId, nNoWorkProcs, num_empty_leaf_nodes);
                for (int i = istart; i <= iend ; i++){
                    m_assigned_local_leaf_indices.push_back(empty_leaf_indices[i]);
                }
            }else{
                m_assigned_local_leaf_indices.push_back
                    (empty_leaf_indices[noWorkProcId%empty_leaf_indices.size()]);
            }
        }
    }

    char msg[256];
    sprintf(msg,"current PE still has NO leaf node. empty_leaf_node_size=%d\
, nNoWorkProcs=%d\n", empty_leaf_indices.size(), nNoWorkProcs);

    D_ASSERT_P(m_myid, m_assigned_local_leaf_indices.size()>0, msg, 
               DERROR_ERROR);
    for (int i = 0; i < m_assigned_local_leaf_indices.size() ; i++)
    {
        distree_node_t &cur_node = m_node_array[m_assigned_local_leaf_indices[i]];
        cur_node.location = distree_node_t::LOCAL;
        
        distree_node_t* ancestor_node = &cur_node;
        while(   ancestor_node->parent >= 0
            && m_node_array[ancestor_node->parent].location
            != distree_node_t::LOCAL)
        {
            ancestor_node = &(m_node_array[ancestor_node->parent]);
            ancestor_node->location = distree_node_t::LOCAL;
        }
    }
    m_assigned_local_leaf_indices.shrink_to_fit();

    collect_all_leaf_assignment();
}

void distree_t::recompute_global_tree_using_leaf_assignment()
{
    feature_global_leaves_t* global_tree_leaf_node_feature
        = new feature_global_leaves_t();
    global_tree_leaf_node_feature->initial_distree_feature_array(*this);

    m_global_leaf_block_array = global_tree_leaf_node_feature;
    set_feature_array(global_tree_leaf_node_feature);
    m_is_building_global_tree = true;
    
    m_node_array[0].processor_indices.clear();
    m_node_array[0].processor_3x_indices.clear();
    m_node_array[0].isoblock_3x_indices.clear();
    m_node_array[0].isoblock_overlap_indices.clear();

    for (int ip = 0; ip < m_groupsize ; ip++)
    {
        m_node_array[0].processor_indices.push_back(ip);
        m_node_array[0].processor_3x_indices.push_back(ip);
    }
    for (int in = 0 ; in < m_global_leaf_block_array->size_feature(); in++)
    {
        {
            m_node_array[0].isoblock_overlap_indices.push_back(in);
            m_node_array[0].isoblock_3x_indices.push_back(in);
        }
    }
    bool ret = node_has_dirty_proc_3x(m_node_array[0]);
    m_node_array[0].bHas_dirty_proc_3x = ret;

    
    int parent_end   = 1;
    int parent_start = 0;
    int child_index  = 1;
    while(parent_start < m_node_array.size())
    {
        for (int parent_id = parent_start; parent_id < parent_end ; parent_id++)
        {
            distree_node_t& cur_parent = m_node_array[parent_id];
            if (!cur_parent.isLeaf)
            {
                if (cur_parent.isoblock_3x_indices.size() > 0){
                    for (int j = 0; j < 8 ; j++)
                    {
                        int cur_child_index = cur_parent.m_children[j];
                        D_ASSERT_P(m_myid, cur_child_index >= 0,
                            "recompute_global_tree_using_leaf_assignment():cur_child_index < 0", DERROR_ERR_MEM);

                        m_features->collect_feature(child_index, parent_id, *this);
                        child_index ++ ;
                    }
                }else{
                    child_index+=8;
                }
            }
        }
        parent_start = parent_end;
        parent_end   = child_index;
    }
    char msg[256];
    sprintf(msg, "parent_end=%d != m_node_array.size()=%d",parent_end, m_node_array.size());
    D_ASSERT_P(m_myid, parent_end == m_node_array.size(), msg , DERROR_ERR_MEM);
}
void distree_t::collect_all_leaf_assignment(){
    svector<int> send_rcounts;
    svector<int> send_displs;
    svector<int> all_assigned_leafs;
    all_assigned_leafs.mpi_allgatherv(m_assigned_local_leaf_indices,send_rcounts,send_displs,MPI_INT);

    svector<int> unique_send_gather;
    unique_send_gather.reserve(all_assigned_leafs.size());
    unique_send_gather.clone(all_assigned_leafs);
    unique_send_gather.remove_duplicates(int_compare);
    char err_msg[256];
    
    if (unique_send_gather.size() != m_leaf_count){
        sprintf(err_msg,"send_gather.size(%d), unique_send_gather.size(%d) != total_leaf_count(%d)\n",
                all_assigned_leafs.size(),unique_send_gather.size(), m_leaf_count);
        D_ASSERT_P(m_myid, unique_send_gather.size() == m_leaf_count, err_msg, DERROR_ERR_MEM);
    }
    int node_count = m_node_array.size();
    for (int i = 0 ; i < node_count ; i++){
       distree_node_t& cur_node = m_node_array[i];
        if (cur_node.isLeaf){
            cur_node.processor_indices.clear();
        }
   }
    for (int iPE = 0 ; iPE < m_groupsize ; iPE++){
        for (int j = 0 ; j < send_rcounts[iPE] ; j++){
            int node_idx = all_assigned_leafs[send_displs[iPE]+j];
            D_ASSERT_P(m_myid, node_idx < m_node_array.size(),
                       "node_idx out of bound\n", DERROR_ERR_MEM);
            if(! m_node_array[node_idx].isLeaf){
                char msg[256];
                sprintf(msg, "****node%d in PE%d is not leaf.***\n",node_idx, iPE);
                D_ASSERT_P(m_myid, m_node_array[node_idx].isLeaf, msg, DERROR_ERR_MEM);
            }
            m_node_array[node_idx].processor_indices.push_back(iPE);
        }
    }
    for (int i = 0 ; i < node_count ; i++){
        distree_node_t& cur_node = m_node_array[i];
        if (cur_node.isLeaf){
            cur_node.processor_indices.remove_duplicates(int_compare);
            cur_node.processor_3x_indices.remove_duplicates(int_compare);
            bool ret = node_has_dirty_proc_3x(cur_node);
            cur_node.bHas_dirty_proc_3x = ret;
        }
    }
    return;
   
    for(int i = 0 ; i < node_count; i++)
    {
        distree_node_t& cur_leaf = m_node_array[i];
        if (cur_leaf.isLeaf){
            cur_leaf.processor_3x_indices.push_back(cur_leaf.processor_indices);
        }
    }
    for(int i = 0 ; i < node_count; i++)
    {
        distree_node_t& cur_leaf = m_node_array[i];
        if (cur_leaf.isLeaf)
        {
            BBox cur_node_triple = cur_leaf.get_triple_bbox();
            for(int j = 0 ; j < node_count; j++)
            {
                if ( j != i)
                {
                    distree_node_t& cur_other_leaf = m_node_array[j];
                    if (  cur_other_leaf.isLeaf
                        &&cur_node_triple.Overlap(cur_other_leaf.m_bound)){
                        cur_leaf.processor_3x_indices.push_back(cur_other_leaf.processor_indices);
                    }
                }
            }

            cur_leaf.processor_3x_indices.remove_duplicates(int_compare);
            bool ret = node_has_dirty_proc_3x(cur_leaf);
            cur_leaf.bHas_dirty_proc_3x = ret;
        }
    }
}

int key_idx_compare(const void *pa,const void *pb)
{
    key_idx_pair_t *a = (key_idx_pair_t *)pa;
    key_idx_pair_t *b = (key_idx_pair_t *)pb;

    if (a->key == b->key) {
        return 0;
    } else {
        if (a->key < b->key)
            return -1;
        else
            return 1;
    }
}

int key_compare(const void *pa,const void *pb)
{
    uint32_t *a = (uint32_t *)pa;
    uint32_t *b = (uint32_t *)pb;

    if (*a == *b) {
        return 0;
    } else {
        if (*a < *b)
            return -1;
        else
            return 1;
    }
}

int distree_vertex_compare(const void *pa,const void *pb)
{
    distree_vertex_t *a = (distree_vertex_t *)pa;
    distree_vertex_t *b = (distree_vertex_t *)pb;

    if (a->key == b->key) {
        return 0;
    } else {
        if (a->key < b->key)
            return -1;
        else
            return 1;
    }
}
void distree_t::arrays_intersection(svector<key_idx_pair_t> &empty_leaf_parent_key_idx_pairs,
                                    svector<uint32_t>       &local_leaf_parent_keys,
                                    svector<int>            &intersection_idx )
{
    qsort(empty_leaf_parent_key_idx_pairs.data_ptr(),
          empty_leaf_parent_key_idx_pairs.size(),
          sizeof(key_idx_pair_t),
          key_idx_compare);
    qsort(local_leaf_parent_keys.data_ptr(),
          local_leaf_parent_keys.size(),
          sizeof(uint32_t),
          key_compare);

    const svector<key_idx_pair_t> &a = empty_leaf_parent_key_idx_pairs;
    const svector<uint32_t>       &b = local_leaf_parent_keys;
          svector<int>            &c = intersection_idx;
    size_t aindex = 0, asize = a.size();
    size_t bindex = 0, bsize = b.size();

    while (aindex < asize && bindex < bsize) {
        if (a[aindex].key == b[bindex]) {
            c.push_back(a[aindex].idx);
            aindex++;
        }
        else if (a[aindex].key < b[bindex]) {
            aindex++;
        }
        else {
            bindex++;
        }
    }
}
bool distree_t::node_has_dirty_proc_3x(const distree_node_t& node)
{
    for (int i = 0 ; i < node.processor_3x_indices.size() ; i++)
    {
        const distree_processor_t& processor = m_processor_array[node.processor_3x_indices[i]];
        if (processor.m_has_features)
        {
            return true;
        }
    }
    return false;
}

void distree_t::print_dirty_proc_3x(const distree_node_t& node)
{
    fprintf(stderr, "\n************Start 3x dirty proc list of node.level=%d node.key=%d ***********\n"
        , node.level
        , node.key
        );

    for (int i = 0 ; i < node.processor_3x_indices.size() ; i++)
    {
        if (m_processor_array[node.processor_3x_indices[i]].m_has_features)
        {
            fprintf(stderr, "node.key=%d.Proc3X[%d]=%d\n"
                , node.key
                , i
                , node.processor_3x_indices[i]
            );
        }
    }
    fprintf(stderr, "************ End ***********\n");
}

void distree_t::print_dirty_proc_1x(const distree_node_t& node)
{
    fprintf(stderr, "\n----------Start: dirty 1x_proc list of node.level=%d node.key=%d ----------n"
            , node.level
            , node.key
            );
    for (int i = 0 ; i < node.processor_indices.size() ; i++)
    {
       if (m_processor_array[node.processor_indices[i]].m_has_features)
       {
          fprintf(stderr, "node.key=%d.Proc1X[%d]=%d\n"
              , node.key
              , i
              , node.processor_3x_indices[i]
              );
       }
    }
    fprintf(stderr,"---------- End ----------\n");
}
void distree_t::create_comm_schedule()
{
    D_ASSERT_P(m_myid, m_node_array.size() > 0, "node_array is empty",  DERROR_ERR_MEM);
    D_ASSERT_P(m_myid, m_isoblock_array->size_feature() > 0, "isoblock_array is empty", DERROR_ERR_MEM);
    
    int local_leaf_count = m_assigned_local_leaf_indices.size();

    for (int i = 0 ; i < local_leaf_count ; i++)
    {
        distree_node_t *cur_node_s =
                        &(m_node_array[m_assigned_local_leaf_indices[i]]);

        D_ASSERT_P(m_myid, cur_node_s->isLeaf, "create_comm_schedule: sent_table: local_leaf is not leaf", DERROR_ERR_MEM);
        D_ASSERT_P(m_myid, cur_node_s->location==distree_node_t::LOCAL, "create_comm_schedule: sent_table: local_leaf is remote", DERROR_ERR_MEM);

        if (   cur_node_s->isLeaf
            && cur_node_s->location == distree_node_t::LOCAL)
        {
            distree_node_t* cur_leaf_s = cur_node_s;
            while(  (
                    cur_node_s->processor_3x_indices.empty()
                     )  
                &&cur_node_s->parent > 0
                )
            {   
                cur_node_s = &(m_node_array[cur_node_s->parent]);
            }

            for (int j = 0 ; j < cur_node_s->processor_3x_indices.size() ; j++)
            {
                int proc_idx = cur_node_s->processor_3x_indices[j];

                if (proc_idx != m_myid
                    && m_processor_array[proc_idx].m_has_features)
                {
                    m_sent_table.push_back(proc_idx);
                }
            }
        }
    }
    
    if ( !m_sent_table.empty())
    {
        m_sent_table.remove_duplicates(int_compare);
    }
    if (m_has_features)
    {
        int node_count = m_node_array.size();
        for (int i = 0 ; i < node_count ; i++)
        {
            distree_node_t *cur_node_r = &(m_node_array[i]);
                
            if (cur_node_r->isLeaf)
            {
                distree_node_t* cur_leaf = cur_node_r;
                while( ( cur_node_r->processor_3x_indices.empty())
                      && cur_node_r->parent > 0 
                    )
                {
                    cur_node_r = &(m_node_array[cur_node_r->parent]);
                }

                svector<int> &processors_3x = cur_node_r->processor_3x_indices;
                if(processors_3x.contain(m_myid))
                {
                    svector<int> &assigned_processors = cur_leaf->processor_indices;

                    for (int j = 0 ; j < assigned_processors.size() ; j++)
                    {
                        if (assigned_processors[j] != m_myid)
                        {
                            m_recv_table.push_back(assigned_processors[j]);
                        }
                    }
                }
            }
        }
    }

    if ( !m_recv_table.empty())
    {
        m_recv_table.remove_duplicates(int_compare);
    }
}

void distree_t::verify_schedule( int groupsize )
{
    /* root gather all blocks */
    svector<int> send_rcounts;
    svector<int> send_displs ;
    svector<int> send_gather ;
    svector<int> recv_rcounts;
    svector<int> recv_displs ;
    svector<int> recv_gather ;
    int root = 0;
    int iPE, j, pe_id;

    char msg[256];
    if (m_myid == root) {
        send_rcounts.resize(groupsize,0);
    }
    //Gather sent_table size from all processors.
    int size = m_sent_table.size();
    const int INVALID_ID = groupsize;
    MPI_Gather(&size, 1, MPI_INT, data_ptr(send_rcounts), 1, MPI_INT, root, MPI_COMM_WORLD);

    if (m_myid == root) {
        send_displs.resize(groupsize,0);
        send_displs[0] = 0;
        for (iPE = 1; iPE < groupsize; iPE++) {
            send_displs[iPE] = send_displs[iPE-1] + send_rcounts[iPE-1];
        }
        send_gather.resize(send_displs[groupsize-1] + send_rcounts[groupsize-1], INVALID_ID);
    }

    MPI_Gatherv(data_ptr(m_sent_table),
                m_sent_table.size(),
                MPI_INT,
                data_ptr(send_gather),
                data_ptr(send_rcounts),
                data_ptr(send_displs),
                MPI_INT,
                root,
                MPI_COMM_WORLD);

    if (m_myid == root) {
        recv_rcounts.resize(groupsize,0);
    }
    size = m_recv_table.size();
    MPI_Gather(&size, 1, MPI_INT, data_ptr(recv_rcounts), 1, MPI_INT, root, MPI_COMM_WORLD);

    if (m_myid == root) {
        recv_displs.resize(groupsize,0);
        recv_displs[0] = 0;
        for (iPE = 1; iPE < groupsize; iPE++) {
            recv_displs[iPE] = recv_displs[iPE-1] + recv_rcounts[iPE-1];
        }
        recv_gather.resize(recv_displs[groupsize-1] + recv_rcounts[groupsize-1], INVALID_ID);
    }

    MPI_Gatherv(data_ptr(m_recv_table),
                m_recv_table.size(),
                MPI_INT,
                data_ptr(recv_gather),
                data_ptr(recv_rcounts),
                data_ptr(recv_displs),
                MPI_INT,
                root,
                MPI_COMM_WORLD);

    if (m_myid == root) {
        FILE* fp = fopen("communication_verification.txt","w");
        D_ASSERT(fp, "Can not open file", DERROR_ERR_IO);
        //create communication matrix based on send&recv table
        //and check if it is a symmetric
        svector<uint8_t>
        matrix_sent(groupsize*groupsize,0);
        svector<uint8_t>
        matrix_rev(groupsize*groupsize,0);
        for (iPE = 0; iPE < groupsize ; iPE++){
            for (j = 0; j < send_rcounts[iPE] ; j++){
                pe_id = send_gather[send_displs[iPE]+j];
                if (pe_id>=groupsize){
                    fprintf(fp,"assert failed.pe_id=%d,groupsize=%d\n",pe_id,groupsize);
                }
                D_ASSERT_P(m_myid, pe_id < groupsize, "Out of memory", DERROR_ERR_MEM);
                matrix_sent[iPE*groupsize + pe_id] = 1;
            }
        }
        for (iPE = 0; iPE < groupsize ; iPE++){
            for (j = 0; j < recv_rcounts[iPE] ; j++){
                pe_id = recv_gather[recv_displs[iPE]+j];
                if (pe_id>=groupsize){
                    fprintf(fp,"assert failed.pe_id=%d,groupsize=%d\n",pe_id,groupsize);
                }
                D_ASSERT_P(m_myid, pe_id < groupsize, "Out of memory", DERROR_ERR_MEM);
                matrix_rev[pe_id*groupsize + iPE] = 1;
            }
        }
        //compare if the two matrix are the same.
        for (iPE = 0; iPE < groupsize ; iPE++){
            for (j = 0; j < groupsize ; j++){
                if (  matrix_sent[iPE*groupsize+j]
                   != matrix_rev[iPE*groupsize+j]){
                       fprintf(fp, "PE%d: mtx_sent[%d,%d]=%d,mtx_rev[%d,%d]=%d\n",m_myid,
                                iPE,j, matrix_sent[iPE*groupsize+j],
                                iPE,j, matrix_rev[iPE*groupsize+j]);
                       D_P_MSG(m_myid,0, "verification failed!\n");
                }
            }
        }
        D_P_MSG(m_myid,0, "###########verification passed!\n");
        fclose(fp);
    }
}

void distree_t::refine_global_distree(bool binit_features)
{
    if(binit_features)
    {
        m_features->initial_distree_feature_array(*this);
    }
    //initial 3x range feature of root node
    D_ASSERT(m_features != NULL, "m_features==NULL\n", DERROR_ERR_MEM);
    distree_node_t*
    cur_root_node   = &(m_node_array[0]);
    cur_root_node->triple_feature_indices.resize(m_features->size_feature());

    for (int i = 0; i < cur_root_node->triple_feature_indices.size() ; i++)
    {
        cur_root_node->triple_feature_indices[i] = i;
    }
    //update isovoxel of existing global_distree node in BFS order.
    int parent_end   = 1;
    int parent_start = 0;
    int child_index  = 1;

    while(parent_start < m_node_array.size())
    {
        for (int parent_id = parent_start; parent_id < parent_end ; parent_id++)
        {
            distree_node_t& cur_parent = m_node_array[parent_id];
            if (!cur_parent.isLeaf)
            {
                if (cur_parent.triple_feature_indices.size() > 0){
                    for (int j = 0; j < 8 ; j++)
                    {
                        int cur_child_index = cur_parent.m_children[j];
                        D_ASSERT_P(m_myid, cur_child_index >= 0,
                            "refine_global_distree():cur_child_index < 0", DERROR_ERR_MEM);
                        m_features->collect_feature(child_index, parent_id, *this);
                        child_index ++ ;
                    }
                }else{
                    child_index+=8;
                }
            }
        }
        parent_start = parent_end;
        parent_end   = child_index;
    }

    char msg[256];
    sprintf(msg, "parent_end=%d != m_node_array.size()=%d",parent_end, m_node_array.size());
    D_ASSERT_P(m_myid, parent_end == m_node_array.size(), msg , DERROR_ERR_MEM);
    {
    }
    int children_start_offset = (int)m_node_array.size();
    int node_size = (int)m_node_array.size();
    sprintf(msg, "global_tree_leaf_node_feature is NULL");
    m_is_building_global_tree = false;

    for (int j = 0 ; j < node_size ; j ++)
    {
        if (     m_node_array[j].isLeaf
            && (!(m_node_array[j].triple_feature_indices.empty())
                ||!(m_node_array[j].isoblock_3x_indices.empty()))
            )
        {
            int parent_start_offset = j;
            int cur_parent_index    = parent_start_offset;
            int parent_count        = 1;
            int parent_depth_id     = m_node_array[cur_parent_index].level;

            while (parent_depth_id < max_depth_id && parent_count)
            {
                int children_count = 0;
                for ( int i = 0 ; i < parent_count ; i++)
                {
                    cur_parent_index = parent_start_offset + i;
                    distree_node_t *cur_parent = &(m_node_array[cur_parent_index]);
                    if (!cur_parent->triple_feature_indices.empty())
                    {
                        bool ret = create_children(cur_parent_index, m_features);
                        children_count = ret ? children_count+8 : children_count;
                        cur_parent = &(m_node_array[cur_parent_index]);
                        cur_parent->isLeaf = !ret;
                        if (cur_parent->location == distree_node_t::LOCAL
                            && !cur_parent->isLeaf
                            && cur_parent->isoblock_3x_indices.size()>0)
                        {
                            for(int k=0 ; k < 8 ; k++)
                            {
                                child_index = cur_parent->m_children[k];
                                D_ASSERT_P(m_myid, child_index < m_node_array.size()
                                            && child_index >= 0,
                                            "refine_global_distree(): child_index out of bound" ,
                                            DERROR_ERR_MEM);
                                m_isoblock_array->collect_feature(child_index, cur_parent_index, *this);
                            }
                        }
                    }
                    parent_depth_id = cur_parent->level;
                }
                parent_count           = children_count;
                parent_start_offset    = children_start_offset;
                children_start_offset  = children_start_offset + children_count;

                D_ASSERT_P(m_myid, children_start_offset == m_node_array.size(),
                            "refine_global_distree(): children_start_offset!=m_node_array.size()" ,
                            DERROR_ERR_MEM);
            }
        }
    }
}

void distree_t::create_local_distree_vertex_array()
{
    const vec3f vertex_offset[8]= {
        vec3f(0, 0, 0), vec3f(1, 0, 0), vec3f(0, 1, 0), vec3f(1, 1, 0),
        vec3f(0, 0, 1), vec3f(1, 0, 1), vec3f(0, 1, 1), vec3f(1, 1, 1)
    };
    for (int i = 0 ; i < m_node_array.size() ; i++)
    {
        distree_node_t &leaf_node = m_node_array[i];
        if (leaf_node.isLeaf)
        {
            float extend = leaf_node.m_bound.MaximumExtentf();

            for (int j = 0; j < 8 ; j++)
            {
                m_vertex_array.push_back(distree_vertex_t());
                distree_vertex_t& cur_vertex = m_vertex_array.back();
                cur_vertex.coords = leaf_node.m_bound.pMin + vertex_offset[j]*extend;
                cur_vertex.node_index = i;
                cur_vertex.node_local_id = j;
                cur_vertex.key = compute_octree_vertex_key(cur_vertex);
            }
        }
    }
    qsort(m_vertex_array.data_ptr(),
          m_vertex_array.size(),
          sizeof(distree_vertex_t),
          distree_vertex_compare);

    int unique_count = 0;
    distree_vertex_t *cur_vertex = &(m_vertex_array[0]);
    int node_index  = cur_vertex->node_index;
    int local_id    = cur_vertex->node_local_id;
    m_vertex_array[0].eucl_dist    = FLOAT_MAX;
    m_vertex_array[0].maxnorm_dist = FLOAT_MAX;
    m_vertex_array[0].is_send      = 0;
    distree_node_t *cur_node = &(m_node_array[node_index]);
    cur_node->vertex_indices[local_id] = 0;

    int vertex_count = m_vertex_array.size();
    for (int vtx_idx = 1; vtx_idx < vertex_count ; vtx_idx++) {
        cur_vertex = &(m_vertex_array[vtx_idx]);
        node_index = cur_vertex->node_index;
        local_id = cur_vertex->node_local_id;
        if (m_vertex_array[unique_count].key
            != cur_vertex->key)
        {
            unique_count++;
            m_vertex_array[unique_count] = *cur_vertex;
            m_vertex_array[unique_count].eucl_dist    = FLOAT_MAX;
            m_vertex_array[unique_count].maxnorm_dist = FLOAT_MAX;
            m_vertex_array[unique_count].is_send      = 0;
        }
        cur_node = &(m_node_array[node_index]);
        cur_node->vertex_indices[local_id] = unique_count;
    }
    unique_count++;
    m_vertex_array.resize(unique_count);
}

void distree_t::save_distree_text(const char* treeName){
    char name[1024];
    FILE *fp = NULL;
    distree_node_t   *cur_node=NULL;
    sprintf(name,"%s%d.txt",treeName,m_myid);
    fp = fopen(name,"w");
    D_ASSERT_P(m_myid, fp, "Can not open file", DERROR_ERR_IO);
  
    int node_array_size = m_node_array.size();
    fprintf(fp,"node_size=%d\n", node_array_size);
    for (int i = 0; i < node_array_size ; i++){
        cur_node = &(m_node_array[i]);
        vec3f center = cur_node->m_bound.Center();
        fprintf(fp,"node%d\n\
                   l=%d,leaf=%d, p=%d\n\
                   k=%d,location=%d\n\
                   %f, %f, %f\n\
                   %f, %f, %f\n\
                   %f, %f, %f\n\
                   %d, %d, %d, %d\n\
                   %d, %d, %d, %d\n\
                   %d, %d, %d, %d\n\
                   %d, %d, %d, %d\n\
                   \n",i
                   ,cur_node->level,cur_node->isLeaf, cur_node->parent
                   ,cur_node->key,cur_node->location
                   ,cur_node->m_bound.pMin.x(),cur_node->m_bound.pMin.y(),cur_node->m_bound.pMin.z()
                   ,cur_node->m_bound.pMax.x(),cur_node->m_bound.pMax.y(),cur_node->m_bound.pMax.z()
                   ,center.x(), center.y(), center.z()
                   ,cur_node->m_children[0],cur_node->m_children[1],cur_node->m_children[2],cur_node->m_children[3]
                   ,cur_node->m_children[4],cur_node->m_children[5],cur_node->m_children[6],cur_node->m_children[7]

                   ,cur_node->vertex_indices[0],cur_node->vertex_indices[1],cur_node->vertex_indices[2],cur_node->vertex_indices[3]
                   ,cur_node->vertex_indices[3],cur_node->vertex_indices[5],cur_node->vertex_indices[6],cur_node->vertex_indices[7]);
    }

    fclose(fp);
}

void distree_t::compute_vertex_distance_parallel()
{
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(parallel_test::DIST_LOCAL);
    HPGV_TIMING_COUNT(parallel_test::DIST_LOCAL);
    compute_local_distance();
    HPGV_TIMING_END(parallel_test::DIST_LOCAL);

    if (m_myid == 0) {
        fprintf(stderr, "^^^ PE%d pass compute_local_distance ^^^^\n", m_myid);
    }
    
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(parallel_test::COM_MEM_TIME);
    int send_count = m_sent_table.size();
    
    MPI_Status status;
    svector<recv_schedule_t> recv_schedule;
    svector<MPI_Status>  my_send_status;
    svector<MPI_Status>  my_recv_status;
    svector<MPI_Request> my_send_request;
    svector<MPI_Request> my_recv_request;
    svector<MPI_Request> dummy_send_request;
    svector<MPI_Request> dummy_recv_request;
    char msg[256];
    if (send_count>0){
        m_send_schedule.resize(send_count);
        my_send_status.resize(send_count);
        my_send_request.resize(send_count);
        dummy_send_request.resize(send_count);
    }
    HPGV_TIMING_END(parallel_test::COM_MEM_TIME);

    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(parallel_test::PREPARE_SEND_BUFFER);
    HPGV_TIMING_COUNT(parallel_test::PREPARE_SEND_BUFFER);

    create_send_schedule_vertex(m_send_schedule);

    HPGV_TIMING_END(parallel_test::PREPARE_SEND_BUFFER);
    
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    {
        D_P_MSG(m_myid, 0
            ,"++++++++++pass create_send_schedule_vertex^^^^^\n");
    }
    HPGV_TIMING_BEGIN(parallel_test::COM_MEM_TIME);
    svector<int> send_size_array(m_send_schedule.size(),-1);
    for (int i = 0; i < send_count ; i++)
    {
        send_size_array[i] =  m_send_schedule[i].send_key_buf.size();
        int err = MPI_Isend(
                    &(send_size_array[i]),
                    1,
                    MPI_INT,
                    m_send_schedule[i].recv_id,
                    MY_DUMMY_MSG,
                    MPI_COMM_WORLD,
                    &(dummy_send_request[i]));
        if(err != MPI_SUCCESS)
        {
            sprintf(msg, "==== err=%d ===\n");
            D_P_MSG(m_myid, m_myid,msg);
        }
        
    }
    int recv_remote_proc_count = m_recv_table.size();
    svector<int> reverse_lookup_table(m_groupsize, -1);
    svector<int> recv_size(recv_remote_proc_count,0);

    if (recv_remote_proc_count > 0)
    {
        dummy_recv_request.resize(recv_remote_proc_count, 0);
        recv_schedule.resize(recv_remote_proc_count);
        for (int i = 0; i < recv_remote_proc_count; i++) {
            recv_schedule[i].send_id = m_recv_table[i];
            reverse_lookup_table[recv_schedule[i].send_id] = i;
        }
        int count = 0, flag = 0;
        while(count < recv_remote_proc_count)
        {
            MPI_Iprobe(MPI_ANY_SOURCE, MY_DUMMY_MSG, MPI_COMM_WORLD,
                        &flag, &status);
            if (flag)
            {
                int recv_schedule_idx = reverse_lookup_table[status.MPI_SOURCE];
                if (status.MPI_SOURCE != recv_schedule[recv_schedule_idx].send_id){
                    char msg[256];
                    sprintf(msg,"status.MPI_SOURCE(%d)!=recv_schedule[].send_id(%d)"
                           ,status.MPI_SOURCE, recv_schedule[recv_schedule_idx].send_id);
                    D_ASSERT_P(m_myid, status.MPI_SOURCE == recv_schedule[recv_schedule_idx].send_id,
                               msg, DERROR_ERR_COMM);
                }

                int size = 0;
                MPI_Get_count(&status, MPI_INT, &size );

                if (size > 0){
                    MPI_Irecv(&(recv_size[recv_schedule_idx]),
                                size,
                                MPI_INT,
                                status.MPI_SOURCE,
                                MY_DUMMY_MSG,
                                MPI_COMM_WORLD,
                                &(dummy_recv_request[recv_schedule_idx]));
                }
                count++;
            }
        }
    }
    MPI_Waitall(send_count, data_ptr(dummy_send_request), MPI_STATUSES_IGNORE);
    MPI_Waitall(recv_remote_proc_count, data_ptr(dummy_recv_request), MPI_STATUSES_IGNORE);
    {
        D_P_MSG(m_myid, 0
            ,"++++++++++pass send vertex buffer size^^^^^\n");
    }

    dummy_recv_request.clear();
    for (int i = 0; i < recv_remote_proc_count ; i++){
        recv_schedule[i].recv_buf.resize(recv_size[i]);
    }
    my_recv_request.resize(recv_remote_proc_count);

    HPGV_TIMING_COUNT(parallel_test::COM_MEM_TIME);
    HPGV_TIMING_END(parallel_test::COM_MEM_TIME);
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(parallel_test::COMM_PHASE_1);
    HPGV_TIMING_COUNT(parallel_test::COMM_PHASE_1);
    HPGV_TIMING_COUNT(parallel_test::MY_SEND_BYTE1);
    for (int i = 0; i < send_count ; i++)
    {
        MPI_Isend(data_ptr(m_send_schedule[i].send_key_buf),
                  m_send_schedule[i].send_key_buf.size(),
                  MPI_INT,
                  m_send_schedule[i].recv_id,
                  MY_SEND_MSG,
                  MPI_COMM_WORLD,
                  &(my_send_request[i]));

        HPGV_TIMING_INCREASE(parallel_test::MY_SEND_BYTE1,
                    m_send_schedule[i].send_key_buf.size()*sizeof(int));
    }
    if (recv_remote_proc_count > 0)
    {
        HPGV_TIMING_COUNT(parallel_test::MY_RECV_BYTE1);
        int count = 0, flag = 0;
        while(count < recv_remote_proc_count)
        {
            MPI_Iprobe(MPI_ANY_SOURCE, MY_SEND_MSG, MPI_COMM_WORLD,
                        &flag, &status);
            if (flag)
            {
                int recv_schedule_idx = reverse_lookup_table[status.MPI_SOURCE];
                D_ASSERT_P(m_myid, status.MPI_SOURCE == recv_schedule[recv_schedule_idx].send_id,
                            "**MPI_SOURCE!=send_id**\n", DERROR_ERR_MEM);
                int isize = 0;
                MPI_Get_count(&status, MPI_INT, &isize );

                if (isize > 0){
                    if(recv_schedule[recv_schedule_idx].recv_buf.size() != isize){
                        char err_msg[256];
                        sprintf(err_msg, "***Proc%d receive recv_schel[%d].recv_buf.size=%d != size=%d from PE%d***\n"
                            ,m_myid, recv_schedule_idx, recv_schedule[recv_schedule_idx].recv_buf.size(), isize
                            ,recv_schedule[recv_schedule_idx].send_id);

                        D_ASSERT_P(m_myid, recv_schedule[recv_schedule_idx].recv_buf.size() == isize,
                                   err_msg, DERROR_ERR_COMM);
                    }

                    recv_schedule[recv_schedule_idx].recv_buf.resize(isize);
                    MPI_Irecv(data_ptr(recv_schedule[recv_schedule_idx].recv_buf),
                              isize,
                              MPI_INT,
                              status.MPI_SOURCE,
                              MY_SEND_MSG,
                              MPI_COMM_WORLD,
                              &(my_recv_request[recv_schedule_idx]));

                    HPGV_TIMING_INCREASE(parallel_test::MY_RECV_BYTE1,
                        recv_schedule[recv_schedule_idx].recv_buf.size()*sizeof(int));
                }
                count++;
            }
        }
    }

    MPI_Waitall(send_count, data_ptr(my_send_request), MPI_STATUSES_IGNORE);
    MPI_Waitall(recv_remote_proc_count, data_ptr(my_recv_request), MPI_STATUSES_IGNORE);
    HPGV_TIMING_END(parallel_test::COMM_PHASE_1);
    
    {
        D_P_MSG(m_myid, 0
            ,"++++++++++pass send vertex to remote^^^^^\n");
    }
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(parallel_test::COMPUT_REMOTE_DIST);
    HPGV_TIMING_COUNT(parallel_test::COMPUT_REMOTE_DIST);
    int nTotalRemoteVertices = 0;
    vec3f coord(0.0f);
    int nBoundary=0;
    int n=0;
    int i=0, j=0;
    for (int i = 0; i < recv_remote_proc_count; i++)
    {
        svector<int> &remote_vertex_keys = recv_schedule[i].recv_buf;
        int vertex_key_count = remote_vertex_keys.size();
        nTotalRemoteVertices += vertex_key_count;
        for (int j=1 ; j < vertex_key_count ; j++)
        {
            int cur_vtx_key = remote_vertex_keys[j];
            compute_from_key_to_coords(cur_vtx_key, coord);
            distree_node_t* cur_node = get_distree_node(coord);
            float eucl_dist = FLOAT_MAX, norm_dist = FLOAT_MAX;
            D_ASSERT_P(m_myid, cur_node, "compute_vertex_distance_parallel:cur_node==NULL!\n", DERROR_ERR_MEM);
            if (cur_node)
            {
                distree_node_t* parent_node = cur_node;
                int generation = 0;
                while(parent_node->triple_feature_indices.empty()
                    &&parent_node->parent>=0)
                {
                    parent_node = &(m_node_array[parent_node->parent]);
                    generation ++;
                }
                compute_vertex_node_feature_distance(coord, *parent_node
                                        ,eucl_dist,norm_dist, generation);
            }else
            {
                nBoundary++;
            }
            float* remote_distance = (float*)&(remote_vertex_keys[j]);
            *remote_distance = eucl_dist;
        }
    }
   
    HPGV_TIMING_END(parallel_test::COMPUT_REMOTE_DIST);
    //--------END of Computing the distance from remote vertices to the local voxel----------------
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    {
        D_P_MSG(m_myid, 0
            ,"++++++++++pass compute remote vertex ^^^^^\n");
    }

    HPGV_TIMING_BEGIN(parallel_test::COMM_PHASE_2);
    HPGV_TIMING_COUNT(parallel_test::COMM_PHASE_2);
    HPGV_TIMING_COUNT(parallel_test::MY_SEND_BYTE2);
    for (int i = 0; i < recv_remote_proc_count; i++)
    {
        MPI_Isend(data_ptr(recv_schedule[i].recv_buf),
                  recv_schedule[i].recv_buf.size(),
                  MPI_INT,
                  recv_schedule[i].send_id,
                  MY_REPLY_MSG,
                  MPI_COMM_WORLD,
                  &(my_recv_request[i]));

        HPGV_TIMING_INCREASE(parallel_test::MY_SEND_BYTE2,
                   recv_schedule[i].recv_buf.size()*sizeof(int));
    }
    HPGV_TIMING_COUNT(parallel_test::MY_RECV_BYTE2);
    if (send_count > 0)
    {
        for (int i = 0; i < reverse_lookup_table.size(); i++) {
            reverse_lookup_table[i] = -1;
        }

        for (int i = 0; i < send_count; i++) {
            reverse_lookup_table[m_send_schedule[i].recv_id] = i;
        }
        int count=0, flag=0;
        while(count < send_count)
        {
            MPI_Iprobe(MPI_ANY_SOURCE, MY_REPLY_MSG, MPI_COMM_WORLD,
                       &flag, &status);
            if (flag)
            {
                int send_key_idx = reverse_lookup_table[status.MPI_SOURCE];
                D_ASSERT_P(m_myid, status.MPI_SOURCE == m_send_schedule[send_key_idx].recv_id,
                            "refine_global_distree(): **MPI_SOURCE!=send_id**\n", DERROR_ERR_MEM);
                int size=0;
                MPI_Get_count(&status, MPI_INT, &size );
                if (size > 0){
                    D_ASSERT_P(m_myid, m_send_schedule[send_key_idx].send_key_buf.size()==size,
                        "compute_vertex_distance_parallel(): The # of your received remote distance != the # of your send keys!\n",DERROR_ERR_MEM);

                    MPI_Irecv(data_ptr(m_send_schedule[send_key_idx].send_key_buf),
                              size,
                              MPI_INT,
                              status.MPI_SOURCE,
                              MY_REPLY_MSG,
                              MPI_COMM_WORLD,
                              &(my_send_request[send_key_idx]));
                    HPGV_TIMING_INCREASE(parallel_test::MY_RECV_BYTE2,
                        m_send_schedule[send_key_idx].send_key_buf.size()*sizeof(int));
                    count++;
                }
            }
        }
    }
    MPI_Waitall(send_count, data_ptr(my_send_request), MPI_STATUSES_IGNORE);
    MPI_Waitall(recv_remote_proc_count, data_ptr(my_recv_request), MPI_STATUSES_IGNORE);
    HPGV_TIMING_END(parallel_test::COMM_PHASE_2);
    HPGV_TIMING_BARRIER( MPI_COMM_WORLD );
    HPGV_TIMING_BEGIN(parallel_test::UPDATE_DIST);
    HPGV_TIMING_COUNT(parallel_test::UPDATE_DIST);
    for (int i = 0; i < send_count ; i++){
        if (m_send_schedule[i].send_key_buf.size() > 1){
            float* remote_vertex_dist_array =
                (float*)data_ptr(m_send_schedule[i].send_key_buf);
            remote_vertex_dist_array++;
            int vertex_count = m_send_schedule[i].send_key_buf.size() - 1;
            for (int j = 0; j < vertex_count ; j++)
            {
                int cur_index = m_send_schedule[i].send_idx_buf[j];
                distree_vertex_t& cur_vtx = m_vertex_array[cur_index];
                float remote_distance = remote_vertex_dist_array[j];
                if (fabsf(remote_distance) < fabsf(cur_vtx.eucl_dist))
                {
                    cur_vtx.eucl_dist = remote_distance;
                }
            }
        }
    }
    HPGV_TIMING_END(parallel_test::UPDATE_DIST);
}

void distree_t::compute_local_distance()
{
    int node_array_size = m_node_array.size();
    int i=0;
    svector<distree_node_t>* p_node_array = &m_node_array;
#ifdef _OPENMP
#pragma omp parallel for default(none), private(i), shared(p_node_array, node_array_size)
#endif
    for (i = 0; i < node_array_size ; i++)
    {  
        distree_node_t &child_node = (*p_node_array)[i];
        if (   child_node.isLeaf
            && child_node.location == distree_node_t::LOCAL
            )
        {
            distree_node_t &parent_node = m_node_array[child_node.parent];
            compute_child_vertex_distance(child_node, parent_node);
        }
    }
}

void distree_t::compute_child_vertex_distance( distree_node_t & child_node,
                                               distree_node_t & parent_node )
{
    int generation = ITSELF;
    distree_node_t* feature_node = &child_node;
    while( feature_node->triple_feature_indices.size() <= 0
        && feature_node->parent >= 0
        && feature_node->level > 0
        )
    {
        feature_node = &(m_node_array[feature_node->parent]);
        generation++;
    }

    for (int vtx_id = 0; vtx_id < 8 ; vtx_id++)
    {
        float eucl_dist = FLOAT_MAX;
        float norm_dist = FLOAT_MAX;
        
        int vtx_idx = child_node.vertex_indices[vtx_id];
        if (!(vtx_idx >= 0 && vtx_idx < m_vertex_array.size())){
            D_ASSERT_P(m_myid, vtx_idx >= 0 && vtx_idx < m_vertex_array.size(),
                        "vertex index out of bounds", DERROR_ERR_MEM);
        }

        distree_vertex_t& cur_vertex = m_vertex_array[vtx_idx];
        compute_vertex_node_feature_distance(cur_vertex, *feature_node, eucl_dist, norm_dist, -1);

        if (fabsf(cur_vertex.eucl_dist) > fabsf(eucl_dist))
        {
            cur_vertex.eucl_dist = eucl_dist;
            cur_vertex.min_dist_type = generation;
        }

        if (fabsf(cur_vertex.maxnorm_dist) > fabsf(norm_dist))
        {
            cur_vertex.maxnorm_dist = norm_dist;
        }

        int level = cur_vertex.min_dist_type;
        int child_local_id = compute_local_child_key(child_node);
        while ( eucl_dist > 2*feature_node->m_bound.SphericalRadius()
             && feature_node->parent >= 0 
             && level++ < 3)
        {
            distree_node_t &parent = m_node_array[feature_node->parent];
            compute_vertex_node_feature_distance(cur_vertex, parent, eucl_dist,
                                                 norm_dist, child_local_id);
            if (fabsf(cur_vertex.eucl_dist) > fabsf(eucl_dist))
            {
                cur_vertex.eucl_dist = eucl_dist;
                cur_vertex.min_dist_type = generation + 1;
            }
            if (fabsf(cur_vertex.maxnorm_dist) > fabsf(norm_dist))
            {
                cur_vertex.maxnorm_dist = norm_dist;
            }
        }
    }
}
void distree_t::compute_vertex_node_feature_distance( vec3f& coords,
                                                      distree_node_t& feature_node,
                                                      float& eucl_dist,
                                                      float& norm_dist, int compute_type/*=-1*/ )
{
    svector<int> *feature_index_array;
    feature_index_array = &(feature_node.triple_feature_indices);
    float min_edist = FLOAT_MAX;
    for (int i = 0; i < feature_index_array->size() ; i++){
        D_ASSERT_P(m_myid, (*feature_index_array)[i] < m_features->size_feature(),
            "Memory overflow", DERROR_ERR_MEM);

        float edist = m_features->compute_vertex_feature_distance(coords,
                                       (*feature_index_array)[i] , *this);

        if( fabsf(min_edist) > fabsf(edist)) {
            min_edist = edist;
        }
    }
    eucl_dist = min_edist;
}
void distree_t::compute_vertex_node_feature_distance( distree_vertex_t& cur_vertex,
                                                      distree_node_t& feature_node,
                                                      float& eucl_dist,
                                                      float& norm_dist, int compute_type/*=-1*/ )
{
    svector<int> *feature_index_array;
    feature_index_array = &(feature_node.triple_feature_indices);
   
    float min_edist = FLOAT_MAX;
    for (int i = 0; i < feature_index_array->size() ; i++){
        D_ASSERT_P(m_myid, (*feature_index_array)[i] < m_features->size_feature(),
                    "Memory overflow", DERROR_ERR_MEM);

        float edist = m_features->compute_vertex_feature_distance(cur_vertex,
                                            (*feature_index_array)[i], *this);
        
        if( fabsf(min_edist) > fabsf(edist)) {
            min_edist = edist;
        }
    }
    eucl_dist = min_edist;
}

void distree_t::create_send_schedule_vertex( svector<send_schedule_t>& send_schedule )
{
    svector<int> reverse_lookup_table(m_groupsize,-1);
    const int send_count = send_schedule.size();

    for (int i = 0; i < send_count ; i++)
    {
        send_schedule[i].recv_id = m_sent_table[i];
        send_schedule[i].send_idx_buf.reserve(m_vertex_array.size());
        reverse_lookup_table[send_schedule[i].recv_id] = i;
    }

    const int node_size = m_node_array.size();

    bool track=false;
    char msg[256];
    for (int i = 0; i < node_size; i++)
    {
        distree_node_t& cur_node = m_node_array[i];
        if (cur_node.isLeaf && cur_node.location == distree_node_t::LOCAL)
        {
            for (int j = 0; j < 8 ; j++)
            {
                int cur_vtx_idx = cur_node.vertex_indices[j];
                distree_vertex_t& cur_vertex = m_vertex_array[cur_vtx_idx];
                if (cur_vertex.min_dist_type >= 3 || cur_vertex.min_dist_type<-1){
                    char msg_err[256];
                    sprintf(msg_err, "+++---min_dist_type=%d-vtx_idx=%d--+++",cur_vertex.min_dist_type
                        ,cur_vtx_idx);
                    D_P_MSG(m_myid, m_myid,msg_err);
                }
                
                distree_node_t* ancestor_node = &cur_node;
                if (cur_vertex.min_dist_type == -1)
                {
                    while(ancestor_node->processor_3x_indices.empty()
                        && ancestor_node->parent > 0 )
                    {
                        sprintf(msg, "ancestor_node->parent=%d out of bound.\n", ancestor_node->parent);
                                   D_ASSERT_P(m_myid, ancestor_node->parent< m_node_array.size()
                                    && ancestor_node->parent>=0, msg, DERROR_ERR_MEM);
                        ancestor_node = &(m_node_array[ancestor_node->parent]);
                    }
                }
                else
                {
                    int level = cur_vertex.min_dist_type;
                    while(level>0)
                    {
                        sprintf(msg, "ancestor_node->parent=%d out of bound.\n", ancestor_node->parent);
                                     D_ASSERT_P(m_myid, ancestor_node->parent< m_node_array.size()
                                      && ancestor_node->parent>=0, msg, DERROR_ERR_MEM);
                        ancestor_node = &(m_node_array[ancestor_node->parent]);
                        --level;
                    }
                }

                const int proc_count = ancestor_node->processor_3x_indices.size();
                
                for (int k = 0; k < proc_count; k++)
                {
                    int send_schedule_index =
                        reverse_lookup_table[ancestor_node->processor_3x_indices[k]];
                    
                    if (send_schedule_index > -1)
                    {
                        sprintf(msg, "ancestor_node->parent=%d out of bound.\n", ancestor_node->parent);
                                      D_ASSERT_P(m_myid, send_schedule_index >=0
                                      && send_schedule_index < send_schedule.size(), msg, DERROR_ERR_MEM);

                        send_schedule_t& cur_send_schedule =
                            send_schedule[send_schedule_index];

                        cur_send_schedule.send_idx_buf.push_back(cur_vtx_idx);
                        
                    }
                }
            }
        }
    }

    for (int i = 0; i < send_count ; i++)
    {
        send_schedule[i].send_idx_buf.shrink_to_fit();
    }
    for (int i = 0; i < send_count ; i++)
    {
        svector<int> &cur_send_idx_buf = send_schedule[i].send_idx_buf;
        cur_send_idx_buf.remove_duplicates(int_compare);
        
        int unique_count = cur_send_idx_buf.size();
        svector<uint32_t> &cur_send_key_buf = send_schedule[i].send_key_buf;

        cur_send_key_buf.resize(unique_count+1);
        cur_send_key_buf[0] = unique_count;
        
        for (int j = 0; j < unique_count ; j++){
            m_vertex_array[cur_send_idx_buf[j]].is_send = 1;
            cur_send_key_buf[j+1] = m_vertex_array[cur_send_idx_buf[j]].key;
        }
    }
    fflush(stderr);
}
int compute_key_0(float x, float y, float z, const vec3f &center){
    int key;
    key =   (( z > center.z()) << 2) +
        (( y > center.y()) << 1) +
        (( x > center.x()) << 0);

    return key;
}

int compute_key_1(float x, float y, float z, const vec3f &center){
    int key;
    key= (( z >  center.z()) << 2) +
        (( y >  center.y()) << 1) +
        (( x >= center.x()) << 0);

    return key;
}

int compute_key_2(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >  center.z()) << 2) +
        (( y >= center.y()) << 1) +
        (( x >  center.x()) << 0);

    return key;
}

int compute_key_3(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >  center.z()) << 2) +
        (( y >= center.y()) << 1) +
        (( x >= center.x()) << 0);

    return key;
}

int compute_key_4(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >= center.z()) << 2) +
        (( y >  center.y()) << 1) +
        (( x >  center.x()) << 0);

    return key;
}

int compute_key_5(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >= center.z()) << 2) +
        (( y >  center.y()) << 1) +
        (( x >= center.x()) << 0);

    return key;
}

int compute_key_6(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >= center.z()) << 2) +
        (( y >= center.y()) << 1) +
        (( x >  center.x()) << 0);

    return key;
}

int compute_key_7(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >= center.z()) << 2) +
          (( y >= center.y()) << 1) +
          (( x >= center.x()) << 0);

    return key;
}
typedef int   (*compute_octree_key_fun_t)(float x, float y, float z, const vec3f &center);

compute_octree_key_fun_t g_key_fun_tables[8]={
    compute_key_0, compute_key_7, compute_key_2, compute_key_3,
    compute_key_4, compute_key_5, compute_key_6, compute_key_1
};

inline distree_node_t* distree_t::get_distree_node(const vec3f& coord )
{
    distree_node_t *node = &(m_node_array[root_index]);
    distree_node_t *prevNode = NULL;
    while( node
        //&& node->location == distree_node_t::LOCAL
        && !node->isLeaf)
    {//iterative refining node and keep searching
        //compute the center of the node's bounding box.
        vec3f center = node->m_bound.Center();
        int localkey = (( coord.z() > center.z()) << 2) +
                       (( coord.y() > center.y()) << 1) +
                       (( coord.x() > center.x()) << 0);
        prevNode = node;
        node = get_node_child(*node, localkey);
    }
    if (!node)
    {
        fprintf(stderr, "PE%d get NULL leaf node. prevNode.key=%d,\
.level=%d isleaf=%d child[0]=%d triple_size=%d extend=%f\n"
            , m_myid, prevNode->key, prevNode->level, prevNode->isLeaf, 
            prevNode->m_children[0], prevNode->triple_feature_indices.size()
            , prevNode->m_bound.getDimension().max_extend1f());
        fflush(stderr);
    }

    return node;
}
distree_node_t* distree_t::get_distree_local_node(const vec3f& coord, bool debug )
{
    distree_node_t *node = &(m_node_array[root_index]);

    svector<int> visited_node_key;//
    while( node
        && node->location == distree_node_t::LOCAL
        && !node->isLeaf)
    {//Iteratively refining node and keep searching
        //compute the center of the node's bounding box.
        vec3f center = node->m_bound.Center();
        int i = 0;
        distree_node_t* tmpnode = node;
        int localkey;

        do{//Make sure we choose the LOCAL leaf node.
           //In boundary case, get_leaf_node() would give you a REMOTE leaf node.
            //int j = i;
            localkey = g_key_fun_tables[i](coord.x(),coord.y(),coord.z(), center);
            tmpnode = get_node_child(*node, localkey);
            i++;
        } while
            (i < 8
            && (tmpnode->location != distree_node_t::LOCAL //REMOTE node is not acceptable
            || visited_node_key.contain(tmpnode->key)) //the branch rooted at tmpnode is already visited and doesn't contains the query point on the node boundary
            );

        if (i < 8 || tmpnode->location == distree_node_t::LOCAL)
        {
            node = tmpnode;
        }
        else
        {//equivalent to if(i>=8 && tmpnode->location == distree_node_t::REMOTE)
            //Special case: when the query position is on the node boundary.
            //It's very likely that we have found the wrong branch even though
            //the root node of the branch indicates that its location is LOCAL
            //Then we should back track to the parent node and start another search
            //from its other branches.
            visited_node_key.push_back(node->key);
            node = &(m_node_array[node->parent]);
        }
    }
    return node;
}

distree_node_t* distree_t::get_node_child(const distree_node_t &parent_node, int childid)
{
    int child_index = parent_node.m_children[childid] ;
    if (child_index == -1) {
        //parent_node is leaf node. no more children.
        return NULL;
    }
    return &(m_node_array[child_index]);
}

void distree_t::create_distance_volume()
{
    int max_assigned_local_leaf_num = get_max_local_region_num();

    int  i = 0;
    for (i = 0; i < m_assigned_local_leaf_indices.size() ; i++)
    {
        distree_node_t& cur_region = m_node_array[m_assigned_local_leaf_indices[i]];
        BBox cur_region_bound  = cur_region.m_bound;
        //cur_region_size should be integer value, but floating point computation would create a tiny error,
        // so we need to cast the value returned by getDimension() to vec3i and then back to vec3f.
        vec3f          cur_region_size   = vec3f(vec3i(cur_region_bound.getDimension()));
        vec3i          cur_distvol_bound = vec3i(cur_region_size/sample_size + 0.5f);

        volume_t       *dist_volume      = new volume_t(cur_distvol_bound);
        dist_volume->m_bound.pMin
            = vec3i((cur_region_bound.pMin - m_global_domain_bound.pMin)/sample_size + 0.5f);
        dist_volume->m_bound.pMax
            = vec3i((cur_region_bound.pMax - m_global_domain_bound.pMin)/sample_size + 0.5f);
        vec3f& offset = cur_region_bound.pMin;
        //compute distance value voxel by voxel
        float maxZ = cur_region_size.z();
        float maxY = cur_region_size.y();
        float maxX = cur_region_size.x();
        for (float z = 0; z < maxZ; z+=sample_size.z()) {
            for (float y = 0; y < maxY; y+=sample_size.y()) {
                for (float x = 0; x < maxX; x+=sample_size.x()) {
//                     HPGV_TIMING_BEGIN(COMPUTE_DIST_VOL_STEP_TIME);
//                     HPGV_TIMING_COUNT(COMPUTE_DIST_VOL_STEP_TIME);
                    float gx = x + offset.x();
                    float gy = y + offset.y();
                    float gz = z + offset.z();
                    float v  = get_tree_value_local(gx, gy, gz);

                    dist_volume->set_volume_value(
                                    x/sample_size.x(),
                                    y/sample_size.y(),
                                    z/sample_size.z(), v);
//                    HPGV_TIMING_END(COMPUTE_DIST_VOL_STEP_TIME);
                }
            }
        }// end of loop through local distance volume voxel.
        m_distance_volumes.push_back(dist_volume);
    }// end of for each assigned local leaf node.
    //create dummy distance volume if
    // max_assigned_local_leaf_num > m_assigned_local_leaf_indices.size()
    // in order to prevent the subsequent collective IO write from getting stuck.
    for (   ; i < max_assigned_local_leaf_num ; i++){
        volume_t *dist_volume = new volume_t(1,1,1);
        dist_volume->m_bound = BBox(vec3f(0.0f),vec3f(1.0f));
        m_distance_volumes.push_back(dist_volume);
    }
}

int distree_t::get_max_local_region_num() const
{
    int my_num = m_assigned_local_leaf_indices.size();
    int max_num = 0;
    MPI_Allreduce(&my_num, &max_num, 1, MPI_INT, MPI_MAX, m_communicator);
    return max_num;
}
float distree_t::get_tree_value_global(float x, float y, float z, bool debug)
{
    const vec3f pos(x,y,z);
    const distree_node_t *node = get_distree_node(pos);
    const int *cur_vertex_indices = node->vertex_indices;
    float vtx_dist_value[8];

    if (node != NULL ){
        //trilinearly interpolate the distance value on the 8 vertices.
        //compute the offset[0,1] of pos relative to the bbox of node.
        vec3f txyz = node->m_bound.Offset(pos);
        for (int i = 0; i < 8 ; i++)
        {
            vtx_dist_value[i] = fabsf(m_vertex_array[cur_vertex_indices[i]].eucl_dist);
        }
        return tlerp(vtx_dist_value, txyz);
    }else {
        //boundary case, simply scan the vertex array and find the nearest vertex.
        //and return the vertex distance.
        return 0.0f;
    }
}
float distree_t::get_tree_value_local(float x, float y, float z, bool debug)
{
    const vec3f pos(x,y,z);
    const distree_node_t *node = get_distree_local_node(pos, debug);
    const int *cur_vertex_indices = node->vertex_indices;
    float vtx_dist_value[8];

    if (node != NULL ){
        //trilinearly interpolate the distance value on the 8 vertices.
        //compute the offset[0,1] of pos relative to the bbox of node.
        vec3f txyz = node->m_bound.Offset(pos);
        for (int i = 0; i < 8 ; i++)
        {
            vtx_dist_value[i] = fabsf(m_vertex_array[cur_vertex_indices[i]].eucl_dist);
        }
        return tlerp(vtx_dist_value, txyz);
    }else {
        //boundary case, simply scan the vertex array and find the nearest vertex.
        //and return the vertex distance.
        return 0.0f;
    }
}

void distree_t::save_distance_volume_parallel(const char* file_name )
{
    MPI_File fh;
    int ret = MPI_File_open(m_communicator,
        const_cast<char*>(file_name),
        MPI_MODE_CREATE | MPI_MODE_WRONLY,
        MPI_INFO_NULL,
        &fh);
    D_ASSERT_P(m_myid,ret == MPI_SUCCESS, "file open failed!\n", DERROR_ERR_IO);

    MPI_Datatype filetype;
    /*printf("**** dist_volume size= %d, assign_local_leaf size=%d***\n"
        ,m_distance_volumes.size(), m_assigned_local_leaf_indices.size());*/
    for (int i = 0; i < m_distance_volumes.size() ; i++)
    {//save each local distance volumes.
        volume_t* cur_dist_volume = m_distance_volumes[i];
        vec3i gsizes = vec3i(vec3f(m_global_domain_size)/sample_size);//m_global_domain_size;
        vec3i psizes = m_np_xyz;
        vec3i lsizes = cur_dist_volume->d_size;
        vec3i start_indices = vec3i(cur_dist_volume->m_bound.pMin);
        if (gsizes.x() <= 0 || gsizes.y() <= 0 || gsizes.z() <= 0
            || psizes.x() <= 0 || psizes.y() <= 0 || psizes.z() <=0
            || lsizes.x() <= 0 || lsizes.y() <= 0 || lsizes.z() <= 0
            || start_indices.x() < 0 || start_indices.y()<0 || start_indices.z()<0){
                cout<<"-----PE"<<m_myid<<"----gsize="<<gsizes<<endl;
                cout<<"-----PE"<<m_myid<<"----psize="<<psizes<<endl;
                cout<<"-----PE"<<m_myid<<"----lsize="<<lsizes<<endl;
                cout<<"-----PE"<<m_myid<<"--dist_volume["<<i<<"]--"
                    <<m_assigned_local_leaf_indices.size()<<"/"
                    <<"---start_indices="<<start_indices<<endl;
        }

        MPI_Type_create_subarray(3, (int*)gsizes, (int*)lsizes, (int*)start_indices,
                                 MPI_ORDER_FORTRAN, MPI_FLOAT, &filetype);
        MPI_Type_commit(&filetype);

        MPI_File_set_view(fh, 0, MPI_FLOAT, filetype, "native", MPI_INFO_NULL);

        int local_array_size = lsizes[0] * lsizes[1] * lsizes[2];
        float* buffer = data_ptr(cur_dist_volume->m_data);

        MPI_File_write_all(fh, buffer, local_array_size, MPI_FLOAT, MPI_STATUS_IGNORE);
    }

    MPI_File_close(&fh);
}

void distree_t::create_distance_volume_allreduce()
{
    vec3i gsizes = vec3i(vec3f(m_global_domain_size)/sample_size);
    int global_voxel_count = gsizes.x() * gsizes.y() * gsizes.z();
    m_distance_vol.resize(global_voxel_count, 3333333.0f/*FLOAT_MAX*/);
    int  ib = 0;
    for (ib = 0; ib < m_assigned_local_leaf_indices.size() ; ib++)
    {
        distree_node_t& cur_region = m_node_array[m_assigned_local_leaf_indices[ib]];
        BBox cur_region_bound  = cur_region.m_bound;
        vec3f  cur_region_size   = vec3f(vec3i(cur_region_bound.getDimension()));
        vec3f& offset = cur_region_bound.pMin;

        vec3i start_indices = vec3i((cur_region_bound.pMin - m_global_domain_bound.pMin)/sample_size + 0.5f);

        float maxZ = cur_region_size.z();
        float maxY = cur_region_size.y();
        float maxX = cur_region_size.x();
        float x,y,z; int ix,iy,iz;
        int m=0;
        for (z = 0, iz=0; z < maxZ; z+=sample_size.z(), iz++) {
            for (y = 0, iy=0; y < maxY; y+=sample_size.y(), iy++) {
                for (x = 0, ix=0; x < maxX; x+=sample_size.x(), ix++, m++) {
                    float gx = x + offset.x();
                    float gy = y + offset.y();
                    float gz = z + offset.z();
                    int idx_global_x = start_indices.x() + ix;
                    int idx_global_y = start_indices.y() + iy;
                    int idx_global_z = start_indices.z() + iz;
                    int global_index = idx_global_z*gsizes.y()*gsizes.x()
                                      +idx_global_y*gsizes.x()+idx_global_x;

                    float v  = get_tree_value_local(gx, gy, gz);
                    
                    m_distance_vol[global_index] = v;
                }
            }
        }
    }

#ifdef USE_MPI
    {
        D_P_MSG(m_myid,0
            ,"^^^^^^^^^^^^Reduce distance volume together.^^^^^\n");
    }
    int rank=-1;
    MPI_Comm_rank(m_communicator, &rank);
    int mpi_datatype = MPI_FLOAT; int op = MPI_MIN; int root = 0;
    int err;
    int ncount = gsizes.x() * gsizes.y();
    float* start_offet = m_distance_vol.data_ptr();

    for (int iz = 0 ; iz < gsizes.z() ; iz++)
    {
        if (rank == root)
        {
            err = MPI_Reduce(MPI_IN_PLACE, start_offet, ncount, mpi_datatype, op, root, m_communicator);
        }else
        {
            err = MPI_Reduce(start_offet, NULL, ncount, mpi_datatype, op, root, m_communicator);
        }

        if (err != MPI_SUCCESS)
        {
            char msg[128];
            sprintf(msg, "svector<T>::mpi_reduce failed!");
            D_P_MSG(rank, rank, msg);
        }
        start_offet += ncount;
    }
#endif
}

void distree_t::save_distance_volume_allreduce( const char* file_name )
{
    if (m_myid == 0)
    {
        FILE* fp = fopen(file_name,"wb");
        if (!fp){
            D_ASSERT_P(m_myid, fp, "distree_t::save_distance_volume_allreduce: File open failed!", DERROR_ERR_IO);
        }
        fwrite(m_distance_vol.data_ptr(), sizeof(float), m_distance_vol.size(),fp);
        fclose(fp);
        D_P_MSG(m_myid,m_myid
            ,"^^^^^^^^^Pass save_distance_volume_parallel^^^^^\n");
    }
}

void distree_t::verify_vertex_distance()
{
    float min_eucl = D_FLOAT_INFINITY, distance;
    int   min_voxel = 0;

    int vertex_array_size = m_vertex_array.size();
    int feature_array_size = m_features->size_feature();
    volume_t* pVolumeFeature = dynamic_cast<volume_t*>(m_features);
    svector<voxel_t> &voxel_array = pVolumeFeature->m_isovoxels;

    for (int vertex_id = 0; vertex_id < vertex_array_size; vertex_id++)
    {
        min_eucl = D_FLOAT_INFINITY;
        distree_vertex_t* vertex = &(m_vertex_array[vertex_id]);

        for (int voxel_id = 0; voxel_id < feature_array_size; voxel_id++)
        {
            voxel_t* voxel = &(voxel_array[voxel_id]);

            distance = (voxel->coords-vertex->coords).length();
            if (min_eucl > distance) {
                min_eucl = distance;
                min_voxel = voxel_id;
            }
        }

        if ( fabsf(vertex->eucl_dist - min_eucl ) > FLOAT_EPSILON) {
            fprintf(stderr, "PE%d: vertex %d (%f, %f, %f) computed %f, \
                            real_min_dist=%f from voxel%d (%f, %f, %f)\n",
                            m_myid,
                            vertex_id,
                            vertex->coords.x(),
                            vertex->coords.y(),
                            vertex->coords.z(),
                            vertex->eucl_dist,
                            min_eucl,
                            min_voxel,
                            voxel_array[min_voxel].coords.x(),
                            voxel_array[min_voxel].coords.y(),
                            voxel_array[min_voxel].coords.z());
        }
    }
}

void distree_t::verify_vertex_triangle_distance()
{
    char msg[128];
    int nVerticesTotal = m_vertex_array.size();
    int nFeatures = m_features->size_feature();
    sprintf(msg, "\n\nVerify PE%d distance computation....\nIt has %d local vertices, %d features\n\n\n",
        m_myid, nVerticesTotal, nFeatures);
    D_P_MSG(m_myid, m_myid, msg);

    feature_geometry_t* pGeometry = dynamic_cast<feature_geometry_t*>(m_features);
    if (pGeometry)
    {
        int count_wrong = 0;
        obj_model_t& obj_model = pGeometry->get_obj_model();
        if (obj_model.vertex_tab == NULL)
        {
            sprintf(msg, "\n\nWARNING! verify_vertex_triangle_distance(): the loaded obj file is freed at the end of feature_geometry_t::initial_distree_feature_array(), verification of vertex_triangle distance will now skipped. \n\n\n",
                m_myid, nVerticesTotal, nFeatures);
            D_P_MSG(m_myid, m_myid, msg);
            return;
        }

        int nTriangle = obj_model.obj_model_get_triangle_count();
        fprintf(stderr, "***send_schedule.size=%d\n",m_send_schedule.size());
        svector<int> sent_indices_all;
        for (int iR = 0 ; iR < m_send_schedule.size() ; iR++)
        {
            int s = m_send_schedule[iR].send_idx_buf.size();
            fprintf(stderr, "%d. PE%d ==> PE%d %d vertices\n",
                     iR, m_myid, m_send_schedule[iR].recv_id, m_send_schedule[iR].send_idx_buf.size());
            sent_indices_all.push_back(m_send_schedule[iR].send_idx_buf);
        }

        fprintf(stderr, "****end******\n");
        sent_indices_all.remove_duplicates(int_compare);

        for (int iv = 0 ; iv < sent_indices_all.size() ; iv++)
        {
            float   min_eucl = FLOAT_MAX;
            int     min_triangle = -1;
            distree_vertex_t& vertex = m_vertex_array[sent_indices_all[iv]];
            for (int it = 0 ; it < nTriangle ; it ++ )
            {
                triangle_t cur_triangle;
                bool isnormal = obj_model.obj_model_get_triangle_at(it, cur_triangle);
                if (isnormal)
                {
                    float distance = cur_triangle.vertex_triangle_distance(vertex.coords);
                    if (fabsf(min_eucl) > fabsf(distance)) {
                        min_eucl = distance;
                        min_triangle = it;
                    }
                }
            }
            if ( fabsf(vertex.eucl_dist - min_eucl ) > FLOAT_EPSILON) {
                count_wrong++;
                fprintf(stderr, "PE%d: vertex %d (%f, %f, %f) computed %f, \
                                real_min_dist=%f from triangle%d\n",
                                m_myid,
                                sent_indices_all[iv],
                                vertex.coords.x(),
                                vertex.coords.y(),
                                vertex.coords.z(),
                                vertex.eucl_dist,
                                min_eucl,
                                min_triangle);
            }
        }
        fprintf(stderr, "PE%d: %d/%d vertices sent to remote\n",
            m_myid,
            sent_indices_all.size(), nVerticesTotal);
        fprintf(stderr, "PE%d: wrong vertices: %d/%d\n",
                        m_myid,
                        count_wrong, nVerticesTotal);
        fflush(stderr);
    }
}

void distree_t::decompose_on_adding( int node_idx, int peid_changed )
{
    size_t nNodes = m_node_array.size();
    size_t leafnum = 0;
    for (int i = 0 ; i < nNodes ; i++)
    {
        if (m_node_array[i].isLeaf)
        {
            leafnum ++;
        }
        
    }
    int parent_depth_id = m_node_array[node_idx].level;
    if ((float)leafnum/(float)m_groupsize < m_epsilon
        && parent_depth_id < max_depth_id )
    {
        
    }
}

void distree_t::removing_isoblock_from_global_tree( int i, int peid_changed )
{

}

void distree_t::decompose_fulltree_on_adding( distree_node_t& leaf_changed, const vec3f& newpos )
{

}

void distree_t::merge_fulltree_leafnode( distree_node_t& parent_node )
{
    for (int i = 0 ; i < 8 ;i++)
    {
        int child_idx = parent_node.m_children[i];
        if (child_idx != -1)
        {
            m_node_array[child_idx].isoblock_overlap_indices.clear();
            m_node_array[child_idx].isoblock_3x_indices.clear();
            m_node_array[child_idx].triple_feature_indices.clear();
            m_node_array[child_idx].processor_3x_indices.clear();
            m_node_array[child_idx].processor_indices.clear();
        }
    }
    
}
