#ifndef _OCTREE_H_
#define _OCTREE_H_
#pragma once
#include <stdint.h>
#include <mpi.h>
#include "BBox.h"
#include "constant.h"
#include "svector.h"
#include "dvolume.h"
#include "dutility.h"
#include "IFeatures.h"
#include "obj_model_t.h"
#include "feature_isoblock_t.h"
#include "distree_vertex_t.h"

typedef enum relation_t{
    ITSELF=0,PARENT=1,GRANDPARENT=2
}relation_t;

typedef enum feature_type_t {
    F_VOLUME=1, F_TRIANGLE=2
}feature_type_t;

class distree_node_t
{
public:
    uint32_t  key;//default: 0
    int       level;//invalid value: -1.
    int       parent;//invalid value: -1.
    int       isLeaf;//default true
    bool      bHas_dirty_proc_3x;//if node's triple has processors that contain features.
    lily::BBox	m_bound;
    int        vertex_indices[8];//invalid value: -1.
    int        m_children[8];//invalid value: -1.
//     int     first_assign;//invalid value: -1.
//     int     second_assign;//invalid value: -1.
    typedef enum location_t {
        LOCAL, REMOTE
    }location_t;
    location_t  location;//0: empty, 1: has isoblock overlapped,2:within assigned region(within isoblock of itself)
                  //invalid value: -1.
    svector<int> triple_feature_indices;//feature_t   triple_features;
//  svector<int> complementary_features[8];//index to the feature that appear in the 3x range of current node but not present in child node.
    svector<int> isoblock_overlap_indices;
    svector<int> isoblock_3x_indices;
    svector<int> processor_indices;//indices of processors whose isoblock intersect the node's bbox.
    svector<int> processor_3x_indices;//indices of processors whose isoblock intersect the node's bbox.

//----------------------------------
    distree_node_t(void);
   // Octant(Point* points, int nPts, const BBox& bound, int level);
    ~distree_node_t();
    lily::BBox get_triple_bbox(void);
};

class distree_processor_t{
public:
    distree_processor_t(void):m_has_features(false){};
    ~distree_processor_t(){};
    svector<int> m_isoblock_indices;
    bool         m_has_features;
};

class key_idx_pair_t{
public:
         key_idx_pair_t(void):key(-1),idx(-1){};
explicit key_idx_pair_t(uint32_t k, uint32_t i):key(k),idx(i){};

    //functor used for comparison
    bool operator()(const key_idx_pair_t &a,const key_idx_pair_t &b)
    {
        return a.key < b.key;
    }

    uint32_t key;
    uint32_t idx;
};

class volume_t;
class feature_isoblock_t;

class distree_t
{
    friend class volume_t;
    friend class feature_isoblock_t;
    friend class feature_pointcloud_t;
    friend class feature_geometry_t;
public:
    size_t               m_myid;//processor id
    size_t               m_groupsize;//# processors.
    MPI_Comm             m_communicator;//MPI communicator
    bool                 m_has_features;//if the distree local region has
    bool                 m_is_building_global_tree;//indicate if current phase is building global tree.
    float                m_epsilon;
    int                  root_index;
    int                  max_depth_id;
    float                min_extend;
    int                  m_leaf_count;//# of leaf node.
    lily::vec3f          sample_size;
    svector<distree_processor_t> m_processor_array;
    svector<distree_node_t>   m_node_array;
    svector<distree_vertex_t> m_vertex_array;
    IFeatures*           m_isoblock_array;//isoblocks
    IFeatures*           m_global_leaf_block_array;//isoblocks
    svector<int>         m_assigned_local_leaf_indices;//array of indices of assigned local leaf nodes.
    //lily::BBox       m_local_region_box;
    feature_type_t       m_feature_type;
    IFeatures*           m_features;//isovoxel, triangle
    IFeatures*           m_cur_active_features;//points to either isovoxel m_features, m_isoblock_array or triangle geometry.
    //int                  feature_array_size;
    lily::vec3f                m_center;
    lily::vec3i                m_np_xyz;//All processors layout dimension.
    lily::vec3i                m_global_domain_size;//global distance field global size;//vec3i or vec3f?
    lily::vec3f                m_global_feature_size;//global feature size
    lily::vec3f                m_local_feature_block_size;
    lily::BBox       m_global_domain_bound;//global distance field domain bounding box is a CUBE.
    lily::BBox       m_local_feature_bound;
    lily::BBox       m_global_feature_bound;
    svector<int>         m_sent_table;//each entry save the processor id to which current processor will send msg.
    svector<int>         m_recv_table;//each entry save the processor id from which current processor will receive msg.
    svector<volume_t*>   m_distance_volumes;
    svector<float>       m_distance_vol;
    svector<send_schedule_t> m_send_schedule;
    void set_feature_array(IFeatures *feature){ m_features = feature;}
    void set_isoblock_feature(IFeatures *isoblock_feature){ m_isoblock_array = isoblock_feature;}
    void set_active_feature(IFeatures *feature){ m_cur_active_features = feature;};
protected:

    typedef bool  (distree_t::*collect_feature_fun_t)(size_t cur_node_index,  \
                                                      size_t parent_node_index,\
                                                      int    parent_depth_id);
    //create children for the node indexed by param parent_node_index.
    //node_array: global node array.
    //return: true if children created, or else false.
    bool create_children( size_t parent_node_index, IFeatures *features
                        , bool is_profiling=false);

    void assign_leaf_node();

    int  compute_global_child_key(const distree_node_t &parent_node,
                                  int lockey/*[0,7]*/) const;

    int  compute_local_child_key(const distree_node_t &child_node) const;
    int  compute_octree_vertex_key(distree_vertex_t &vtx) const;
    int  compute_parent_key(int child_key,int child_depth_id) const;
    int  compute_parent_key(const distree_node_t &child_node) const;
    void compute_from_key_to_coords(int key, lily::vec3f &coords) const;
    //********Collect Isoblock*********
    // Parameter: int cur_node_index: the node that participates collecting
    // features and further decide if it is a leaf node.
    // Parameter: int parent_node_index: parent node of the "cur_node".
    // Parameter: const voxel_t * isovoxels: global array of isovoxels.
    // Parameter: parent_depth_id: parent_node's depth id.
    // return: if the cur_node_index is leaf node.
    //**********************************************
//     bool collect_block_in_range(size_t cur_node_index,
//                                 size_t parent_node_index,
//                                 int    parent_depth_id);

    void arrays_intersection(svector<key_idx_pair_t> &empty_leaf_parent_key_idx_pairs,
                             svector<uint32_t> &local_leaf_parent_keys,
                             svector<int>      &intersection_idx );

    //Gather the global tree leaf node assignment of all processors.
    void collect_all_leaf_assignment();

    void create_send_schedule_vertex(svector<send_schedule_t>& send_schedule );

    int get_max_local_region_num() const;
public:
    distree_t(int peid, feature_type_t feature_type, MPI_Comm comm, int max_depth,
           float resolution=1.0f, float epsilon=1.0f);

    ~distree_t(void);
    size_t  get_processor_id() const {return m_myid;}
    void    set_processor_id(int id) { m_myid = id;}
    lily::vec3i   get_processor_layout() const { return m_np_xyz;}
    void    set_processor_layout(lily::vec3i np_xyz) { m_np_xyz = np_xyz;}

    void    set_sample_size(const lily::vec3f& size3d)
                            { sample_size = size3d;}

    void    set_leaf_node_size(const lily::vec3f &size3d)
    {
        min_extend  = size3d.max_extend1f();
    }
    //***//
    bitmap_t * generate_bitmap(int groupsize, color_t mycolor, MPI_Comm mpicomm);
    //***//
    void create_global_distree(const  lily::vec3i &global_domain_size,
                               const lily::vec3f &global_feature_size,const lily::vec3i &np_xyz,
                               int groupsize);
    //***//
    void create_comm_schedule();
    bool node_has_dirty_proc_3x(const distree_node_t& node);

    void print_dirty_proc_3x(const distree_node_t& node);
    void print_dirty_proc_1x(const distree_node_t& node);

    void verify_schedule( int groupsize );

    typedef void (*initial_distree_feature_array_fun_t)(int idxBase/*=0,1*/,
                                                        void* feature_obj);
    //***//
    void refine_global_distree(bool binit_features=true);
    //***//
    void create_local_distree_vertex_array();
    //
    void save_distree_binary(const char *treeName);
    void save_distree_text(const char* treeName);
    void compute_vertex_distance_parallel();
    void compute_local_distance();
    void compute_child_vertex_distance( distree_node_t & child_node, distree_node_t & parent_node );
    //compute the distance from vtx_coords to all the "comp_feature_id" features of the "node"
    //The "node" must has features.
    //vtx_coords :coordinates of the vertex.
    //node: the node whose 3x BBox gives the feature search range.
    //eucl: Returned min Euclidean distance to all "node"'s features
    //maxnorm: Returned min maxnorm distance to all "node"'s features
    void compute_vertex_node_feature_distance( distree_vertex_t& cur_vertex,
                                               distree_node_t& feature_node, float& eucl_dist,
                                               float& norm_dist, int compute_type=-1 );

    void compute_vertex_node_feature_distance( lily::vec3f& coords,
                                               distree_node_t& feature_node, float& eucl_dist,
                                               float& norm_dist, int compute_type=-1 );

    distree_node_t* get_distree_node(const lily::vec3f& coord);
    distree_node_t* get_distree_local_node(const lily::vec3f& coord, bool debug=false );
    distree_node_t* get_node_child(const distree_node_t &parent_node, int childid);
    float           get_tree_value_local(float x, float y, float z, bool debug=false);
    float           get_tree_value_global(float x, float y, float z, bool debug=false);
    void create_distance_volume();
    void create_distance_volume_allreduce();
    void save_distance_volume_parallel(const char* file_name );
    void save_distance_volume_allreduce(const char* file_name );
    void verify_vertex_distance();
    void verify_vertex_triangle_distance();
    void recompute_global_tree_using_leaf_assignment();
    void decompose_on_adding( int node_idx, int peid_changed );
    void removing_isoblock_from_global_tree( int i, int peid_changed );
    void decompose_fulltree_on_adding( distree_node_t& leaf_changed, const lily::vec3f& newpos );
    void merge_fulltree_leafnode( distree_node_t& parent_node );
    
};

//int int_compare(const void *pa,const void *pb);

#endif

