#include <iostream>
#include <algorithm>
#include <stdint.h>
#include "distree_t.h"
#include "DError.h"
#include "mathtool.h"
#include "utility.h"
#include "hpgv_utiltiming.h"
#include "parallel_test.h"
#include "distree_vertex_t.h"
#include "feature_geometry_t.h"

using namespace std;
distree_node_t::distree_node_t( void )
    :isLeaf(1),level(-1),parent(-1),key(0)
    ,location(REMOTE)
    //,first_assign(-1),second_assign(-1)
{
    for (int i = 0 ; i < 8 ; i ++){
        m_children[i] = -1;
        vertex_indices[i] = -1;
    }
}

distree_node_t::~distree_node_t()
{
}
distree_t::distree_t(int peid, feature_type_t feature_type, MPI_Comm comm,
    int max_depth, float resolution/*=1.0f*/, float epsilon/*=1.0*/)
    :m_myid(peid),m_feature_type(feature_type)
    ,m_features(NULL),m_cur_active_features(NULL)
    ,root_index(0),max_depth_id(max_depth)
    ,min_extend(resolution),m_epsilon(epsilon),m_communicator(comm)
    ,m_has_features(false),m_leaf_count(0)
{
    set_leaf_node_size(vec3f(1.0f,1.0f,1.0f));
}

distree_t::~distree_t(void)
{
    //Clear up m_distance_volumes
    for (int i = 0; i < m_distance_volumes.size() ; i++)
    {
        delete m_distance_volumes[i];
        m_distance_volumes[i] = NULL;
    }
    m_distance_volumes.clear();
}

bitmap_t * distree_t::generate_bitmap(int groupsize, color_t mycolor, MPI_Comm mpicomm)
{
    uint8_t  *lmap   = NULL;
    uint8_t  *gmap   = NULL;
    bitmap_t *bitmap = NULL;
    int     mapsize = (int)(ceil(groupsize / 8.0f));
    uint8_t mycode = 0;
    int nbits = 0;
    size_t myelement, mybit, i;

    lmap = (uint8_t *)calloc(mapsize, sizeof(uint8_t));
    D_ASSERT_P(m_myid, lmap != 0, "Out of memory", DERROR_ERR_MEM);

    gmap = (uint8_t *)calloc(mapsize, sizeof(uint8_t));
    D_ASSERT_P(m_myid, gmap != 0, "Out of memory", DERROR_ERR_MEM);
    //Generate local bitmap.
    if (mycolor == BLACK) {
        myelement = m_myid / 8;
        mybit  = m_myid % 8;
        mycode = 1 << (/*7 -*/ mybit);
        lmap[myelement] = mycode;
    }

    //Combine all PEs' bitmap.
    MPI_Allreduce(lmap, gmap, mapsize, MPI_BYTE, MPI_BOR, mpicomm);

    free(lmap);

#ifdef DEBUG_0
    fprintf(stderr, "id %d ", m_myid);
    for (i = 0; i < mapsize; i++) {
        fprintf(stderr, " %s ",  byte_to_binary(gmap[i]));
    }
    fprintf(stderr, "\n");
    fflush(stderr);
#endif

    bitmap = (bitmap_t *)calloc(1, sizeof(bitmap_t));
    D_ASSERT_P(m_myid, bitmap != 0, "Out of memory", DERROR_ERR_MEM);

    //Count the # of processor which contains isovoxels.
    for (i = 0; i < mapsize ; i++){
        //Scan the gmap in a chunk of 8bits as step size.
        //and count the number of bit 1 in a chunk by looking up the table.
        nbits += bit_count_table[gmap[i]];//count the number of bit 1.
    }
    bitmap->data = (uint8_t*)gmap;
    bitmap->size = mapsize;
    bitmap->nbits= nbits;

    return bitmap;
}
int distree_t::compute_parent_key(int child_key,int child_depth_id) const
{
    int nShift = ((max_depth_id-child_depth_id+1)*3);
    //make the right most nShift bits 0.
    return (child_key >> nShift ) << nShift;//(child_key >> 3)<<3;
}
int distree_t::compute_parent_key(const distree_node_t &child_node) const
{
    int nShift = ((max_depth_id-child_node.level+1)*3);
    //make the right most nShift bits 0.
    return (child_node.key >> nShift ) << nShift;//(child_key >> 3)<<3;
}
//compute_global_child_key from parent key by appending lockey to
//parent key at an appropriate position indicated by param i_depth.
//param: parent_depth_id: the depth id of parent node
int distree_t::compute_global_child_key(const distree_node_t &parent_node,
                                    int lockey/*[0,7]*/) const
{
    int parent_key = parent_node.key;
    int parent_depth_id = parent_node.level;
    int child_key
        = parent_key | (lockey << ((max_depth_id - parent_depth_id - 1 )*3));
    return  child_key;
}

int distree_t::compute_local_child_key(const distree_node_t &child_node) const
{
    int global_key = child_node.key;
    int child_depth_id = child_node.level;
    //extract the right most 3 bits as local key, to determine the children id.
    return ((global_key >> ((max_depth_id - child_depth_id)*3)) & 0x00000007);
}

//the max_depth_id of octant is one less than the max_depth_id of octant vertex
int distree_t::compute_octree_vertex_key(distree_vertex_t &vtx) const
{
    int i_depth = 0;
    int key=0;
    int locKey=0;
    float threshold  = min_extend*0.5f;//vertex key resolution is 2 time the one of octant resolution
    //therefore, the radius threshold for key is half of the one of octant.
    float px = vtx.coords.x(), py = vtx.coords.y(), pz = vtx.coords.z();
    const vec3f center_offset[8] = {//{dx,dy,dz}
        vec3f(-1,-1,-1), vec3f(1,-1, -1), vec3f(-1, 1,-1), vec3f(1, 1, -1),
        vec3f(-1,-1, 1), vec3f(1,-1,  1), vec3f(-1, 1, 1), vec3f(1, 1,  1)
    };
    //initial bbox at depth=0
    float radius = m_global_domain_size.max_extend1i()*0.5f;//radius of octant at each iteration.
    vec3f c = m_center;//center coordinates of bbox .c[0]:x,c[1]:y,c[2]:z

    for (i_depth = 1; i_depth <= max_depth_id+1 && radius >= threshold; i_depth++)
    {
        locKey = ( ( pz > c[2]) << 2 ) +	/*z*/
                 ( ( py > c[1]) << 1 ) +	/*y*/
                 ( ( px > c[0]) << 0 ) ;	/*x*/
        key |= locKey << ((max_depth_id+1 - i_depth)*3); //global key.
        //update bbox for next depth.
        radius *= 0.5f;//shrink size by half.
        c += center_offset[locKey] * radius;
    }
    return key;
}

void distree_t::compute_from_key_to_coords(int key, vec3f &coords) const
{
    const vec3f center_offset[8] = {//{dx,dy,dz}
        vec3f(-1,-1,-1), vec3f(1,-1, -1), vec3f(-1, 1,-1), vec3f(1, 1, -1),
        vec3f(-1,-1, 1), vec3f(1,-1,  1), vec3f(-1, 1, 1), vec3f(1, 1,  1)
    };
    //initial bbox at depth=0
    float radius = m_global_domain_size.max_extend1i()*0.5f;//dsize*0.5f;
    vec3f c(m_center);
    int   locKey=0;
    int i_depth = 0;
    for (i_depth = 1; i_depth <= max_depth_id/*+1*/ && radius >= min_extend; i_depth++) {
         locKey = key >> ((max_depth_id+1 - i_depth)*3); //global key.
         locKey &= 0x07;
        //update bbox for next depth.
        radius *= 0.5;//shrink size by half.
        c[0] = c[0] + center_offset[locKey][0]*radius;
        c[1] = c[1] + center_offset[locKey][1]*radius;
        c[2] = c[2] + center_offset[locKey][2]*radius;
    }
    locKey = key >> ((max_depth_id+1 - i_depth)*3); //global key.
    locKey &= 0x07;
    c += center_offset[locKey]*radius;
    coords = c;
}

void distree_t::create_global_distree(
    bitmap_t *bitmap, const vec3i &global_domain_size,
    const vec3f &global_feature_size,
    const vec3i &np_xyz, int groupsize)
{
    m_np_xyz = np_xyz;
    m_groupsize = groupsize;
    //Initialize geometric info of global domain and global feature.
    //global domain m_bound is a CUBE. And the origin of the coordinates
    //is and the lower-left corner of the global_feature bbox.
    m_global_feature_size = global_feature_size;
    m_center = m_global_feature_size * 0.5f;
    float global_extend        = (float)global_domain_size.max_extend1i();
    m_global_domain_size       = vec3i((int)global_extend);
    m_global_domain_bound.pMin = m_center - global_extend*0.5f;
    m_global_domain_bound.pMax = m_center + global_extend*0.5f;
    vec3f global_dimension = m_global_domain_bound.getDimension();

    vec3i pxyz   = vec3i::index1D_to_index3D((int)m_myid, np_xyz);
    m_local_feature_block_size = BLOCK_SIZE_FLOAT3D(pxyz, np_xyz, global_feature_size);//not used
    m_local_feature_bound.pMin =  BLOCK_LOW_FLOAT3D(pxyz, np_xyz, global_feature_size);//not used
    m_local_feature_bound.pMax = BLOCK_HIGH_FLOAT3D(pxyz, np_xyz, global_feature_size);//not used

    //global feature bound not necessarily a cube.
    m_global_feature_bound.pMin= m_center - global_feature_size*0.5f;
    m_global_feature_bound.pMax= m_center + global_feature_size*0.5f;

    //---------initialize isoblock array--------------
    HPGV_TIMING_BEGIN(parallel_test::T_g_init);
    HPGV_TIMING_COUNT(parallel_test::T_g_init);
    m_isoblock_array = new feature_isoblock_t(bitmap);
    m_isoblock_array->initial_distree_feature_array(*this);
    HPGV_TIMING_END(parallel_test::T_g_init);
    //---------initialize root node--------------
    distree_node_t node;
    m_node_array.push_back(node,true);
    m_node_array[0].level  = 0;
    m_node_array[0].parent = -1;
    m_node_array[0].isLeaf = 1;//only one root node at this time.
    m_node_array[0].isoblock_3x_indices.resize(m_isoblock_array->size_feature(), true);
    //Initialize root BBox.
    //root node bound is trivially the whole domain bound.
    m_node_array[0].m_bound = m_global_domain_bound;
    for (int i=0; i< m_node_array[0].isoblock_3x_indices.size() ; i++){
        //root node contains all isoblocks. it is trivial to initialize
        //its isoblock_indices array.
        m_node_array[0].isoblock_3x_indices[i] = i;
    }
    //root node bound trivially covers all processors.
    m_node_array[0].processor_indices.resize(groupsize, true);
    m_node_array[0].processor_3x_indices.resize(groupsize, true);
    for (int i=0; i< m_node_array[0].processor_indices.size() ; i++){
        //root node processor list contains all processor ids.
        //it is trivial to initialize its processor list.
        m_node_array[0].processor_indices[i] = i;
        m_node_array[0].processor_3x_indices[i] = i;
    }

    //memset(m_node_array[0].complementary_features,NULL, sizeof(feature_t)*8);
    //root node should also overlap all isoblocks, it is trivial to initialize
    //isoblock_overlap_indices the same as isoblock_3x_indices.
    m_node_array[0].isoblock_overlap_indices = m_node_array[0].isoblock_3x_indices;

    int leafnum=1;//The leaf node number
    size_t start=0, end=m_node_array.size();
    int parent_depth_id = 0;
    HPGV_TIMING_BEGIN(parallel_test::T_g_2);
    HPGV_TIMING_COUNT(parallel_test::T_g_2);
    while (true){
        //scan parents. one level at a time.
        for (size_t parent_idx = start ; parent_idx < end ; parent_idx++ ){
            //First determine if current parent needs to be decomposed
            //by checking if it contains features.
            distree_node_t &parent_node = m_node_array[parent_idx];
            parent_depth_id = parent_node.level;
            if (parent_node.isoblock_3x_indices.size() > 0){
                //parent node 3x range has isoblock
                //Split current parent by generating 8 children.
                HPGV_TIMING_BEGIN(parallel_test::T_g_child);
                bool bSplitted = create_children(parent_idx, m_isoblock_array, true);
                HPGV_TIMING_END(parallel_test::T_g_child);
                //parent_node address may be change in create_children due to
                //insertion child nodes to node_array causing reallocation of buffer.
                distree_node_t &parent_node = m_node_array[parent_idx];
                if (bSplitted){
                    parent_node.isLeaf = 0;
                    leafnum += 7;
                }
            }
        }
        start = end;
        end = m_node_array.size();

        if ((float)leafnum/(float)groupsize >= m_epsilon
            || parent_depth_id > max_depth_id
            || end==start){
                break;
        }
    }
    HPGV_TIMING_END(parallel_test::T_g_2);
    distree_node_t& lastNode = m_node_array.back();
    HPGV_TIMING_COUNT(parallel_test::C_global_tree_h);
    HPGV_TIMING_INCREASE(parallel_test::C_global_tree_h,lastNode.level);
    HPGV_TIMING_INCREASE(parallel_test::C_global_tree_size,m_node_array.size());
    char msg[128];
    sprintf(msg, "****global tree height = %d****\n", lastNode.level);
    D_P_MSG(m_myid, 0, msg);
    HPGV_TIMING_BEGIN(parallel_test::ASSIGN_LOCAL_REGION);
    HPGV_TIMING_COUNT(parallel_test::ASSIGN_LOCAL_REGION);

    assign_leaf_node(bitmap, np_xyz, global_domain_size, groupsize);

    HPGV_TIMING_END(parallel_test::ASSIGN_LOCAL_REGION);
}

bool distree_t::create_children(size_t parent_node_index, IFeatures *features, bool is_profiling/*=false*/)
{
    distree_node_t &parent_node = m_node_array[parent_node_index];
    int parent_depth_id = parent_node.level;
    //extend of child bbox
    float child_extend = parent_node.m_bound.MaximumExtentf() * 0.5f;

    int num_none_leaf = 0;
    const vec3f pt_min_offset[8]= {
                vec3f(0, 0, 0), vec3f(1, 0, 0), vec3f(0, 1, 0), vec3f(1, 1, 0),
                vec3f(0, 0, 1), vec3f(1, 0, 1), vec3f(0, 1, 1), vec3f(1, 1, 1)
                };
    //fprintf(stderr, "child_extend=%f\n",child_extend);
    if (child_extend < min_extend){
        return false;//no split.
    }
    if (!is_profiling){
        //Create 8 children.
        for (int i = 0; i < 8 ; i++){
            //for each new child
            //insert new child node
            m_node_array.push_back(distree_node_t());
            size_t cur_child_index = m_node_array.size()-1;
            distree_node_t&
                current_child =	m_node_array[cur_child_index];
            current_child.level = parent_depth_id + 1;
            //Compute its BBox and key.
            distree_node_t &parent_node = m_node_array[parent_node_index];
            current_child.m_bound.pMin = parent_node.m_bound.pMin + pt_min_offset[i]*child_extend;
            current_child.m_bound.pMax = current_child.m_bound.pMin + child_extend;
            current_child.location = parent_node.location;

            //don't forget who is your parent.
            current_child.parent = (int)parent_node_index;
            //compute child key based on parent key.
            current_child.key = compute_global_child_key(parent_node, i);
            parent_node.m_children[i] = (int)cur_child_index;
            //collect feature that is within 3x BBox(or sphere) of current child.
            //1. For serial version, we only consider isovoxels
            //2. For parallel version of local distree, we consider both isovoxels and remote isoblocks.
            //3. For parallel version of block distree, we only consider remote isoblocks.
            //return: FALSE: current child node is not leaf node.
            current_child.isLeaf = features->collect_feature(cur_child_index,
                                                            parent_node_index,
                                                            *this,is_profiling);
        }
    }else{
        //Create 8 children.
        for (int i = 0; i < 8 ; i++){
            //for each new child
            //insert new child node
            m_node_array.push_back(distree_node_t(),is_profiling);
            size_t cur_child_index = m_node_array.size()-1;
            distree_node_t&
                current_child =	m_node_array[cur_child_index];
            current_child.level = parent_depth_id + 1;
            //Compute its BBox and key.
            distree_node_t &parent_node = m_node_array[parent_node_index];
            current_child.m_bound.pMin = parent_node.m_bound.pMin + pt_min_offset[i]*child_extend;
            current_child.m_bound.pMax = current_child.m_bound.pMin + child_extend;
            current_child.location = parent_node.location;

            //don't forget who is your parent.
            current_child.parent = (int)parent_node_index;
            HPGV_TIMING_BEGIN(parallel_test::T_g_key);
            //compute child key based on parent key.
            current_child.key = compute_global_child_key(parent_node, i);
            HPGV_TIMING_END(parallel_test::T_g_key);
            parent_node.m_children[i] = (int)cur_child_index;
            //collect feature that is within 3x BBox(or sphere) of current child.
            //1. For serial version, we only consider isovoxels
            //2. For parallel version of local distree, we consider both isovoxels and remote isoblocks.
            //3. For parallel version of block distree, we only consider remote isoblocks.
            //return: FALSE: current child node is not leaf node.
            HPGV_TIMING_BEGIN(parallel_test::T_g_collect);
            current_child.isLeaf = features->collect_feature(cur_child_index,
                                                            parent_node_index,
                                                            *this,is_profiling);
            HPGV_TIMING_END(parallel_test::T_g_collect);
        }
    }

    return true;
    //If all 8 children are none leaf node. We can safely remove the parent's
    //voxel_indices array.
    //    if (num_none_leaf == 8) {
    //        free(parent_node->voxel_indices);
    //        parent_node->voxel_indices  = 0;
    //        parent_node->voxelnum       = 0;
    //    }
}

void distree_t::assign_leaf_node( const bitmap_t* bitmap,const vec3i &np_xyz,const vec3i &global_size, int groupsize )
{
    //Select local volume region. From now on the local volume region
    //of each PE is not necessarily the region of its regular sampled region,
    //it maybe the union of multiple regions covered by several consecutive leaf nodes
    //of global distree.
    //In the first pass, For each leaf node whose processor_array
    //is nonempty(has isoblocks). We assign such an leaf node to
    //the processor in its "processors" list to minimize data movement. Some
    //leaf nodes maybe assigned to only one processor, while some leaf nodes
    //may be assigned to multiple processors.
    bool bHasWorkLoad = false;
    //std::vector<int> leaf_indices;
    m_leaf_count = 0;
    for (int i=0 ; i < m_node_array.size() ; i++){
        distree_node_t &cur_node = m_node_array[i];
        if (cur_node.isLeaf)
        {
            m_leaf_count++;
            //leaf_indices.push_back(i);
            if (cur_node.processor_indices.size()>0){
                //check if processor_indices contains myid.
                //if yes, set cur_node dirty;
                size_t proc_count = cur_node.processor_indices.size();
                for (size_t proc=0; proc < proc_count ; proc++){
                    size_t pid = cur_node.processor_indices[proc];
                    distree_processor_t &cur_proc = m_processor_array[pid];
                    if (pid == m_myid && cur_proc.m_has_features){
                        bHasWorkLoad = true;
                        m_assigned_local_leaf_indices.push_back(i);
                    }
                }
            }
        }
    }
    int N = m_assigned_local_leaf_indices.size();
    if (m_myid==0 ){
        printf("====PE%d has %d assigned local leaf indices, # total node=%d====\n",
            m_myid, m_assigned_local_leaf_indices.size(), m_node_array.size());
        for (int i = 0 ; i < N ; i++){
            printf("====PE%d: assigned_leaf_node[%d/%d]=%d=====\n",m_myid,i,N, m_assigned_local_leaf_indices[i]);
        }
    }
    if ( m_myid==122){
        printf("====PE%d has %d assigned local leaf indices====\n",
                    m_myid, m_assigned_local_leaf_indices.size());
        for (int i = 0 ; i < N; i++){
            printf("----PE%d: assigned_leaf_node[%d/%d]=%d----\n",m_myid,i,N, m_assigned_local_leaf_indices[i]);
        }
    }
    // MPI_Finalize();
    // exit(0);
    //In the second pass, for each leaf octant whose processors array is empty,
    //it does not have any elements within its region, and we call it an empty
    //leaf octant.(such leaf octant's indices are stored in empty_leaf_indice),
    //In the first case, all processors have been already assigned with workload in the first pass.
    //All gather workload flag to see if all processors have workload in the first pass.
    svector<uint8_t> hasWorkloadFlags;
    hasWorkloadFlags.resize(groupsize);

    MPI_Allgather(&(bHasWorkLoad),      1, MPI_UNSIGNED_CHAR,
                  &hasWorkloadFlags[0], 1, MPI_UNSIGNED_CHAR,
                  m_communicator);

    bool allHasWorkLoad = true;
    for (int i = 0; allHasWorkLoad && i < hasWorkloadFlags.size() ; i++){
        //if there is one(or more) processor has no workload, allHasWorkLoad==false.
        allHasWorkLoad = allHasWorkLoad & (bool)hasWorkloadFlags[i];
    }
    svector<int> all_assigned_dirty_leaf;
    svector<int> count_array, displ_array;
    all_assigned_dirty_leaf.mpi_allgatherv(m_assigned_local_leaf_indices,
                                          count_array, displ_array, MPI_INT);

    if (m_myid==0 ){
        //for debugging.
        FILE* fp = fopen(".\output_debug.txt","w");
        if (!fp){
            D_ASSERT_P(m_myid, fp, "File open failed!", DERROR_ERR_IO);
        }

        fprintf(fp, "+++++++++++++++++++++++++++++++++++++++\n");
        fprintf(fp, "++++++ PE%d: allHasWorkLoad=%d++# total node=%d====\n",
            m_myid,allHasWorkLoad, m_node_array.size());
        fprintf(fp, "+++++++++++++++++++++++++++++++++++++++\n");
        //Gather m_assigned_local_leaf_indices array from all processors and those
        //local leaf nodes have isoblocks corresponding to that processor.
        //check if all_assigned_dirty_leaf are in ascending order.
        for (int iPE = 0 ; iPE < groupsize ; iPE ++){
            fprintf(fp, "++++++++++Start of PE%d++++++++++++++++++++\n",iPE);
            //sort subarray of each processor.
            void* start = &(all_assigned_dirty_leaf[displ_array[iPE]]);
            //void* end   = &(all_assigned_dirty_leaf[displ_array[iPE]+count_array[iPE]]);
            qsort(start, count_array[iPE], sizeof(int), int_compare);

            for (int j = 0; j < count_array[iPE] ; j++){
                int k = all_assigned_dirty_leaf[displ_array[iPE]+j];
                fprintf(fp, "====assign_dirty_leaf[%d]=%d\n",j, k);
            }
            fprintf(fp, "-----------End of PE%d----------------------\n\n",iPE);
        }

        fclose(fp);
    }
    //Compute the empty leaf node array from all_assigned_dirty_leaf.
    all_assigned_dirty_leaf.remove_duplicates(int_compare);
//     qsort(all_assigned_dirty_leaf.data_ptr(), all_assigned_dirty_leaf.size(),
//         sizeof(int), int_compare);
    //all_assigned_dirty_leaf is now sorted and has no duplicates.
    svector<int> empty_leaf_indices;
    int total_node_count = m_node_array.size();
    int i = total_node_count - m_leaf_count;//i: index of first leaf node in m_node_array.
    int j = 0;//j: index of assigned leaf node in assign_dirty_leaf_count.
    int assign_dirty_leaf_count = all_assigned_dirty_leaf.size();
    //Collect the leaf nodes that are not assigned yet by subtracting
    //the elements in assign_dirty_leaf from entire leaf nodes list.
    while ( i < total_node_count && j < assign_dirty_leaf_count){
        D_ASSERT_P(m_myid, m_node_array[i].isLeaf, "not leaf node!", DERROR_ERR_IO);
        if (i < all_assigned_dirty_leaf[j]) {
            empty_leaf_indices.push_back(i);
            i++;
        }
        else if ( i > all_assigned_dirty_leaf[j] ){
            j++;
        }else if (i == all_assigned_dirty_leaf[j]) {
            i++; j++;
        }
    }
    while (i < total_node_count){
        empty_leaf_indices.push_back(i++);
    }
    //printf("--empty_leaf_indices.size()=%d+assigned_leaf_node.size()=%d-=%d---total_leaf_node=%d--\n",empty_leaf_indices.size(),\
        all_assigned_dirty_leaf.size(), empty_leaf_indices.size()+all_assigned_dirty_leaf.size(),leaf_count);
    if (allHasWorkLoad){
        //In the first case, if all processors has workloads, Then, we assign each empty leaf
        //octant oct to the processor that is responsible for the octs
        //neighboring leaf octants in leaf array and these neighboring leaf
        //octants have the closest ancestor with oct.
        svector<key_idx_pair_t> empty_leaf_parent_key_idx_pairs;
        for (int i = 0 ; i < empty_leaf_indices.size() ; i++){
            //int key = node_array[empty_leaf_indices[i]].key;
            int idx = empty_leaf_indices[i];
            int parent_key = compute_parent_key(m_node_array[idx]);
            empty_leaf_parent_key_idx_pairs.push_back(key_idx_pair_t(parent_key, idx));
        }
        svector<uint32_t> local_leaf_parent_keys;
        for (int i=0 ; i < m_assigned_local_leaf_indices.size() ; i++){
            int parent_key = compute_parent_key(m_node_array[m_assigned_local_leaf_indices[i]]);
            local_leaf_parent_keys.push_back(parent_key);
        }
        //compute the intersection of the above two indices arrays.
        //And add the the idx component of the intersection part
        //to m_assigned_local_leaf_indices.
        arrays_intersection(empty_leaf_parent_key_idx_pairs,
                            local_leaf_parent_keys,
                            m_assigned_local_leaf_indices);
    }else{
    //In the second case, some processors have not been assigned workload in the first pass.
    //In this case we adopt a simple rule that we evenly partition and distribute the
    //empty leaf octants according to their order in leaf array among the rest of processors.
        //Check if current processor has no work load.
        //we only consider the processors with no job in the second case.
        if (m_assigned_local_leaf_indices.size() <= 0){
            //1. scan hasWorkloadFlags array to count how many processor has no work load.
            int nNoWorkProcs = 0;
            for (int i = 0; i < hasWorkloadFlags.size() ; i++){
                nNoWorkProcs = hasWorkloadFlags[i] ? nNoWorkProcs : nNoWorkProcs+1;
            }
            //evenly partition and distribute the empty leaf nodes to the
            //processor with no job.
            size_t num_empty_leaf_nodes = empty_leaf_indices.size();
            int noWorkProcId=0;
            for (int i = 0; i < m_myid && i < hasWorkloadFlags.size() ; i++){
                if (! hasWorkloadFlags[i]){
                    noWorkProcId++;
                }
                //noWorkProcId = hasWorkloadFlags[i] ? noWorkProcId : noWorkProcId+1;
            }
            int istart = BLOCK_LOW( noWorkProcId, nNoWorkProcs, num_empty_leaf_nodes);
            int iend   = BLOCK_HIGH(noWorkProcId, nNoWorkProcs, num_empty_leaf_nodes);
            //add newly partitioned leaf node to local_leaf_indices.
            for (int i = istart; i <= iend ; i++){
                m_assigned_local_leaf_indices.push_back(empty_leaf_indices[i]);
            }
        }
    }
    //Now we have partitioned local leaf nodes of global distree, so that every processor
    //has job to do, and every empty leaf node is assigned to at least one processor.
    //The partition naturally follows Z-order space filling curve which preserves locality.
    //Then each processor goes through each octant oct, set oct:location=local if oct is assigned
    //to itself, and otherwise set oct:location = remote.
    D_ASSERT_P(m_myid, m_assigned_local_leaf_indices.size()>0,
               "current processor still has NO leaf node assigned to!\n",
               DERROR_ERROR);

    for (int i = 0; i < m_assigned_local_leaf_indices.size() ; i++){
        distree_node_t &cur_node = m_node_array[m_assigned_local_leaf_indices[i]];
        D_ASSERT_P(m_myid, cur_node.isLeaf,
                   "None leaf node is not allowed to set location flag!\n",
                   DERROR_ERROR);
        //default location is REMOTE.
        cur_node.location = distree_node_t::LOCAL;
        //compute the bbox of assigned leaf node region
        m_local_region_box = m_local_region_box.Union(cur_node.m_bound);
        //tag its parent.location==LOCAL if its parent is not LOCAL.
        distree_node_t* ancestor_node = &cur_node;
        while(   ancestor_node->parent >= 0
            && m_node_array[ancestor_node->parent].location
            != distree_node_t::LOCAL)
        {
            ancestor_node = &(m_node_array[ancestor_node->parent]);
            ancestor_node->location = distree_node_t::LOCAL;
        }
    }
    m_assigned_local_leaf_indices.shrink_to_fit();
    //svector<int>(m_assigned_local_leaf_indices).swap(m_assigned_local_leaf_indices);
    //if (m_myid==0){
//         FILE* fp = fopen(".\output_debug.txt","a");
//         fprintf(fp, "PE%d Before collect_all_leaf_assignment()\n\n\n",m_myid);
    //}
//     MPI_Finalize();
//     exit(0);
    //    fclose(fp);
    collect_all_leaf_assignment();
}

//Gather the global tree leaf node assignment of all processors.
void distree_t::collect_all_leaf_assignment(){
    svector<int> send_rcounts;//send size array
    svector<int> send_displs;//displacement array.
    svector<int> all_assigned_leafs;//array for all gathered leaf node indices.
    all_assigned_leafs.mpi_allgatherv(m_assigned_local_leaf_indices,send_rcounts,send_displs,MPI_INT);
//     int  local_leaf_count = m_assigned_local_leaf_indices.size();
//     send_rcounts.resize(m_groupsize, 0);
//     //1. gather how many ints each processor is going to send in step 2.
//     MPI_Allgather(&local_leaf_count, 1, MPI_INT, data_ptr(send_rcounts), 1, MPI_INT, m_communicator);
//
//     send_displs.resize(m_groupsize, 0);
//     send_displs[0] = 0;
//     for (int iPE = 1; iPE < m_groupsize ; iPE++){
//         send_displs[iPE] = send_displs[iPE-1] + send_rcounts[iPE-1];
//     }
//     send_gather.resize(send_displs[m_groupsize-1]+send_rcounts[m_groupsize-1], -1);//
//     //2.Gather leaf node assignment from all processor, so that every processor
//     //has a copy of this gathered information.
//     MPI_Allgatherv(data_ptr(m_assigned_local_leaf_indices),//assigned local leaf node array of current PE
//                    local_leaf_count,//send element count for current PE
//                    MPI_INT,//send element type
//                    data_ptr(send_gather),//receive buffer address for ROOT
//                    data_ptr(send_rcounts),//receive element counts from each PE.
//                    data_ptr(send_displs),//displacement in receive buffer for each PE.
//                    MPI_INT,
//                    m_communicator);
    if (m_myid==0){
        FILE* fp = fopen(".\output_debug.txt","a");
        if (!fp){
            D_ASSERT_P(m_myid, fp, "File open failed!", DERROR_ERR_IO);
        }
        fprintf(fp, "After case one step 2.\n\n\n\n");
        for (int iPE = 0 ; iPE < m_groupsize ; iPE ++){
            void* start = &(all_assigned_leafs[send_displs[iPE]]);
            //void* end   = &(all_assigned_leafs[send_displs[iPE]+send_rcounts[iPE]]);
            qsort(start, send_rcounts[iPE], sizeof(int), int_compare);
            fprintf(fp, "++++++++++Start of PE%d++++++++++++++++++++\n",iPE);
            for (int j = 0; j < send_rcounts[iPE] ; j++){
                int k = all_assigned_leafs[send_displs[iPE]+j];
                fprintf(fp, "====assign_leaf[%d]=%d\n",j, k);
            }
            fprintf(fp, "-----------End of PE%d----------------------\n\n",iPE);
        }
        fclose(fp);
    }

    //int total_leaf_count = 0;
    int node_count = m_node_array.size();
    //3. Update the processor array of each leaf node.
    //3.1. clear processor array of every global tree leaf node.
    for (int i = node_count - m_leaf_count ; i < node_count ; i++){
        distree_node_t& cur_node = m_node_array[i];
        if (cur_node.isLeaf){
            cur_node.processor_indices.clear();
            cur_node.processor_3x_indices.clear();
        }
    }
    //sanity check: Check if the # of unique node index in send_gather equals the total
    //# of global tree leaf node, to make sure every leaf node is assigned to at
    //least one processor.
    svector<int> unique_send_gather;
    unique_send_gather.reserve(all_assigned_leafs.size());
    unique_send_gather.clone(all_assigned_leafs);
    unique_send_gather.remove_duplicates(int_compare);
    char err_msg[256];

    sprintf(err_msg,"send_gather.size(%d), unique_send_gather.size(%d) != total_leaf_count(%d)\n",
            all_assigned_leafs.size(),unique_send_gather.size(), m_leaf_count);
    D_ASSERT_P(m_myid, unique_send_gather.size() == m_leaf_count, err_msg, DERROR_ERR_MEM);
    //3.2 Update processor array based on all gathered array: send_gather.
    //node.process_indices includes the processor id which is responsible
    //for the 'node'.
    for (int iPE = 0 ; iPE < m_groupsize ; iPE++){
        for (int j = 0 ; j < send_rcounts[iPE] ; j++){
            int node_idx = all_assigned_leafs[send_displs[iPE]+j];
            D_ASSERT_P(m_myid, node_idx < m_node_array.size(),
                       "node_idx out of bound\n", DERROR_ERR_MEM);
            m_node_array[node_idx].processor_indices.push_back(iPE);
        }
    }
    for (int i = node_count - m_leaf_count ; i < node_count ; i++)
    {//processor list of triple must contains the processor_list that overlap cur_leaf.
        distree_node_t& cur_leaf = m_node_array[i];
        D_ASSERT_P(m_myid, cur_leaf.isLeaf, "cur_leaf is not leaf!", DERROR_ERR_IO);
        cur_leaf.processor_3x_indices.push_back(cur_leaf.processor_indices);
    }
    MPI_Barrier(m_communicator);
    //3.3 find the neighboring PE id in every global tree leaf node's triple range.
    //O(n^2)
    svector<distree_node_t*> leaf_nodes;
    FILE* fp = fopen(".\output_debug.txt","a");
    if (!fp){
        D_ASSERT_P(m_myid, fp, "File open failed!", DERROR_ERR_IO);
    }
    for (int i = node_count - m_leaf_count ; i < node_count ; i++)
    {
        //for each leaf node(no matter local or remote)
        distree_node_t& cur_leaf = m_node_array[i];
        D_ASSERT_P(m_myid, cur_leaf.isLeaf, "cur_leaf is not leaf!", DERROR_ERR_IO);
        if (cur_leaf.isLeaf){
            leaf_nodes.push_back(&m_node_array[i]);
            geometry::BBox cur_nodel_triple = cur_leaf.m_bound;
            cur_nodel_triple.Expand(cur_nodel_triple.MaximumExtentf());

            //cur_leaf.processor_3x_indices.push_back(cur_leaf.processor_indices);
//             if (cur_leaf.location == distree_node_t::LOCAL){
//                 cur_leaf.processor_3x_indices.push_back(m_myid);
//             }
            for (int j = node_count - m_leaf_count ; j < node_count ; j++){
                //for each other leaf node.
                if ( j != i)
                {
                    distree_node_t& cur_other_leaf = m_node_array[j];
                    if (  cur_other_leaf.isLeaf
                       // &&cur_remote_leaf.location == distree_node_t::REMOTE
                        &&cur_nodel_triple.Overlap(cur_other_leaf.m_bound)){
                            int send_proc_id = 5, recv_proc_id = 1;
                            if (m_myid== send_proc_id )
                            {
                                int searchedPid = recv_proc_id;
                                if (cur_leaf.location == distree_node_t::LOCAL
                                    && cur_other_leaf.location == distree_node_t::REMOTE)
                                {
                                    if (cur_other_leaf.processor_indices.contain(searchedPid)){
                                        fprintf(fp, "//////////////////////\n");
                                        fprintf(fp, "/////// PE%d 's cur_leaf(LOCAL)%d triple has--> cur_other_leaf(REMOTE)%d of PE=%d /////////\n",m_myid,i,j, searchedPid);
                                        fprintf(fp, "/////////////////////\n");
                                    }
                                }else if (cur_leaf.location == distree_node_t::REMOTE
                                    && cur_other_leaf.location == distree_node_t::LOCAL)
                                {
                                    if (cur_other_leaf.processor_indices.contain(m_myid)
                                        && cur_leaf.processor_indices.contain(searchedPid)){
                                        fprintf(fp, "/$/$/$/$/$/$/$/$/$/$/$/$/$/$/$/$\n");
                                        fprintf(fp, "/$/$/$/$ PE%d 's cur_other_leaf%d(LOCAL)'s triple has <---- cur_leaf(REMOTE)%d of PE=%d /$/$/$/$\n",m_myid,j,i,searchedPid);
                                        fprintf(fp, "/$/$/$/$/$/$/$/$/$/$/$/$/$/$/$/$\n");
                                    }else if (cur_other_leaf.processor_indices.contain(searchedPid)
                                        && cur_leaf.processor_indices.contain(m_myid)){
                                            fprintf(fp, "/#/#/#?#?/#/#/#?#?/#/#/#?#?/#/#/#?#?\n");
                                            fprintf(fp, "/#/#/#?#? PE%d 's cur_other_leaf%d(LOCAL)'s triple has <---- cur_leaf(REMOTE)%d of PE=%d /#/#/#?#?\n",m_myid,j,i,searchedPid);
                                            fprintf(fp, "/#/#/#?#?/#/#/#?#?/#/#/#?#?/#/#/#?#?/#/#/#?#?\n");
                                    }
                                }
                            }

                        //if cur_remote_leaf overlap with cur_nodel_triple,
                        //then add cur_remote_leaf.processor_indices to cur_local_leaf.processor_3x_range.
                        cur_leaf.processor_3x_indices.push_back(cur_other_leaf.processor_indices);
                    }
                }
            }//end of for j

            //remove duplicates.
            cur_leaf.processor_3x_indices.remove_duplicates(int_compare);
        }// end of if
    }
    fclose(fp);
    //print processor_3x_indices
//     for (int i = 0 ; i < leaf_nodes.size(); i++){
//         distree_node_t& cur_leaf = *(leaf_nodes[i]);
//         int proc_count = cur_leaf.processor_3x_indices.size();
//         for (int j = 0 ; j < proc_count ; j++){
//
//         }
//     }
}

int key_idx_compare(const void *pa,const void *pb)
{
    key_idx_pair_t *a = (key_idx_pair_t *)pa;
    key_idx_pair_t *b = (key_idx_pair_t *)pb;

    if (a->key == b->key) {
        return 0;
    } else {
        if (a->key < b->key)
            return -1;
        else
            return 1;
    }
}

int key_compare(const void *pa,const void *pb)
{
    uint32_t *a = (uint32_t *)pa;
    uint32_t *b = (uint32_t *)pb;

    if (*a == *b) {
        return 0;
    } else {
        if (*a < *b)
            return -1;
        else
            return 1;
    }
}
int int_compare(const void *pa,const void *pb)
{
    int *a = (int *)pa;
    int *b = (int *)pb;

    if (*a == *b) {
        return 0;
    } else {
        if (*a < *b)
            return -1;
        else
            return 1;
    }
}
int distree_vertex_compare(const void *pa,const void *pb)
{
    distree_vertex_t *a = (distree_vertex_t *)pa;
    distree_vertex_t *b = (distree_vertex_t *)pb;

    if (a->key == b->key) {
        return 0;
    } else {
        if (a->key < b->key)
            return -1;
        else
            return 1;
    }
}
//find the intersection of two arrays:empty_leaf_parent_key_idx_pairs and local_leaf_parent_keys,
//and save the values shared by these two arrays in intersection_idx.
void distree_t::arrays_intersection(svector<key_idx_pair_t> &empty_leaf_parent_key_idx_pairs,
                                    svector<uint32_t>       &local_leaf_parent_keys,
                                    svector<int>            &intersection_idx )
{
    //key_idx_pair_t key_idx_compare;

//     sort(empty_leaf_parent_key_idx_pairs.begin(),
//          empty_leaf_parent_key_idx_pairs.end(),
//          key_idx_compare);
    qsort(empty_leaf_parent_key_idx_pairs.data_ptr(),
          empty_leaf_parent_key_idx_pairs.size(),
          sizeof(key_idx_pair_t),
          key_idx_compare);
//     sort(local_leaf_parent_keys.begin(),
//          local_leaf_parent_keys.end());
    qsort(local_leaf_parent_keys.data_ptr(),
          local_leaf_parent_keys.size(),
          sizeof(uint32_t),
          key_compare);

    const svector<key_idx_pair_t> &a = empty_leaf_parent_key_idx_pairs;
    const svector<uint32_t>       &b = local_leaf_parent_keys;
          svector<int>            &c = intersection_idx;
    size_t aindex = 0, asize = a.size();
    size_t bindex = 0, bsize = b.size();

    while (aindex < asize && bindex < bsize) {
        if (a[aindex].key == b[bindex]) {
            c.push_back(a[aindex].idx);
            aindex++;//greedily add the a[i], whose key appears in b[], to c[]
                    //therefore, different a[i]s with the same key are acceptable in our case.
            //bindex++;
        }
        else if (a[aindex].key < b[bindex]) {
            aindex++;
        }
        else {
            bindex++;
        }
    }
}
//create communication schedule in the form of sent_table and recev_table
void distree_t::create_comm_schedule( bitmap_t * bitmap)
{
    //sanity check.
    D_ASSERT_P(m_myid, m_node_array.size() > 0, "node_array is empty",  DERROR_ERR_MEM);
    D_ASSERT_P(m_myid, m_isoblock_array->size_feature() > 0, "isoblock_array is empty", DERROR_ERR_MEM);
    FILE* fp = NULL;
    int send_proc_id = 5, recv_proc_id = 1;
    if (m_myid== send_proc_id || m_myid == recv_proc_id){
        fp = fopen(".\output_debug.txt","a");
        if (!fp){
            D_ASSERT_P(m_myid, fp, "File open failed!", DERROR_ERR_IO);
        }
    }
    for (int i = 0 ; i < m_assigned_local_leaf_indices.size() ; i++)
    {
        distree_node_t *local_leaf =
                        &(m_node_array[m_assigned_local_leaf_indices[i]]);
        //Update cur_local_leaf to the closest ancestor
        //node of leaf whose processors list is not empty.
//         while(
//               ancestor_local_leaf->processor_3x_indices.empty()
//             &&ancestor_local_leaf->parent /*>=*/ > 0 //has parent
//             )
//         {
//             ancestor_local_leaf = &(m_node_array[ancestor_local_leaf->parent]);
//         }
        D_ASSERT_P(m_myid, local_leaf->isLeaf, "create_comm_schedule: sent_table: local_leaf is not leaf", DERROR_ERR_MEM);
        D_ASSERT_P(m_myid, local_leaf->location==distree_node_t::LOCAL, "create_comm_schedule: sent_table: local_leaf is remote", DERROR_ERR_MEM);
        if (local_leaf->isLeaf
            && local_leaf->location == distree_node_t::LOCAL){
                if (m_myid==send_proc_id)
            {
                fprintf(fp, "/////PE%d's leaf%d has the following processor://///////\n", m_myid,i);
            }

            for (int j = 0 ; j < local_leaf->processor_3x_indices.size() ; j++)
            {//select those remote processors that are in ancestor_local_leaf's concentric triple.
                int proc_idx = local_leaf->processor_3x_indices[j];
                if (m_myid==send_proc_id)
                {
                    fprintf(fp, "/////proc%d://///////\n", proc_idx);
                }

                if (proc_idx != m_myid
                    && m_processor_array[proc_idx].m_has_features){
                    m_sent_table.push_back(proc_idx);
                }
            }
            if (m_myid==send_proc_id)
            {
                fprintf(fp, "//////////////////////////////\n");
            }
        }
    }
    ///debug:
    /*
    if (m_myid==127){
       int neighbor = 60;
       bool has = m_processor_array[neighbor].m_has_features;
       int nIsoblk = m_processor_array[neighbor].m_isoblock_indices.size();
       bool bSentNeighbor = false;
       for (int i =0 ; i < m_sent_table.size() ; i++){
           if (m_sent_table[i] == neighbor){
               bSentNeighbor = true;
               break;
           }
       }
       printf("----------------------------------------------------\n");
       printf("---PE%d has_feature=%d, its remote processor %d(total=%d)\
           (has_features=%d,#isoblk=%d----\n)"
           ,m_myid,m_has_features,neighbor,m_processor_array.size(), has, nIsoblk);
       printf("----PE%d--%s-->PE%d----\n",m_myid,bSentNeighbor?"-":"X",neighbor);
       printf("----------------------------------------------------\n");
    }
    */
    //remove duplicates in m_sent_table.
    if ( !m_sent_table.empty()){
        if (m_myid==send_proc_id)
        {
            fprintf(fp, "////PE%d:m_sent_table.size=%d// has_feature=%d/////////\n",m_myid,m_sent_table.size(),m_has_features);
        }else if (m_myid==recv_proc_id)
        {
            fprintf(fp, "$$$ PE%d:m_sent_table.size=%d// has_feature=%d $$$\n",m_myid,m_sent_table.size(),m_has_features);
        }
        m_sent_table.remove_duplicates(int_compare);
        //sort(m_sent_table.begin(), m_sent_table.end());
//         qsort(m_sent_table.data_ptr(),
//               m_sent_table.size(),
//               sizeof(int),
//               int_compare);
//
//         svector<int> unique_sent_table;
//         unique_sent_table.reserve(m_sent_table.size());
//         if (m_sent_table.size() > 0){
//             unique_sent_table.push_back(m_sent_table[0]);
//         }
//         for ( int i= 1 ; i < m_sent_table.size() ; i++)
//         {
//             if (unique_sent_table[unique_sent_table.size()-1]
//                 != m_sent_table[i])
//             {//find new different one.
//                 unique_sent_table.push_back(m_sent_table[i]);
//             }
//         }
//         unique_sent_table.resize(unique_sent_table.size());
//         m_sent_table.swap(unique_sent_table);
    }
    if (m_myid==send_proc_id)
    {
        fprintf(fp, "/:(///After removal m_sent_table.size=%d////:(////\n",m_sent_table.size());
    }else if (m_myid==recv_proc_id)
    {
        fprintf(fp, "$$$ After removal m_sent_table.size=%d////:( $$$\n",m_sent_table.size());
    }
    //-------------------- end  of  sent_table-----------------------
    //--------------------start of recev_table-----------------------
    //only those PEs that have local isoblock will receive msg.
    //check if current PE has isoblock by referencing bitmap.
    if (m_has_features)/*bitmap->isSet(m_myid)*/
    {
        int total_leaf_node = m_node_array.size();
        for (int i = total_leaf_node - m_leaf_count; i < total_leaf_node/*m_assigned_local_leaf_indices.size()*/ ; i++)
//         int node_count = m_node_array.size();//m_assigned_local_leaf_indices.size();
//         for (int i = 0 ; i < node_count ; i++)
        {//scan
            distree_node_t *local_leaf = &(m_node_array[i]);
                            //&(m_node_array[m_assigned_local_leaf_indices[i]]);
//             distree_node_t *ancestor_local_leaf = &(m_node_array[i]);
            D_ASSERT_P(m_myid, local_leaf->isLeaf, "comm_schedule():local_leaf is not leaf\n",DERROR_ERR_MEM);
            if (local_leaf->isLeaf
                //&& local_leaf->location == distree_node_t::REMOTE
                ){
                //Update cur_local_leaf to the closest ancestor
                //node of leaf whose processors list is not empty.
    //             while(
    //                 ancestor_local_leaf->processor_3x_indices.empty()
    //                 &&ancestor_local_leaf->parent /*>=*/ > 0 //has parent
    //                 )
    //             {
    //                 ancestor_local_leaf = &(m_node_array[ancestor_local_leaf->parent]);
    //             }
                svector<int> &processors_3x = local_leaf->processor_3x_indices;
                //int proc_idx=0;

                if(processors_3x.contain(m_myid))
                {//processor list contains myid. Then the remote processors in the leaf's
                 //or ancestor's concentric triple are the ones that will send message to current PE.
                    svector<int> &assigned_processors = local_leaf->processor_indices;
                    if (m_myid==recv_proc_id && assigned_processors.contain(send_proc_id)){
                        fprintf(fp,"#####PE%d's leaf%d's triple contains remote PE%d#####\n",m_myid,i, send_proc_id);
                    }
                    for (int j = 0 ; j < assigned_processors.size() ; j++)
                    {//locate the leaf node leaf or its closest ancestor node
                     //ancestor whose ancestor.processors_3x contains pi's ID.
                        if (assigned_processors[j] != m_myid){
                            m_recv_table.push_back(assigned_processors[j]);
                        }
                    }
                }
            }//end of remote leaf verification
        }
    }
    ///debug:
    /*
    if (m_myid==60){
        int neighbor = 120;
        bool has = m_processor_array[neighbor].m_has_features;
        int nIsoblk = m_processor_array[neighbor].m_isoblock_indices.size();
        bool bRecvNeighbor = false;
        for (int i =0 ; i < m_sent_table.size() ; i++){
            if (m_recv_table[i] == neighbor){
                bRecvNeighbor = true;
                break;
            }
        }
        printf("=======================================\n");
        printf("=====PE%d has feature=%d, its remote processor %d(total=%d)\
               (has_features=%d,#isoblk=%d)====\n"
            ,m_myid, m_has_features,neighbor,m_processor_array.size(),has, nIsoblk);
        printf("=====PE%d--<--%s--PE%d=====\n",m_myid,bRecvNeighbor?"-":"X",neighbor);
        printf("=======================================\n");
    }
    */
    //remove duplicates in m_sent_table.
    if ( !m_recv_table.empty()){
        m_recv_table.remove_duplicates(int_compare);
        //sort(m_recv_table.begin(), m_recv_table.end());
//         qsort(m_recv_table.data_ptr(),
//               m_recv_table.size(),
//               sizeof(int),
//               int_compare);
//
//         svector<int> unique_rev_table;
//         unique_rev_table.reserve(m_recv_table.size());
//         unique_rev_table.push_back(m_recv_table[0]);
//
//         for ( int i = 1 ; i < m_recv_table.size() ; i++)
//         {
//             if (unique_rev_table[unique_rev_table.size()-1]
//                 != m_recv_table[i])
//             {//find new different one.
//                 unique_rev_table.push_back(m_recv_table[i]);
//             }
//         }
//         unique_rev_table.resize(unique_rev_table.size());
//         m_recv_table.swap(unique_rev_table);
    }

    if (m_myid== send_proc_id ){
        if (!fp){
            D_ASSERT_P(m_myid, fp, "File open failed!", DERROR_ERR_IO);
        }
        fprintf(fp, "/######/*/*/PE0 Sent table/#########/\n");
        for (int i = 0 ; i < m_sent_table.size() ; i++)
        {
            fprintf(fp, "/####/ PE%d ---> PE%d /####/\n ", m_myid, m_sent_table[i]);
        }
        fprintf(fp, "/##############################/\n");
        fprintf(fp, "/***/*/*/PE0 Receive table/*/*/*/*/*/*/*/*\n");
        for (int i = 0 ; i < m_recv_table.size() ; i++)
        {
            fprintf(fp, "/*****/ PE%d <--- PE%d /****/\n ", m_myid, m_recv_table[i]);
        }
        fprintf(fp, "/***/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*/*\n");
        fclose(fp);
    }else if (m_myid==recv_proc_id)
    {
        if (!fp){
            D_ASSERT_P(m_myid, fp, "File open failed!", DERROR_ERR_IO);
        }
        fprintf(fp, "/$$$$$$$/PE%d Sent table/$$$$$$$$/\n",m_myid);
        for (int i = 0 ; i < m_sent_table.size() ; i++)
        {
            fprintf(fp, "/$$$$$$$/ PE%d ---> PE%d /$$$$$$$/\n ", m_myid, m_sent_table[i]);
        }
        fprintf(fp, "/$$$$$$$$$$$$$$$$$$$$$$$$$$$$/\n");
        fprintf(fp, "/$$$$$$$/PE%d Receive table/$$$$$$$*\n",m_myid);
        for (int i = 0 ; i < m_recv_table.size() ; i++)
        {
            fprintf(fp, "/$$$$$$$/ PE%d <--- PE%d /$$$$$$$/\n ", m_myid, m_recv_table[i]);
        }
        fprintf(fp, "/$$$$$$$$$$$$$$$$$$$$$$$$$$$$/\n");
        fclose(fp);
    }

    /*
    if (m_myid==0){
        //print sent_table
        printf("--#assigned leaf=%d--sent_table.size=%d----\n",
             m_assigned_local_leaf_indices.size(),m_sent_table.size());
        for (int i = 0 ; i < m_sent_table.size(); i++){
            printf("===PE%d==sent_table[%d]=%d===\n",m_myid, i, m_sent_table[i]);
        }
        //print recv_table
        printf("--has_feature=%d-recv_table.size=%d----\n",m_has_features,m_recv_table.size());
        for (int i = 0 ; i < m_recv_table.size(); i++){
            printf("-----PE%d---m_recv_table[%d]=%d----\n",m_myid, i, m_recv_table[i]);
        }
    }
    */
//     int com_count = m_sent_table.size();
//     char msg[128];
//     for (int i = 0; i <  com_count; i++){
//     	sprintf(msg, "PE%d====>PE%d\n",m_myid, m_sent_table[i]);
//         D_P_MSG(m_myid, m_myid, msg);
//     }
//     com_count = m_recv_table.size();
//     for (int i = 0; i <  com_count; i++){
//         sprintf(msg, "PE%d<====PE%d\n",m_myid, m_recv_table[i]);
//         D_P_MSG(m_myid, m_myid, msg);
//     }
}

void distree_t::verify_schedule( int groupsize )
{
    //Verify.
    /* root gather all blocks */
    svector<int> send_rcounts;
    svector<int> send_displs ;
    svector<int> send_gather ;
    svector<int> recv_rcounts;
    svector<int> recv_displs ;
    svector<int> recv_gather ;
    int root = 0;
    int iPE, j, pe_id;

    char msg[256];
    if (m_myid == root) {
        send_rcounts.resize(groupsize,0);//send_rcounts = (int *)calloc(groupsize, sizeof(int));
    }
    //Gather sent_table size from all processors.
    int size = m_sent_table.size();
    const int INVALID_ID = groupsize;
    MPI_Gather(&size, 1, MPI_INT, data_ptr(send_rcounts), 1, MPI_INT, root, MPI_COMM_WORLD);

    if (m_myid == root) {
        send_displs.resize(groupsize,0);
        //exclusive prefix sum(exclusive scan).
        send_displs[0] = 0;
        for (iPE = 1; iPE < groupsize; iPE++) {
            send_displs[iPE] = send_displs[iPE-1] + send_rcounts[iPE-1];
        }
        send_gather.resize(send_displs[groupsize-1] + send_rcounts[groupsize-1], INVALID_ID);
    }

    MPI_Gatherv(data_ptr(m_sent_table),//send buffer address for current PE
                m_sent_table.size(),//send element count for current PE
                MPI_INT,//send element type
                data_ptr(send_gather),//receive buffer address for ROOT
                data_ptr(send_rcounts),//receive element counts from each PE.
                data_ptr(send_displs),//displacement in receive buffer for each PE.
                MPI_INT,
                root,//The id of the PE that will receive the gather info.
                MPI_COMM_WORLD);

    if (m_myid == root) {
        recv_rcounts.resize(groupsize,0);
    }
    size = m_recv_table.size();
    MPI_Gather(&size, 1, MPI_INT, data_ptr(recv_rcounts), 1, MPI_INT, root, MPI_COMM_WORLD);

    if (m_myid == root) {
        recv_displs.resize(groupsize,0);
        recv_displs[0] = 0;
        for (iPE = 1; iPE < groupsize; iPE++) {
            recv_displs[iPE] = recv_displs[iPE-1] + recv_rcounts[iPE-1];
        }
        recv_gather.resize(recv_displs[groupsize-1] + recv_rcounts[groupsize-1], INVALID_ID);
    }

    MPI_Gatherv(data_ptr(m_recv_table),
                m_recv_table.size(),
                MPI_INT,
                data_ptr(recv_gather),
                data_ptr(recv_rcounts),
                data_ptr(recv_displs),
                MPI_INT,
                root,
                MPI_COMM_WORLD);

    if (m_myid == root) {
        //create communication matrix based on send&recv table
        //and check if it is a symmetric
        svector<uint8_t>
        matrix_sent(groupsize*groupsize,0);//(uint8_t*) calloc(groupsize*groupsize, sizeof(char));
        svector<uint8_t>
        matrix_rev(groupsize*groupsize,0);//(uint8_t*) calloc(groupsize*groupsize, sizeof(char));
        //matrix_sent[iPE][pe_id]==1: PE i would sent msg to PE pe_id.
        for (iPE = 0; iPE < groupsize ; iPE++){//for each sent PE
            for (j = 0; j < send_rcounts[iPE] ; j++){//for each destination PE
                pe_id = send_gather[send_displs[iPE]+j];//destination PE id
                if (pe_id>=groupsize){
                    sprintf(msg,"assert failed.pe_id=%d,groupsize=%d\n",pe_id,groupsize);
                    D_P_MSG(m_myid,0,msg);
                }
                D_ASSERT_P(m_myid, pe_id < groupsize, "Out of memory", DERROR_ERR_MEM);
                matrix_sent[iPE*groupsize + pe_id] = 1;
            }
        }
        //fprintf(stderr, "+++++Proc%d receive %d request+++++++++ %d\n",m_myid, recv_rcounts[0]);
        //matrix_rev[iPE][pe_id]==1: PE i would receive msg from PE pe_id.
        for (iPE = 0; iPE < groupsize ; iPE++){
            for (j = 0; j < recv_rcounts[iPE] ; j++){
                pe_id = recv_gather[recv_displs[iPE]+j];
                if (pe_id>=groupsize){
                    sprintf(msg,"assert failed.pe_id=%d,groupsize=%d\n",pe_id,groupsize);
                    D_P_MSG(m_myid,0,msg);
                }
                D_ASSERT_P(m_myid, pe_id < groupsize, "Out of memory", DERROR_ERR_MEM);
                matrix_rev[pe_id*groupsize + iPE] = 1;
            }
        }
        //compare if the two matrix are the same.
        for (iPE = 0; iPE < groupsize ; iPE++){
            for (j = 0; j < groupsize ; j++){
                if (  matrix_sent[iPE*groupsize+j]
                   != matrix_rev[iPE*groupsize+j]){
                       sprintf(msg, "PE%d: mtx_sent[%d,%d]=%d,mtx_rev[%d,%d]=%d\n",m_myid,
                                iPE,j, matrix_sent[iPE*groupsize+j],
                                iPE,j, matrix_rev[iPE*groupsize+j]);
                                //iPE,j, matrix_rev[iPE*groupsize+j]);
                       D_P_MSG(m_myid,0, msg);
                       D_P_MSG(m_myid,0, "verification failed!\n");
                }
            }
        }
        D_P_MSG(m_myid,0, "verification passed!\n");
    }
}

void distree_t::refine_global_distree()
{
    m_features->initial_distree_feature_array(*this);
    //-------------------------------------------------------
    //initial 3x range feature of root node
    distree_node_t*
    cur_root_node   = &(m_node_array[0]);
    cur_root_node->triple_feature_indices.resize(m_features->size_feature());

    for (int i = 0; i < cur_root_node->triple_feature_indices.size() ; i++)
    {//root node contains all voxels. it is trivial to initialize
      //its triple feature indices array.
        cur_root_node->triple_feature_indices[i] = i;
    }
    //update isovoxel of existing global_distree node in BFS order.
    int parent_end   = 1;//level 0 has only root node
    int parent_start = 0;//start with root node
    int child_index  = 1;

    while(parent_start < m_node_array.size())
    {
        for (int parent_id = parent_start; parent_id < parent_end ; parent_id++)
        {//for each parent in current level. locate its 8 children if there is any.
            //int cur_parent_index = parent_start + parent_id;
            distree_node_t& cur_parent = m_node_array[parent_id];//m_node_array[cur_parent_index];
            if (!cur_parent.isLeaf)
                //&& cur_parent.triple_feature_indices.size() > 0)//don't need this, this would cause infinite loop, if the root node has no feature at all.
            {
                if (cur_parent.triple_feature_indices.size() > 0){
                    //parent has 8 children and has features. distribute the features
                    //to its 8 children.
                    for (int j = 0; j < 8 ; j++)
                    {//for each child.
                        int cur_child_index = cur_parent.m_children[j];
                        D_ASSERT_P(m_myid, cur_child_index >= 0,
                            "refine_global_distree():cur_child_index < 0", DERROR_ERR_MEM);
//                         D_ASSERT_P(m_myid, cur_child_index == child_index,
//                                     "refine_global_distree(): cur_child_index != child_index", DERROR_ERR_MEM);
                        m_features->collect_feature(child_index, parent_id, *this);
                        child_index ++ ;
                    }
                }else{
                    //if parent has no feature, we simply skip its 8 child nodes.
                    child_index+=8;
                }
            }
        }
        //parent offset index move to next level.
        parent_start = parent_end;
        parent_end   = child_index;
    }

    char msg[256];
    sprintf(msg, "parent_end=%d != m_node_array.size()=%d",parent_end, m_node_array.size());
    D_ASSERT_P(m_myid, parent_end == m_node_array.size(), msg , DERROR_ERR_MEM);
    {
        D_P_MSG(m_myid, m_myid
            ,"^^^pass propagating triple features^^^^^\n");
    }
    //---------------------Refined Global Distree--------------------------
    //subdivide the tree node specified in tree->local_region in BFS order.
    int children_start_offset = (int)m_node_array.size();
    int node_size = (int)m_node_array.size();//node size before growing local tree.
    for (int j = 0 ; j < node_size ; j ++)
    {//scan each leaf node of global tree
        if (     m_node_array[j].isLeaf //for each leaf node
            && (!(m_node_array[j].triple_feature_indices.empty())
                ||!(m_node_array[j].isoblock_3x_indices.empty()))
            )//whose elements is not empty
        {
            //----------subdivide current leaf node------------------
            int parent_start_offset = j;//local_region[j].node_index;
            int cur_parent_index    = parent_start_offset;
            int parent_count        = 1;
            int parent_depth_id     = m_node_array[cur_parent_index].level;

            while (parent_depth_id < max_depth_id && parent_count)
            {
                int children_count = 0;
                for ( int i = 0 ; i < parent_count ; i++)
                {//for each parent in current level.
                 //the leaf octants are recursively split if its elements is not empty.
                    cur_parent_index = parent_start_offset + i;
                    distree_node_t *cur_parent = &(m_node_array[cur_parent_index]);
                    if (!cur_parent->triple_feature_indices.empty())
                    {
                        bool ret = create_children(cur_parent_index, m_features);
                        children_count = ret ? children_count+8 : children_count;
                        cur_parent = &(m_node_array[cur_parent_index]);
                        cur_parent->isLeaf = !ret;//children created, parent is no longer leaf.
                        //if parent_node is local and has 3x range isoblock,
                        //then collect 3x range isoblock of each new child node.
                        if (cur_parent->location == distree_node_t::LOCAL
                            && !cur_parent->isLeaf
                            && cur_parent->isoblock_3x_indices.size()>0)
                        {//For a local leaf octant, its elementblocks list also
                         //needs to be split and propagated among the new children.
                            for(int k=0 ; k < 8 ; k++)
                            {
                                child_index = cur_parent->m_children[k];
                                D_ASSERT_P(m_myid, child_index < m_node_array.size()
                                            && child_index >= 0,
                                            "refine_global_distree(): child_index out of bound" ,
                                            DERROR_ERR_MEM);
                                m_isoblock_array->collect_feature(child_index, cur_parent_index, *this);
                            }
                        }
                    }
                    parent_depth_id = cur_parent->level;
                }
                //parent child offset index move to next level.
                parent_count           = children_count;
                parent_start_offset    = children_start_offset;
                children_start_offset  = children_start_offset + children_count;

                D_ASSERT_P(m_myid, children_start_offset == m_node_array.size(),
                            "refine_global_distree(): children_start_offset!=m_node_array.size()" ,
                            DERROR_ERR_MEM);
            }
        }//end if is leaf node
    }
}

void distree_t::create_local_distree_vertex_array()
{
    const vec3f vertex_offset[8]= {
        vec3f(0, 0, 0), vec3f(1, 0, 0), vec3f(0, 1, 0), vec3f(1, 1, 0),
        vec3f(0, 0, 1), vec3f(1, 0, 1), vec3f(0, 1, 1), vec3f(1, 1, 1)
    };
    //The vertex array of each processor contains the vertices of
    //both local leaf nodes and remote leaf nodes
    for (int i = 0 ; i < m_node_array.size() ; i++)
    {
        distree_node_t &leaf_node = m_node_array[i];
        if (leaf_node.isLeaf
            //&& leaf_node.location == distree_node_t::LOCAL
            //Not just local leaf, but also remote leaf node. Or else
            //remote vertex distance can not be computed in local PE
            )
        {
            float extend = leaf_node.m_bound.MaximumExtentf();

            for (int j = 0; j < 8 ; j++)
            {
                m_vertex_array.push_back(distree_vertex_t());
                distree_vertex_t& cur_vertex = m_vertex_array.back();
                cur_vertex.coords = leaf_node.m_bound.pMin + vertex_offset[j]*extend;
                cur_vertex.node_index = i;
                cur_vertex.node_local_id = j;
                cur_vertex.key = compute_octree_vertex_key(cur_vertex);
            }
        }
    }
    //Remove duplicates.
    //1.Sort vertex_array based on vertex's key.
    //distree_vertex_t compare_vertex;
    //sort(m_vertex_array.begin(), m_vertex_array.end(), compare_vertex);
    qsort(m_vertex_array.data_ptr(),
          m_vertex_array.size(),
          sizeof(distree_vertex_t),
          distree_vertex_compare);

    int unique_count = 0;
    distree_vertex_t *cur_vertex = &(m_vertex_array[0]);
    int node_index  = cur_vertex->node_index;
    int local_id    = cur_vertex->node_local_id;
    m_vertex_array[0].eucl_dist    = FLOAT_MAX;
    m_vertex_array[0].maxnorm_dist = FLOAT_MAX;
    m_vertex_array[0].is_send      = 0;
    distree_node_t *cur_node = &(m_node_array[node_index]);
    cur_node->vertex_indices[local_id] = 0;

    int vertex_count = m_vertex_array.size();
    for (int vtx_idx = 1; vtx_idx < vertex_count ; vtx_idx++) {
        //scan vertex_array, and collect different vertices.
        cur_vertex = &(m_vertex_array[vtx_idx]);
        node_index = cur_vertex->node_index;
        local_id = cur_vertex->node_local_id;
        if (m_vertex_array[unique_count].key
            != cur_vertex->key)
        {
            unique_count++;
            m_vertex_array[unique_count] = *cur_vertex;
            m_vertex_array[unique_count].eucl_dist    = FLOAT_MAX;
            m_vertex_array[unique_count].maxnorm_dist = FLOAT_MAX;
            m_vertex_array[unique_count].is_send      = 0;
        }
        cur_node = &(m_node_array[node_index]);
        cur_node->vertex_indices[local_id] = unique_count;//-1;
    }
    unique_count++;
    m_vertex_array.resize(unique_count);
    return;

//     svector<distree_vertex_t> unique_vertex_array(unique_count);
//     for (int i = 0; i < unique_count; i++){
//     	unique_vertex_array[i] = m_vertex_array[i];
//     }
//     m_vertex_array.swap(unique_vertex_array);
    //printf("++++++++++end of create vertex array++++++");

    //m_vertex_array.resize(unique_count+1);
    //reserve no larger than m_vertex_array.
    /*
    svector<distree_vertex_t> unique_vertex_array(m_vertex_array.size());
    int unique_count = 0;
    unique_vertex_array[0] = m_vertex_array[0];
    unique_vertex_array[0].eucl_dist    = FLOAT_MAX;
    unique_vertex_array[0].maxnorm_dist = FLOAT_MAX;
    unique_vertex_array[0].is_send      = 0;
    unique_count++;
    distree_vertex_t *cur_vertex = &(m_vertex_array[0]);
    distree_node_t *cur_node = &(m_node_array[cur_vertex->node_index]);
    cur_node->vertex_indices[cur_vertex->node_local_id] = 0;

    int vertex_count = m_vertex_array.size();
    for (int vtx_idx = 1; vtx_idx < vertex_count ; vtx_idx++) {
        //scan vertex_array, and collect different vertices.
        cur_vertex = &(m_vertex_array[vtx_idx]);
        if (unique_vertex_array[unique_count - 1].key
            != cur_vertex->key)
        {
            unique_vertex_array[unique_count] = *cur_vertex;
            unique_vertex_array[unique_count].eucl_dist    = FLOAT_MAX;
            unique_vertex_array[unique_count].maxnorm_dist = FLOAT_MAX;
            unique_vertex_array[unique_count].is_send      = 0;
            unique_count++;
        }

        cur_node = &(m_node_array[cur_vertex->node_index]);
        cur_node->vertex_indices[cur_vertex->node_local_id]
                = unique_count-1;
    }
    //Compact array.
    //printf("********unique count=%d********\n",unique_count);
    unique_vertex_array.resize(unique_count);
    m_vertex_array.swap(unique_vertex_array);
    */
}
void distree_t::load_distree_binary(char *filename)
{
    FILE* fp = fopen(filename,"rb");
    D_ASSERT_P(m_myid, fp, "Can not open file", DERROR_ERR_IO);
    //center
    fread((float*)m_center, sizeof(vec3f), 1, fp);
    //global bounding box
    fread((float*)(m_global_domain_bound.pMin), sizeof(vec3f), 1, fp );
    fread((float*)(m_global_domain_bound.pMax), sizeof(vec3f), 1, fp );

    //Node array size
    int node_array_size = 0;
    fread(&node_array_size, sizeof(int), 1, fp );
    printf("node_array_size=%d\n", node_array_size);
    printf("sizeof(distree_node_t)=%d\n",sizeof(distree_node_t));
    if (node_array_size > 0){
        m_node_array.resize(node_array_size);
        fread(&(m_node_array[0]), sizeof(distree_node_t), node_array_size, fp );
    }

    //region array
    int assigned_local_leaf_num = 0;
    fread(&assigned_local_leaf_num, sizeof(int), 1, fp);
    if (assigned_local_leaf_num > 0){
        m_assigned_local_leaf_indices.resize(assigned_local_leaf_num);
        fread(&(m_assigned_local_leaf_indices[0]), sizeof(int), assigned_local_leaf_num, fp );
    }

    //vertex array size
    int vertex_array_size = 0;
    fread(&(vertex_array_size), sizeof(int), 1, fp );
    printf("sizeof(distree_vertex_t)=%d\n", sizeof(distree_vertex_t));
    printf("Proc%d vertex_array_size=%d\n", m_myid, vertex_array_size);
    //vertex array
    if (vertex_array_size > 0){
        m_vertex_array.resize(vertex_array_size);
        fread(&(m_vertex_array[0]), sizeof(distree_vertex_t), vertex_array_size, fp );
    }
    //feature type
    fread(&(m_feature_type), sizeof(feature_type_t), 1, fp);
    //feature size
    int feature_size;
    fread(&(feature_size), sizeof(int), 1, fp );
    printf("feature_array_size=%d\n", feature_size);
    if (feature_size > 0){
        switch(m_feature_type){
        case F_VOLUME:{
            //point feature
            m_features = new volume_t(vec3i(1,1,1), 0.1f);
            volume_t *feature_volume = dynamic_cast<volume_t*>(m_features);
            if (feature_volume){
                feature_volume->m_isovoxels.resize(feature_size);
                fread(data_ptr(feature_volume->m_isovoxels), sizeof(voxel_t),
                      feature_size, fp );
            }
            break;
                      }
        case F_TRIANGLE:{
            //triangle feature
            m_features = new feature_geometry_t();
            feature_geometry_t* triangles = dynamic_cast<feature_geometry_t*>(m_features);
            if (triangles){
                triangles->m_features_triangle.resize(feature_size);
                fread(data_ptr(triangles->m_features_triangle), sizeof(triangle_t),
                      feature_size, fp );
            }
            break;
                        }
        }
    }
    fclose(fp);
}
void distree_t::save_distree_binary(char *treeName)
{
    char name[1024];
    FILE *fp = NULL;
    distree_node_t   *cur_node=NULL;
    distree_vertex_t *cur_vertex=NULL;
    //int dist_type_count[3];
    int local_dist_count=0;
    int parent_dist_count=0;
    int grandparent_dist_count=0;
    sprintf(name,"%s%d.dat",treeName,m_myid);
    D_P_MSG(m_myid, m_myid, name);
    fp = fopen(name,"wb");
    D_ASSERT_P(m_myid, fp, "Can not open file", DERROR_ERR_IO);
    //center
    fwrite((float*)m_center, sizeof(vec3f), 1, fp );
    //global_min/max
    fwrite((float*)(m_global_domain_bound.pMin), sizeof(vec3f), 1, fp );
    fwrite((float*)(m_global_domain_bound.pMax), sizeof(vec3f), 1, fp );

    //Node array size
    int node_array_size = m_node_array.size();
    fwrite(&node_array_size, sizeof(int), 1, fp );
    if (m_myid==0){
        printf("node_array_size=%d\n", node_array_size);
        //Node array
        printf("sizeof(distree_node_t)=%d\n",sizeof(distree_node_t));
    }
    if (node_array_size > 0){
        fwrite(&(m_node_array[0]), sizeof(distree_node_t), node_array_size, fp );
    }

    //region array
    int assigned_local_leaf_num = m_assigned_local_leaf_indices.size();
    fwrite(&assigned_local_leaf_num, sizeof(int), 1, fp);
    if (assigned_local_leaf_num > 0){
        fwrite(&(m_assigned_local_leaf_indices[0]), sizeof(int), assigned_local_leaf_num, fp );
    }

    //vertex array size
    int vertex_array_size = m_vertex_array.size();
    fwrite(&(vertex_array_size), sizeof(int), 1, fp );
    if (m_myid==0){
        printf("sizeof(distree_vertex_t)=%d\n", sizeof(distree_vertex_t));
    }
    printf("Proc%d vertex_array_size=%d\n",m_myid, vertex_array_size);
    //vertex array
    if (vertex_array_size > 0){
//         for (int i = 0; i < vertex_array_size ; i++){
//         	distree_vertex_t& cur_vtx = m_vertex_array[i];
//             if (cur_vtx.is_send){
//                 printf("===PE%d vtx[%d] send====\n",m_myid, i);
//             }
//         }
        fwrite(&(m_vertex_array[0]), sizeof(distree_vertex_t), vertex_array_size, fp );
    }
    /*
    dist_type_count[0]=0;
    dist_type_count[1]=0;
    dist_type_count[2]=0;
    for (i = 0; i < vertex_array_size; i++){
        dist_type_count[m_vertex_array[i].min_dist_type]++;
    }
    */
//     printf("Proc%d min dist self=%d\n", m_myid, dist_type_count[0]);
//     printf("Proc%d min dist parent=%d\n", m_myid, dist_type_count[1]);
//     printf("Proc%d min dist grandparent=%d\n", m_myid, dist_type_count[2]);
    //feature type
    fwrite(&(m_feature_type), sizeof(int), 1, fp );
    //feature size
    int feature_array_size = m_features->size_feature();
    fwrite(&(feature_array_size), sizeof(int), 1, fp );
    //if (m_myid==0){
        printf("feature_array_size=%d\n", feature_array_size);
    //}
    if (feature_array_size > 0 ){
        switch(m_feature_type){
        case F_VOLUME:{
            //point feature
            if (m_myid==0){
                printf("save voxel array\n");
            }
            volume_t* volume = dynamic_cast<volume_t *>(m_features);
            if (volume){
                fwrite(volume->data(), sizeof(voxel_t), feature_array_size, fp );
            }
            break;
                      }
        case F_TRIANGLE:{
            //triangle feature
            if (m_myid==0){
                printf("save triangle array\n");
            }
            feature_geometry_t* triangles = dynamic_cast<feature_geometry_t *>(m_features);
            if (triangles){
                fwrite(triangles->data(), sizeof(triangle_t), feature_array_size, fp );
            }

            break;
                        }
        }
    }
    fclose(fp);
}
void distree_t::save_distree_text(char* treeName){
    char name[1024];
    FILE *fp = NULL;
    distree_node_t   *cur_node=NULL;
    distree_vertex_t *cur_vertex=NULL;
    sprintf(name,"%s%d.txt",treeName,m_myid);
    fp = fopen(name,"w");
    D_ASSERT_P(m_myid, fp, "Can not open file", DERROR_ERR_IO);
    /*
    //Vertex Array
    for (i = 0; i < tree->vertex_array_size ; i++){
        cur_vertex = &(tree->vertex_array[i]);

            fprintf(fp,"vertex%d\n\
                       key=%d\n\
                       node_index=%d\n\
                       node_local_id=%d\n\
                       ecul_dist=%f\n\
                       norm_dist=%f\n\
                       %f, %f, %f\n"
                       ,i
                       ,cur_vertex->key
                       ,cur_vertex->node_index
                       ,cur_vertex->node_local_id
                       ,cur_vertex->eucl_dist
                       ,cur_vertex->maxnorm_dist
                       ,cur_vertex->coords.x3d,cur_vertex->coords.y3d,cur_vertex->coords.z3d
                       );
    }
    */
    //Node array.
    int node_array_size = m_node_array.size();
    fprintf(fp,"node_size=%d\n", node_array_size);
    for (int i = 0; i < node_array_size ; i++){
        cur_node = &(m_node_array[i]);
        vec3f center = cur_node->m_bound.Center();
    	fprintf(fp,"node%d\n\
                   l=%d,leaf=%d, p=%d\n\
                   k=%d,location=%d\n\
                   %f, %f, %f\n\
                   %f, %f, %f\n\
                   %f, %f, %f\n\
                   %d, %d, %d, %d\n\
                   %d, %d, %d, %d\n\
                   %d, %d, %d, %d\n\
                   %d, %d, %d, %d\n\
                   \n",i
                   ,cur_node->level,cur_node->isLeaf, cur_node->parent
                   ,cur_node->key,cur_node->location
                   ,cur_node->m_bound.pMin.x(),cur_node->m_bound.pMin.y(),cur_node->m_bound.pMin.z()
                   ,cur_node->m_bound.pMax.x(),cur_node->m_bound.pMax.y(),cur_node->m_bound.pMax.z()
                   ,center.x(), center.y(), center.z()
                   ,cur_node->m_children[0],cur_node->m_children[1],cur_node->m_children[2],cur_node->m_children[3]
                   ,cur_node->m_children[4],cur_node->m_children[5],cur_node->m_children[6],cur_node->m_children[7]

                   //,cur_node->triple_features.feature_num,     cur_node->isoblock_3x_num
                   ,cur_node->vertex_indices[0],cur_node->vertex_indices[1],cur_node->vertex_indices[2],cur_node->vertex_indices[3]
                   ,cur_node->vertex_indices[3],cur_node->vertex_indices[5],cur_node->vertex_indices[6],cur_node->vertex_indices[7]);
    }

    fclose(fp);
}

void distree_t::compute_vertex_distance_parallel()
{
    //compute distance from leaf node vertices to the nearest voxel features.
    //in the LOCAL VOLUME.
    HPGV_TIMING_BEGIN(parallel_test::DIST_LOCAL);
    HPGV_TIMING_COUNT(parallel_test::DIST_LOCAL);
    compute_local_distance();
    HPGV_TIMING_END(parallel_test::DIST_LOCAL);
    //MPI_Barrier(MPI_COMM_WORLD);

    //compute distance field cross PE by sending local_distree's vertex_array to other
    //PEs listed in the sent table and prepare for incoming requests from other PEs
    //listed in recev_table.
    //send first.
    HPGV_TIMING_BEGIN(parallel_test::COM_MEM_TIME);
    int send_count = m_sent_table.size();;//send_table->size;
    //send_schedule_t *send_schedule = NULL;
    //recv_schedule_t *recv_schedule = NULL;
    MPI_Status status;
    svector<send_schedule_t> send_schedule;
    svector<recv_schedule_t> recv_schedule;
    svector<MPI_Status>  my_send_status;
    svector<MPI_Status>  my_recv_status;
    svector<MPI_Request> my_send_request;
    svector<MPI_Request> my_recv_request;
    svector<MPI_Request> dummy_send_request;
    svector<MPI_Request> dummy_recv_request;
    char msg[256];
    if (send_count>0){
        send_schedule.resize(send_count);
        my_send_status.resize(send_count);
        my_send_request.resize(send_count);
        dummy_send_request.resize(send_count);
    }
    HPGV_TIMING_END(parallel_test::COM_MEM_TIME);

    HPGV_TIMING_BEGIN(parallel_test::PREPARE_SEND_BUFFER);
    HPGV_TIMING_COUNT(parallel_test::PREPARE_SEND_BUFFER);

    create_send_schedule_vertex(send_schedule);

    HPGV_TIMING_END(parallel_test::PREPARE_SEND_BUFFER);
    /*{
        D_P_MSG(m_myid, m_myid
            ,"++++++++++pass create_send_schedule_vertex^^^^^\n");
    }*/
    //----------Dummy send and receive------------.
    //This phase of communication tells the remote PEs how many vertex keys it is expected
    //to receive from current PE, so that remote PE can preallocate the buffer for the incoming
    //vertex key stream in the phase 1 step.
    HPGV_TIMING_BEGIN(parallel_test::COM_MEM_TIME);
    for (int i = 0; i < send_count ; i++)
    {
        int send_size = send_schedule[i].send_key_buf.size();
        MPI_Isend(&send_size,//dummy value
                    1,
                    MPI_INT,
                    send_schedule[i].recv_id,
                    MY_DUMMY_MSG,
                    MPI_COMM_WORLD,
                    &(dummy_send_request[i]));
//         if (m_myid==17){
//             sprintf(msg, "$$$$$$$ PE%d --sendsize=%d--> PE%d $$$$$$$$$$$$$$$\n"
//                 ,m_myid
//                 ,send_size
//                 ,send_schedule[i].recv_id);
//             D_P_MSG(m_myid, m_myid, msg);
//         }
    }
    int recv_remote_proc_count = m_recv_table.size();
    svector<int> reverse_lookup_table(m_groupsize, -1);
    svector<int> recv_size(recv_remote_proc_count,0);

    if (recv_remote_proc_count > 0)
    {
        dummy_recv_request.resize(recv_remote_proc_count, 0);
        recv_schedule.resize(recv_remote_proc_count);
        for (int i = 0; i < recv_remote_proc_count; i++) {
            recv_schedule[i].send_id = m_recv_table[i];
            reverse_lookup_table[recv_schedule[i].send_id] = i;
        }
        //receive the number of vertex keys from remote PE.
        int count = 0, flag = 0;
        while(count < recv_remote_proc_count)
        {
            MPI_Iprobe(MPI_ANY_SOURCE, MY_DUMMY_MSG, MPI_COMM_WORLD,
                        &flag, &status);
            if (flag)
            {
                int recv_schedule_idx = reverse_lookup_table[status.MPI_SOURCE];
                    if (m_myid == 2
                        //m_myid == 45 ||m_myid == 49 || m_myid == 54 || m_myid ==  59
                        ){
                        //print out recv_schedule.
//                         for (int i = 0; i < recv_remote_proc_count ; i++){
//                         	printf("+++PE%d:recv_schedule[%d].send_id=%d, recv_schedule_idx=%d++\n",m_myid
//                                  ,i, recv_schedule[i].send_id,recv_schedule_idx);
//                         }
                    }
                if (status.MPI_SOURCE != recv_schedule[recv_schedule_idx].send_id){
                    char msg[256];
                    sprintf(msg,"status.MPI_SOURCE(%d)!=recv_schedule[].send_id(%d)"
                           ,status.MPI_SOURCE, recv_schedule[recv_schedule_idx].send_id);
                    D_ASSERT_P(m_myid, status.MPI_SOURCE == recv_schedule[recv_schedule_idx].send_id,
                               msg, DERROR_ERR_COMM);
//                     volatile char* p = NULL;
//                     *p = 'a';//suicide bomber!
                }

                int size = 0;
                MPI_Get_count(&status, MPI_INT, &size );

                if (size > 0){
                    MPI_Irecv(&(recv_size[recv_schedule_idx]),
                                size,
                                MPI_INT,
                                status.MPI_SOURCE,
                                MY_DUMMY_MSG,
                                MPI_COMM_WORLD,
                                &(dummy_recv_request[recv_schedule_idx]));
                }
                count++;
            }
        }
    }
    MPI_Waitall(send_count, data_ptr(dummy_send_request), MPI_STATUSES_IGNORE);
    MPI_Waitall(recv_remote_proc_count, data_ptr(dummy_recv_request), MPI_STATUSES_IGNORE);
    MPI_Barrier(MPI_COMM_WORLD);

    dummy_recv_request.clear();
    //allocate buffer for recv_schedule
    for (int i = 0; i < recv_remote_proc_count ; i++){
        recv_schedule[i].recv_buf.resize(recv_size[i]);
        //             if (m_myid==0){
        //                 sprintf(msg, "+++++++++++recv_schedule[from PE%d].recv_size=%d != , recv_size[%d]=%d+++++++++++\n"
        //                     ,recv_schedule[i].send_id
        //                     ,recv_schedule[i].recv_buf.size()
        //                     ,i
        //                     ,recv_size[i]);
        //                 D_P_MSG(m_myid, m_myid, msg);
        //             }
    }
    /*{
        D_P_MSG(m_myid, m_myid
            ,"++++++++++pass Exchange dummy vertex count^^^^^\n");
    }*/
    my_recv_request.resize(recv_remote_proc_count);

    HPGV_TIMING_COUNT(parallel_test::COM_MEM_TIME);
    HPGV_TIMING_END(parallel_test::COM_MEM_TIME);
//========================end of Receive buffer allocation===============
    //---------Phase 1 send selected vertices to remote PE for distance computation.
    HPGV_TIMING_BEGIN(parallel_test::COMM_PHASE_1);
    HPGV_TIMING_COUNT(parallel_test::COMM_PHASE_1);
    HPGV_TIMING_COUNT(parallel_test::MY_SEND_BYTE1);
    for (int i = 0; i < send_count ; i++)
    {
        MPI_Isend(data_ptr(send_schedule[i].send_key_buf),
                    send_schedule[i].send_key_buf.size(),
                    MPI_INT,
                    send_schedule[i].recv_id,
                    MY_SEND_MSG,
                    MPI_COMM_WORLD,
                    &(my_send_request[i]));

        HPGV_TIMING_INCREASE(parallel_test::MY_SEND_BYTE1,
                    send_schedule[i].send_key_buf.size()*sizeof(int));
        //sprintf(msg, "PE%d==>PE%d\n",myid, send_schedule[i].recv_id);
        //D_P_MSG(myid, myid, msg);
    }
    if (recv_remote_proc_count > 0)
    {
		HPGV_TIMING_COUNT(parallel_test::MY_RECV_BYTE1);
        int count = 0, flag = 0;
        while(count < recv_remote_proc_count)
        {
            MPI_Iprobe(MPI_ANY_SOURCE, MY_SEND_MSG, MPI_COMM_WORLD,
                        &flag, &status);
            if (flag)
            {
                int recv_schedule_idx = reverse_lookup_table[status.MPI_SOURCE];
                D_ASSERT_P(m_myid, status.MPI_SOURCE == recv_schedule[recv_schedule_idx].send_id,
                            "**MPI_SOURCE!=send_id**\n", DERROR_ERR_MEM);
                int size = 0;
                MPI_Get_count(&status, MPI_INT, &size );

                if (size > 0){
//                     if(m_myid==0 /*&& status.MPI_SOURCE==20*/)
//                     {
//                         sprintf(msg, "--------------recv_schedule[from PE%d].recv_size=%d != size=%d, recv_size[%d]=%d--------------\n"
//                             ,status.MPI_SOURCE
//                             ,recv_schedule[recv_schedule_idx].recv_buf.size()
//                             ,size
//                             ,recv_schedule_idx
//                             ,recv_size[recv_schedule_idx]);
//                         D_P_MSG(m_myid, m_myid, msg);
//                    }
                    D_ASSERT_P(m_myid, recv_schedule[recv_schedule_idx].recv_buf.size() == size,
                                msg, DERROR_ERR_COMM);

                    MPI_Irecv(data_ptr(recv_schedule[recv_schedule_idx].recv_buf),
                              size,
                              MPI_INT,
                              status.MPI_SOURCE,
                              MY_SEND_MSG,
                              MPI_COMM_WORLD,
                              &(my_recv_request[recv_schedule_idx]));

                    HPGV_TIMING_INCREASE(parallel_test::MY_RECV_BYTE1,
                        recv_schedule[recv_schedule_idx].recv_buf.size()*sizeof(int));
                }
                count++;
            }
        }//end of while
    }

    MPI_Waitall(send_count, data_ptr(my_send_request), MPI_STATUSES_IGNORE);
    MPI_Waitall(recv_remote_proc_count, data_ptr(my_recv_request), MPI_STATUSES_IGNORE);
    HPGV_TIMING_END(parallel_test::COMM_PHASE_1);
    /*{
        D_P_MSG(m_myid, m_myid
            ,"++++++++++pass Exchange vertex keys^^^^^\n");
    }*/
    //--------Computing the distance from remote vertices to the local voxel----------------
    HPGV_TIMING_BEGIN(parallel_test::COMPUT_REMOTE_DIST);
    HPGV_TIMING_COUNT(parallel_test::COMPUT_REMOTE_DIST);
    int nTotalRemoteVertices = 0;
    vec3f coord(0.0f);
    int nBoundary=0;
    for (int i = 0; i < recv_remote_proc_count; i++)
    {
        svector<int> &remote_vertex_keys = recv_schedule[i].recv_buf;
        int vertex_key_count = remote_vertex_keys.size();
        nTotalRemoteVertices += vertex_key_count;
        for (int j=1 ; j < vertex_key_count ; j++)//j start from 1 in order to skip the first element.
        {
            int cur_vtx_key = remote_vertex_keys[j];
            compute_from_key_to_coords(cur_vtx_key, coord);
            //locate which local node does this remote vertex reside in.
            //compute the distance from the remote vertex to the features
            //in the local node's 3x range.
            distree_node_t* cur_node = get_distree_node(coord);
            float eucl_dist = FLOAT_MAX, norm_dist = FLOAT_MAX;
            if (cur_node)
            {
                distree_node_t* parent_node = cur_node;
                int generation = 0;
                while(parent_node->triple_feature_indices.empty()
                    &&parent_node->parent>=0)
                {
                    parent_node = &(m_node_array[parent_node->parent]);
                    generation ++;
                }
                compute_vertex_node_feature_distance(coord, *parent_node
                                        ,eucl_dist,norm_dist, generation);
            }else{
                nBoundary++;
                //use brutal force search
            }
            //the key entry now serves as holder for computed float distance.
            //and will be send back to where it came from.
            float* remote_distance = (float*)&(remote_vertex_keys[j]);
            *remote_distance = eucl_dist;
//             if (/*cur_vtx.key*/cur_vtx_key == 662663168){
//                 printf("---------PE%d----remote_dist=%f---\n"
//                     ,m_myid, eucl_dist);
//             }
        }
    }
    /*{
        D_P_MSG(m_myid, m_myid
            ,"++++++++++pass Compute remote distances^^^^^\n");
    }*/
//     sprintf(msg, "***PE%d, #boundary case=%d, out of %d, boundary rate=%f\n",m_myid,
//            nBoundary, nTotalRemoteVertices, (float)nBoundary/(float)nTotalRemoteVertices );
//     D_P_MSG(m_myid,m_myid, msg);
    HPGV_TIMING_END(parallel_test::COMPUT_REMOTE_DIST);
    //--------END of Computing the distance from remote vertices to the local voxel----------------
    ////-----------------------PHASE 2---------------------------------------------
    //    MPI_Barrier(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(parallel_test::COMM_PHASE_2);
    HPGV_TIMING_COUNT(parallel_test::COMM_PHASE_2);
    HPGV_TIMING_COUNT(parallel_test::MY_SEND_BYTE2);
    for (int i = 0; i < recv_remote_proc_count; i++)
    {
        MPI_Isend(data_ptr(recv_schedule[i].recv_buf),
                    recv_schedule[i].recv_buf.size(),
                    MPI_INT,//vertex_type,
                    recv_schedule[i].send_id,
                    MY_REPLY_MSG,
                    MPI_COMM_WORLD,
                    &(my_recv_request[i]));

        HPGV_TIMING_INCREASE(parallel_test::MY_SEND_BYTE2,
                   recv_schedule[i].recv_buf.size()*sizeof(int));
    }
    //Prepare to receive the updated vertices distance from other PEs that we
    //just asked for help to compute at the beginning of this function.
    HPGV_TIMING_COUNT(parallel_test::MY_RECV_BYTE2);
    if (send_count > 0)
    {
        //Reset reverse_lookup_table!!
        for (int i = 0; i < reverse_lookup_table.size(); i++) {
            reverse_lookup_table[i] = -1;
        }

        for (int i = 0; i < send_count; i++) {
            reverse_lookup_table[send_schedule[i].recv_id] = i;
        }
        int count=0, flag=0;
        while(count < send_count)
        {
            MPI_Iprobe(MPI_ANY_SOURCE, MY_REPLY_MSG, MPI_COMM_WORLD,
                        &flag, &status);
            if (flag)
            {
                int send_key_idx = reverse_lookup_table[status.MPI_SOURCE];
                D_ASSERT_P(m_myid, status.MPI_SOURCE == send_schedule[send_key_idx].recv_id,
                            "refine_global_distree(): **MPI_SOURCE!=send_id**\n", DERROR_ERR_MEM);
                int size=0;
                MPI_Get_count(&status, MPI_INT, &size );
                if (size > 0){
                    //send_schedule[send_key_idx].send_size = size;
                    //send_schedule[i].send_key_buf = (char *)calloc(size, sizeof(int) );
                    D_ASSERT_P(m_myid, send_schedule[send_key_idx].send_key_buf.size()==size,
                        "refine_global_distree(): The # of your receive remote distance!= the # of your send keys!\n",DERROR_ERR_MEM);

                    MPI_Irecv(data_ptr(send_schedule[send_key_idx].send_key_buf),
                                size,
                                MPI_INT,
                                status.MPI_SOURCE,
                                MY_REPLY_MSG,
                                MPI_COMM_WORLD,
                                &(my_send_request[send_key_idx]));
                    HPGV_TIMING_INCREASE(parallel_test::MY_RECV_BYTE2,
                        send_schedule[send_key_idx].send_key_buf.size()*sizeof(int));
                    count++;
                }
            }
        }
    }
    MPI_Waitall(send_count, data_ptr(my_send_request), MPI_STATUSES_IGNORE);
    MPI_Waitall(recv_remote_proc_count, data_ptr(my_recv_request), MPI_STATUSES_IGNORE);
    HPGV_TIMING_END(parallel_test::COMM_PHASE_2);
    /*{
        D_P_MSG(m_myid, m_myid
            ,"++++++++++pass Exchange reply result^^^^^\n");
    }*/
    HPGV_TIMING_BEGIN(parallel_test::UPDATE_DIST);
    HPGV_TIMING_COUNT(parallel_test::UPDATE_DIST);
    //Update local distance field using what we just received from other PEs.
    for (int i = 0; i < send_count ; i++){
        if (send_schedule[i].send_key_buf.size() > 1){
            float* remote_vertex_dist_array =
                (float*)data_ptr(send_schedule[i].send_key_buf);
            remote_vertex_dist_array++;//skip the first dummy element.
            int vertex_count = send_schedule[i].send_key_buf.size() - 1;
            for (int j = 0; j < vertex_count ; j++)
            {
                int cur_index = send_schedule[i].send_idx_buf[j];
                distree_vertex_t& cur_vtx = m_vertex_array[cur_index];
                float remote_distance = remote_vertex_dist_array[j];
                if (remote_distance < cur_vtx.eucl_dist)
                {
                    cur_vtx.eucl_dist = remote_distance;
                }
            }
        }
    }
//     int vtx_count = m_vertex_array.size();
//     for (int i = 0; i <  vtx_count; i++){
//         distree_vertex_t& cur_vtx = m_vertex_array[i];
//         if (cur_vtx.key == 662663168){
//             printf("---------PE%d-cur_vtx.localMinDist=%f----\n"
//                 ,m_myid, cur_vtx.eucl_dist);
//         }
//     }
    HPGV_TIMING_END(parallel_test::UPDATE_DIST);
}

void distree_t::compute_local_distance()
{
    int node_array_size = m_node_array.size();
    for (int i = 0; i < node_array_size ; i++)
    {  //scan node_array and locate leaf node
        distree_node_t &child_node = m_node_array[i];
        if (   child_node.isLeaf
            && child_node.location == distree_node_t::LOCAL
            //Not just local leaf, but also remote leaf node. Or else
            //remote vertex distance can not be computed in local PE
            )
        {
            distree_node_t &parent_node = m_node_array[child_node.parent];
            //compute distance field of 8 vertices of cur_node in LOCAL FEATURE BLOCK.
            //And update the respective min distance of cur_node->dist[8];
            //Two cases to deal with.(see ppt p5 for illustration)
            compute_child_vertex_distance(child_node, parent_node);
        }
    }
}

void distree_t::compute_child_vertex_distance( distree_node_t & child_node,
                                               distree_node_t & parent_node )
{
    //find the node whose 3x range has features.
    int generation = ITSELF;
    distree_node_t* feature_node = &child_node;
    while( feature_node->triple_feature_indices.size() <= 0//cur_node has no features
        && feature_node->parent >= 0 //cur_node  has parent
        && feature_node->level <= 0/*>0*/)
    {//go to parent node
        feature_node = &(m_node_array[feature_node->parent]);
        generation++;
    }

    for (int vtx_id = 0; vtx_id < 8 ; vtx_id++)
    {//for each of the eight vertices of child_node.
        float eucl_dist = FLOAT_MAX, min_eucl_dist = FLOAT_MAX;
        float norm_dist = FLOAT_MAX, min_norm_dist = FLOAT_MAX;
        //for each vertex, first compute max_norm
        //if the cur_node has no feature or its min max_norm to all features of cur_node
        // is partially outside of the cur_node's 3x BBox range, find nearest
        // features within parent_node's 3xBBox range. Else it is done.
        int vtx_idx = child_node.vertex_indices[vtx_id];
        if (!(vtx_idx >= 0 && vtx_idx < m_vertex_array.size())){
            D_ASSERT_P(m_myid, vtx_idx >= 0 && vtx_idx < m_vertex_array.size(),
                        "vertex index out of bounds", DERROR_ERR_MEM);
        }

        distree_vertex_t& cur_vertex = m_vertex_array[vtx_idx];
        compute_vertex_node_feature_distance(cur_vertex, *feature_node, eucl_dist, norm_dist, -1);

        if (cur_vertex.eucl_dist > eucl_dist)
        {
            cur_vertex.eucl_dist = eucl_dist;
            cur_vertex.min_dist_type = generation;
        }
//         else{
//             char msg_err[256];
//             sprintf(msg_err, "^^^^---lastest eucl_dist=%f-vtx_idx=%d--^^^^^"
//                 ,eucl_dist,vtx_idx);
//             D_P_MSG(m_myid, m_myid,msg_err);
//             char* p=NULL;
//             *p ='0';
//         }
        if (cur_vertex.maxnorm_dist > norm_dist)
        {
            cur_vertex.maxnorm_dist = norm_dist;
        }

        int level = cur_vertex.min_dist_type;
        int child_local_id = compute_local_child_key(child_node);
        while ( eucl_dist > 2*feature_node->m_bound.SphericalRadius()//need to search parent's triple feature?
             && feature_node->parent >= 0 //has parent node?
             && level++ < 3)//not beyond grandparent node?
        {//If all "YES", then (3x,5x) range of the cur_node may have closer features. Check it out!
            //Search in 3x BBox range of parent_node's features.
            distree_node_t &parent = m_node_array[feature_node->parent];
            //The child_local_id here provides a hint on how to further compute
            //the distance from cur_vertex to the parent's features without revisiting
            //the features that are in child_node.
            compute_vertex_node_feature_distance(cur_vertex, parent, eucl_dist,
                                                 norm_dist, child_local_id);
            if (cur_vertex.eucl_dist > eucl_dist)
            {
                cur_vertex.eucl_dist = eucl_dist;
                cur_vertex.min_dist_type = generation + 1;
            }
            if (cur_vertex.maxnorm_dist > norm_dist)
            {
                cur_vertex.maxnorm_dist = norm_dist;
            }
        }
//         if (cur_vertex.min_dist_type < 0 || cur_vertex.min_dist_type>=3){
//             char msg_err[256];
//             sprintf(msg_err, "###---min_dist_type=%d-vtx_idx=%d--###"
//                 ,cur_vertex.min_dist_type,vtx_idx);
//             D_P_MSG(m_myid, m_myid,msg_err);
//             D_ASSERT_P(m_myid, cur_vertex.min_dist_type < 3 && cur_vertex.min_dist_type>=0,
//                 "min_dist_type out of bound\n",DERROR_ERR_MEM);
//         }
    }
}
void distree_t::compute_vertex_node_feature_distance( vec3f& coords,
                                                      distree_node_t& feature_node,
                                                      float& eucl_dist,
                                                      float& norm_dist, int compute_type/*=-1*/ )
{
    svector<int> *feature_index_array;
    feature_index_array = &(feature_node.triple_feature_indices);
    /***not to use complementary_feature array, it's not working rightnow.
    if (compute_type == -1){
        feature_index_array = &(feature_node.triple_feature_indices);
    }else{
        feature_index_array = &(feature_node.complementary_features[compute_type]);
        if (feature_index_array->empty()){
            eucl_dist = FLOAT_MAX;
            norm_dist = FLOAT_MAX;
            return;
        }
    }
    */
    //iterate features index array specified by comp_feature_id and find minimum distance
    float min_edist = FLOAT_MAX;
    for (int i = 0; i < feature_index_array->size() ; i++){
        D_ASSERT_P(m_myid, (*feature_index_array)[i] < m_features->size_feature(),
            "Memory overflow", DERROR_ERR_MEM);

        float edist = m_features->compute_vertex_feature_distance(coords,
                                       (*feature_index_array)[i] , *this);

        if(min_edist > edist) {
            min_edist = edist;
        }
    }
    eucl_dist = min_edist;
}
void distree_t::compute_vertex_node_feature_distance( distree_vertex_t& cur_vertex,
                                                      distree_node_t& feature_node,
                                                      float& eucl_dist,
                                                      float& norm_dist, int compute_type/*=-1*/ )
{
    svector<int> *feature_index_array;
    feature_index_array = &(feature_node.triple_feature_indices);
    /***not to use complementary_feature array, it's not working rightnow.
    if (compute_type == -1){
        feature_index_array = &(feature_node.triple_feature_indices);
    }else{
        feature_index_array = &(feature_node.complementary_features[compute_type]);
        if (feature_index_array->empty()){
            eucl_dist = FLOAT_MAX;
            norm_dist = FLOAT_MAX;
            return;
        }
    }
    */
    //iterate features index array specified by comp_feature_id and find minimum distance
    float min_edist = FLOAT_MAX;
    for (int i = 0; i < feature_index_array->size() ; i++){
        D_ASSERT_P(m_myid, (*feature_index_array)[i] < m_features->size_feature(),
                    "Memory overflow", DERROR_ERR_MEM);

        float edist = m_features->compute_vertex_feature_distance(cur_vertex,
                                            (*feature_index_array)[i], *this);
        if(min_edist > edist) {
            min_edist = edist;
        }
    }
    eucl_dist = min_edist;
}

void distree_t::create_send_schedule_vertex( svector<send_schedule_t>& send_schedule )
{
    svector<int> reverse_lookup_table(m_groupsize,-1);
    const int send_count = send_schedule.size();
    for (int i = 0; i < send_count ; i++)
    {//the initial size of the send_buf is at most as large as the size of the vertex array.
        send_schedule[i].recv_id = m_sent_table[i];
        send_schedule[i].send_idx_buf.reserve(m_vertex_array.size());
        reverse_lookup_table[send_schedule[i].recv_id] = i;
    }

    const int node_size = m_node_array.size();

    for (int i = 0; i < node_size; i++)
    {//The selected vertex v must belongs to a leaf node whose
     //bounding box overlaps the assigned local region.
        distree_node_t& cur_node = m_node_array[i];
        if (cur_node.isLeaf && cur_node.location == distree_node_t::LOCAL)
        {
            for (int j = 0; j < 8 ; j++)
            {
                int cur_vtx_idx = cur_node.vertex_indices[j];
                distree_vertex_t& cur_vertex = m_vertex_array[cur_vtx_idx];
                if (cur_vertex.min_dist_type >= 3 || cur_vertex.min_dist_type<-1){
//                     char* p=NULL;
//                     *p = 'a';
                    char msg_err[256];
                    sprintf(msg_err, "+++---min_dist_type=%d-vtx_idx=%d--+++",cur_vertex.min_dist_type
                        ,cur_vtx_idx);
                    D_P_MSG(m_myid, m_myid,msg_err);
                    D_ASSERT_P(m_myid, cur_vertex.min_dist_type < 3 && cur_vertex.min_dist_type>=0,
                        "min_dist_type out of bound\n",DERROR_ERR_MEM);
                }

                distree_node_t* ancestor_node = &cur_node;
                int level = cur_vertex.min_dist_type;
                //The v:min dist node denotes the node where the local
                //minimum vertex distance has been computed. If the
                //node.processors list is nonempty, the v:key should be send
                //to all the neighboring processors whose ID appears in
                //min_dist_node.processors list.
                while(level>0)//level>0
                {
                    ancestor_node = &(m_node_array[ancestor_node->parent]);
                    --level;
                }
                const int proc_count = ancestor_node->processor_3x_indices.size();
                for (int k = 0; k < proc_count; k++)
                {
                    int send_schedule_index =
                        reverse_lookup_table[ancestor_node->processor_3x_indices[k]];
                    if (send_schedule_index > -1)
                    {
                        send_schedule_t& cur_send_schedule =
                                         send_schedule[send_schedule_index];

                        cur_send_schedule.send_idx_buf.push_back(cur_vtx_idx);
                    }
                }
            }// for each vertex of leaf node
        }
    }//for each leaf node

    for (int i = 0; i < send_count ; i++)
    {//shrink to fit
        //svector<int>(send_schedule[i].send_idx_buf).swap(send_schedule[i].send_idx_buf);
        send_schedule[i].send_idx_buf.shrink_to_fit();
    }
    //remove duplicates
    for (int i = 0; i < send_count ; i++)
    {
        svector<int> &cur_send_idx_buf = send_schedule[i].send_idx_buf;
        //sort(cur_send_idx_buf.begin(), cur_send_idx_buf.end());
        qsort(cur_send_idx_buf.data_ptr(),
              cur_send_idx_buf.size(),
              sizeof(int),
              int_compare);

        svector<int> unique_idx_array;
        unique_idx_array.reserve(cur_send_idx_buf.size());

        if (cur_send_idx_buf.size()>0){
            unique_idx_array.push_back(cur_send_idx_buf[0]);
        }

        for (int j = 1; j < cur_send_idx_buf.size(); j++){
            if (unique_idx_array[unique_idx_array.size()-1] != cur_send_idx_buf[j]){
                unique_idx_array.push_back(cur_send_idx_buf[j]);
            }
        }
        //save the send vertices's key to send_schedule[i].send_key_buf.
        //it is the send_schedule[i].send_key_buf that will be send to remote PE.
        int unique_count = unique_idx_array.size();
        svector<uint32_t> &cur_send_key_buf = send_schedule[i].send_key_buf;

        cur_send_key_buf.resize(unique_count+1);
        cur_send_key_buf[0] = unique_count;//The first element is used to count the # of keys.
        char msg[256];
        /*sprintf(msg, "Proc%d: %d vertice out of %d---->PE%d\n",
            m_myid, unique_count, m_vertex_array.size(), send_schedule[i].recv_id);
        D_P_MSG(m_myid, 1, msg);*/
        for (int j = 0; j < unique_count ; j++){
            //printf("===PE%d vtx[%d] send====\n",m_myid, j);
            m_vertex_array[unique_idx_array[j]].is_send = 1;//For debug: tag the vertices being sent
            cur_send_key_buf[j+1] = m_vertex_array[unique_idx_array[j]].key;
            //unique_idx_array[j] = vertex_array[unique_idx_array[j]].key;
        }
        send_schedule[i].send_idx_buf.swap(unique_idx_array);
    }
}
int compute_key_0(float x, float y, float z, const vec3f &center){
    int key;
    key =   (( z > center.z()) << 2) +
        (( y > center.y()) << 1) +
        (( x > center.x()) << 0);

    return key;
}

int compute_key_1(float x, float y, float z, const vec3f &center){
    int key;
    key= (( z >  center.z()) << 2) +
        (( y >  center.y()) << 1) +
        (( x >= center.x()) << 0);

    return key;
}

int compute_key_2(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >  center.z()) << 2) +
        (( y >= center.y()) << 1) +
        (( x >  center.x()) << 0);

    return key;
}

int compute_key_3(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >  center.z()) << 2) +
        (( y >= center.y()) << 1) +
        (( x >= center.x()) << 0);

    return key;
}

int compute_key_4(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >= center.z()) << 2) +
        (( y >  center.y()) << 1) +
        (( x >  center.x()) << 0);

    return key;
}

int compute_key_5(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >= center.z()) << 2) +
        (( y >  center.y()) << 1) +
        (( x >= center.x()) << 0);

    return key;
}

int compute_key_6(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >= center.z()) << 2) +
        (( y >= center.y()) << 1) +
        (( x >  center.x()) << 0);

    return key;
}

int compute_key_7(float x, float y, float z, const vec3f &center){
    int key;
    key = (( z >= center.z()) << 2) +
          (( y >= center.y()) << 1) +
          (( x >= center.x()) << 0);

    return key;
}
typedef int   (*compute_octree_key_fun_t)(float x, float y, float z, const vec3f &center);

compute_octree_key_fun_t g_key_fun_tables[8]={
    compute_key_0, compute_key_7, compute_key_2, compute_key_3,
    compute_key_4, compute_key_5, compute_key_6, compute_key_1
};

inline distree_node_t* distree_t::get_distree_node(const vec3f& coord )
{
    distree_node_t *node = &(m_node_array[root_index]);
    while( node
        //&& node->location == distree_node_t::LOCAL
        && !node->isLeaf)
    {//iterative refining node and keep searching
        //compute the center of the node's bounding box.
        vec3f center = node->m_bound.Center();
        int localkey = (( coord.z() > center.z()) << 2) +
                       (( coord.y() > center.y()) << 1) +
                       (( coord.x() > center.x()) << 0);

        node = get_node_child(*node, localkey);
    }
    return node;
}
distree_node_t* distree_t::get_distree_local_node(const vec3f& coord )
{
    distree_node_t *node = &(m_node_array[root_index]);
    while( node
        && node->location == distree_node_t::LOCAL
        && !node->isLeaf)
    {//iterative refining node and keep searching
        //compute the center of the node's bounding box.
        vec3f center = node->m_bound.Center();
//         int localkey = (( coord.z() > center.z()) << 2) +
//             (( coord.y() > center.y()) << 1) +
//             (( coord.x() > center.x()) << 0);
        int i = 0;
        distree_node_t* tmpnode = node;
        int localkey;
        do{//make sure we choose the LOCAL leaf node.
           //In boundary case, get_leaf_node() would give you a REMOTE leaf node.
            //int j = i;
            localkey = g_key_fun_tables[i](coord.x(),coord.y(),coord.z(), center);
             tmpnode = get_node_child(*node, localkey);
            i++;
        } while (i < 8 && tmpnode->location != distree_node_t::LOCAL);
        node = tmpnode;
    }
    return node;
}

distree_node_t* distree_t::get_node_child(const distree_node_t &parent_node, int childid)
{
    int child_index = parent_node.m_children[childid] ;
    if (child_index == -1) {
        //parent_node is leaf node. no more children.
        return NULL;
    }
    return &(m_node_array[child_index]);
}

void distree_t::create_distance_volume()
{
    int max_assigned_local_leaf_num = get_max_local_region_num();

    int  i = 0;
    for (i = 0; i < m_assigned_local_leaf_indices.size() ; i++)
    {
        distree_node_t& cur_region = m_node_array[m_assigned_local_leaf_indices[i]];
        geometry::BBox cur_region_bound  = cur_region.m_bound;
        //cur_region_size should be integer value, but floating point computation would create a tiny error,
        // so we need to cast the value returned by getDimension() to vec3i and then back to vec3f.
        vec3f          cur_region_size   = vec3f(vec3i(cur_region_bound.getDimension()));
        vec3i          cur_distvol_bound = vec3i(cur_region_size/sample_size + 0.5f);

        volume_t       *dist_volume      = new volume_t(cur_distvol_bound);
        dist_volume->m_bound.pMin
            = vec3i((cur_region_bound.pMin - m_global_domain_bound.pMin)/sample_size + 0.5f);
        dist_volume->m_bound.pMax
            = vec3i((cur_region_bound.pMax - m_global_domain_bound.pMin)/sample_size + 0.5f);
        vec3f& offset = cur_region_bound.pMin;
        //compute distance value voxel by voxel
        float maxZ = cur_region_size.z();
        float maxY = cur_region_size.y();
        float maxX = cur_region_size.x();
        for (float z = 0; z < maxZ; z+=sample_size.z()) {
            for (float y = 0; y < maxY; y+=sample_size.y()) {
                for (float x = 0; x < maxX; x+=sample_size.x()) {
//                     HPGV_TIMING_BEGIN(COMPUTE_DIST_VOL_STEP_TIME);
//                     HPGV_TIMING_COUNT(COMPUTE_DIST_VOL_STEP_TIME);
                    float gx = x + offset.x();
                    float gy = y + offset.y();
                    float gz = z + offset.z();
                    float v  = get_tree_value(gx, gy, gz);
                    dist_volume->set_volume_value(
                                    x/sample_size.x(),
                                    y/sample_size.y(),
                                    z/sample_size.z(), v);
//                    HPGV_TIMING_END(COMPUTE_DIST_VOL_STEP_TIME);
                }
            }
        }// end of loop through local distance volume voxel.
        m_distance_volumes.push_back(dist_volume);
    }// end of for each assigned local leaf node.
    //create dummy distance volume if
    // max_assigned_local_leaf_num > m_assigned_local_leaf_indices.size()
    // in order to prevent the subsequent collective IO write from getting stuck.
    /*printf("****PE%d max_assigned_local_leaf_num=%d***\n"
        ,m_myid, max_assigned_local_leaf_num);*/

    for (   ; i < max_assigned_local_leaf_num ; i++){
        volume_t *dist_volume = new volume_t(1,1,1);
        dist_volume->m_bound = geometry::BBox(vec3f(0.0f),vec3f(1.0f));
        m_distance_volumes.push_back(dist_volume);
    }
}

int distree_t::get_max_local_region_num() const
{
    int my_num = m_assigned_local_leaf_indices.size();
    int max_num = 0;
    MPI_Allreduce(&my_num, &max_num, 1, MPI_INT, MPI_MAX, m_communicator);
    return max_num;
}

float distree_t::get_tree_value(float x, float y, float z)
{
    const vec3f pos(x,y,z);
    const distree_node_t *node = get_distree_local_node(pos);
    const int *cur_vertex_indices = node->vertex_indices;
    float vtx_dist_value[8];
    if (node != NULL ){
        //trilinearly interpolate the distance value on the 8 vertices.
        //compute the offset[0,1] of pos relative to the bbox of node.
        vec3f txyz = node->m_bound.Offset(pos);
        for (int i = 0; i < 8 ; i++){
            vtx_dist_value[i] = m_vertex_array[cur_vertex_indices[i]].eucl_dist;
        }
    //linear interpolation along x axis
        float sx0 = vtx_dist_value[0]*(1-txyz.x()) + vtx_dist_value[1]*txyz.x();
        float sx1 = vtx_dist_value[2]*(1-txyz.x()) + vtx_dist_value[3]*txyz.x();
        float sx2 = vtx_dist_value[4]*(1-txyz.x()) + vtx_dist_value[5]*txyz.x();
        float sx3 = vtx_dist_value[6]*(1-txyz.x()) + vtx_dist_value[7]*txyz.x();
    //linear interpolation along y axis
        float sy0 = sx0 * (1-txyz.y()) + sx1 * txyz.y();
        float sy1 = sx2 * (1-txyz.y()) + sx3 * txyz.y();
    //linear interpolation along z axis
        return sy0 * (1-txyz.z()) + sy1 * txyz.z();//fabsf(sz0);
    }else {
        //boundary case, simply scan the vertex array and find the nearest vertex.
        //and return the vertex distance.
        return 0.0f;
    }
}

void distree_t::save_distance_volume_parallel(const char* file_name )
{
    MPI_File fh;
    int ret = MPI_File_open(m_communicator,
        const_cast<char*>(file_name),
        MPI_MODE_CREATE | MPI_MODE_WRONLY,
        MPI_INFO_NULL,
        &fh);
    D_ASSERT_P(m_myid,ret == MPI_SUCCESS, "file open failed!\n", DERROR_ERR_IO);

    MPI_Datatype filetype;
    /*printf("**** dist_volume size= %d, assign_local_leaf size=%d***\n"
        ,m_distance_volumes.size(), m_assigned_local_leaf_indices.size());*/
    for (int i = 0; i < m_distance_volumes.size() ; i++)
    {//save each local distance volumes.
        volume_t* cur_dist_volume = m_distance_volumes[i];
        vec3i gsizes = vec3i(vec3f(m_global_domain_size)/sample_size);//m_global_domain_size;
        vec3i psizes = m_np_xyz;
        vec3i lsizes = cur_dist_volume->d_size;
        vec3i start_indices = vec3i(cur_dist_volume->m_bound.pMin);
        if (gsizes.x() <= 0 || gsizes.y() <= 0 || gsizes.z() <= 0
            || psizes.x() <= 0 || psizes.y() <= 0 || psizes.z() <=0
            || lsizes.x() <= 0 || lsizes.y() <= 0 || lsizes.z() <= 0
            || start_indices.x() < 0 || start_indices.y()<0 || start_indices.z()<0){
                cout<<"-----PE"<<m_myid<<"----gsize="<<gsizes<<endl;
                cout<<"-----PE"<<m_myid<<"----psize="<<psizes<<endl;
                cout<<"-----PE"<<m_myid<<"----lsize="<<lsizes<<endl;
                cout<<"-----PE"<<m_myid<<"--dist_volume["<<i<<"]--"
                    <<m_assigned_local_leaf_indices.size()<<"/"
                    <<"---start_indices="<<start_indices<<endl;
        }

        MPI_Type_create_subarray(3, (int*)gsizes, (int*)lsizes, (int*)start_indices,
                                 MPI_ORDER_FORTRAN, MPI_FLOAT, &filetype);
        MPI_Type_commit(&filetype);

        MPI_File_set_view(fh, 0, MPI_FLOAT, filetype, "native", MPI_INFO_NULL);

        int local_array_size = lsizes[0] * lsizes[1] * lsizes[2];
        float* buffer = data_ptr(cur_dist_volume->m_data);

        MPI_File_write_all(fh, buffer, local_array_size, MPI_FLOAT, MPI_STATUS_IGNORE);
    }

    MPI_File_close(&fh);
}

void distree_t::verify_vertex_distance()
{
    float min_eucl = D_FLOAT_INFINITY, distance;
    int   min_voxel = 0;

    //voxel_t *voxel;
    //voxel_t *voxel_array = (voxel_t *)(distree->feature_array);

    int vertex_array_size = m_vertex_array.size();
    int feature_array_size = m_features->size_feature();
    volume_t* pVolumeFeature = dynamic_cast<volume_t*>(m_features);
    svector<voxel_t> &voxel_array = pVolumeFeature->m_isovoxels;

    for (int vertex_id = 0; vertex_id < vertex_array_size; vertex_id++)
    {
        min_eucl = D_FLOAT_INFINITY;
        distree_vertex_t* vertex = &(m_vertex_array[vertex_id]);

        for (int voxel_id = 0; voxel_id < feature_array_size; voxel_id++)
        {
            voxel_t* voxel = &(voxel_array[voxel_id]);

            distance = (voxel->coords-vertex->coords).length();
            //compute_dist_eucl(&(voxel->coords), &(vertex->coords));
            if (min_eucl > distance) {
                min_eucl = distance;
                min_voxel = voxel_id;
            }
        }

        if ( fabs(vertex->eucl_dist - min_eucl ) > FLOAT_EPSILON) {
            fprintf(stderr, "PE%d: vertex %d (%f, %f, %f) computed %f, \
                            real_min_dist=%f from voxel%d (%f, %f, %f)\n",
                            m_myid,
                            vertex_id,
                            vertex->coords.x(),
                            vertex->coords.y(),
                            vertex->coords.z(),
                            vertex->eucl_dist,
                            min_eucl,
                            min_voxel,
                            voxel_array[min_voxel].coords.x(),
                            voxel_array[min_voxel].coords.y(),
                            voxel_array[min_voxel].coords.z());
        }
    }
}