/**
 * hpgv_utiltiming.c
 *
 * Copyright (c) Hongfeng Yu
 *
 * Contact:
 * Hongfeng Yu
 * hfstudio@gmail.com
 *
 *
 * All rights reserved.  May not be used, modified, or copied
 * without permission.
 *
 */

#include "hpgv_utiltiming.h"
#include "DError.h"
#include <string.h>
#include <math.h>

#define LOCAL_TIMING_MSG    1000
#define LOCAL_COUNT_MSG     1001
#define MAXLINE             1024
#define GETTIME   MPI_Wtime()
#define SQR(x) ((x) * (x))
/**
 * timing_t
 *
 */
typedef struct timing_t {
    /* mpi */
    int mpiid, groupsize, root;
    MPI_Comm mpicomm;
    int countroot;

    /* data */
    double timing[HPGV_MAX_TIMING_UNIT];
    int count[HPGV_MAX_TIMING_UNIT];
    char name[HPGV_MAX_TIMING_UNIT][MAXLINE];

    /* statistics */
    double **t_timing;
    int    **t_count;
    double *t_proc[HPGV_MAX_TIMING_UNIT];
    double sum[HPGV_MAX_TIMING_UNIT];
    double min[HPGV_MAX_TIMING_UNIT];
    double max[HPGV_MAX_TIMING_UNIT];
    double mean[HPGV_MAX_TIMING_UNIT];
    double stddev[HPGV_MAX_TIMING_UNIT];

    /* io */
    int savelocal;
    int showlocal;
    int saveglobal;
    int showglobal;
    char contextlocal[MAXLINE];
    char contextglobal[MAXLINE];
} timing_t;

timing_t *theTiming = NULL;

/**
 * hpgv_timing_contextglobal
 *
 */
void
hpgv_timing_contextglobal(char *context)
{
#ifdef HPGV_TIMING_DEBUG
    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    strcpy(theTiming->contextglobal, context);
}

/**
 * hpgv_timing_contextlocal
 *
 */
void
hpgv_timing_contextlocal(char *context)
{
#ifdef HPGV_TIMING_DEBUG
    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    strcpy(theTiming->contextlocal, context);
}

/**
 * hpgv_timing_showlocal
 *
 */
void
hpgv_timing_showlocal(int local)
{
#ifdef HPGV_TIMING_DEBUG
    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->showlocal = local;
}

/**
 * hpgv_timing_showglobal
 *
 */
void
hpgv_timing_showglobal(int global)
{
#ifdef HPGV_TIMING_DEBUG
    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->showglobal = global;
}

/**
 * hpgv_timing_savelocal
 *
 */
void
hpgv_timing_savelocal(int local)
{
#ifdef HPGV_TIMING_DEBUG
    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->savelocal = local;
}

/**
 * hpgv_timing_saveglobal
 *
 */
void
hpgv_timing_saveglobal(int global)
{
#ifdef HPGV_TIMING_DEBUG
    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->saveglobal = global;
}

/**
 * hpgv_timing_countroot
 *
 */
void
hpgv_timing_countroot(int countroot)
{
#ifdef HPGV_TIMING_DEBUG
    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->countroot = countroot;
}

/**
 * hpgv_timing_clear
 *
 */
void
hpgv_timing_clear()
{
    int i;
    if (theTiming != NULL) {
        if (theTiming->t_timing != NULL) {
            for (i = 0; i < theTiming->groupsize; i++) {
                if (theTiming->t_timing[i] != NULL) {
                    free(theTiming->t_timing[i]);
                }
            }
            free(theTiming->t_timing);
        }

        if (theTiming->t_count != NULL) {
            for (i = 0; i < theTiming->groupsize; i++) {
                if (theTiming->t_count[i] != NULL) {
                    free(theTiming->t_count[i]);
                }
            }
            free(theTiming->t_count);
        }

        for (i = 0; i < HPGV_MAX_TIMING_UNIT; i++) {
            if (theTiming->t_proc[i] != NULL) {
                free(theTiming->t_proc[i]);
            }
        }

        free(theTiming);
        theTiming = NULL;
    }
}

/**
 * hpgv_timing_finalize
 *
 */
void
hpgv_timing_finalize()
{
    hpgv_timing_clear();
}

/**
 * hpgv_timing_init
 *
 */
void
hpgv_timing_init(int root, MPI_Comm mpicomm)
{
    if (theTiming) {
        free(theTiming);
    }

    theTiming = (timing_t *)calloc(1, sizeof(timing_t));
    D_ASSERT_P(0, theTiming, "Out of memory.", DERROR_ERR_MEM);

    theTiming->root = root;
    theTiming->mpicomm = mpicomm;

    MPI_Comm_rank(mpicomm, &(theTiming->mpiid));
    MPI_Comm_size(mpicomm, &(theTiming->groupsize));

    D_P_MSG(theTiming->mpiid, root,
               "\n\n========== Start timing module. ========== \n");
}

/**
 * hpgv_timing_valid
 *
 */
int
hpgv_timing_valid()
{
    if (!theTiming) {
        return DFALSE;
    }

    return DTRUE;
}

/**
 * hpgv_timing_name
 *
 */
void
hpgv_timing_name(int unit,const char *name)
{
    if (unit >= HPGV_MAX_TIMING_UNIT) {
        D_ERR_MSG_P(0, "Exceed maximum timing unit number.");
        return;
    }

    if (!theTiming) {
        D_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }

    sprintf(theTiming->name[unit], "%s", name);
}

/**
 * hpgv_timing_decrease
 *
 */
void
hpgv_timing_decrease(int unit, double value)
{
#ifdef HPGV_TIMING_DEBUG
    if (unit >= HPGV_MAX_TIMING_UNIT) {
        HPGV_ERR_MSG_P(0, "Exceed maximum timing unit number.");
        return;
    }

    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->timing[unit] -= value;
}

/**
 * hpgv_timing_increase
 *
 */
void
hpgv_timing_increase(int unit, double value)
{
#ifdef HPGV_TIMING_DEBUG
    if (unit >= HPGV_MAX_TIMING_UNIT) {
        HPGV_ERR_MSG_P(0, "Exceed maximum timing unit number.");
        return;
    }

    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->timing[unit] += value;
}

/**
 * hpgv_timing_set
 *
 */
void
hpgv_timing_set(int unit, double value)
{
#ifdef HPGV_TIMING_DEBUG
    if (unit >= HPGV_MAX_TIMING_UNIT) {
        HPGV_ERR_MSG_P(0, "Exceed maximum timing unit number.");
        return;
    }

    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->timing[unit] = value;
}

/**
 * hpgv_timing_get
 *
 */
double
hpgv_timing_get(int unit)
{
#ifdef HPGV_TIMING_DEBUG
    if (unit >= HPGV_MAX_TIMING_UNIT) {
        HPGV_ERR_MSG_P(0, "Exceed maximum timing unit number.");
        return;
    }

    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    return theTiming->timing[unit];
}

/**
 * hpgv_timing_getgather
 *
 */
double
hpgv_timing_getgather(int proc, int unit)
{
#ifdef HPGV_TIMING_DEBUG
    if (unit >= HPGV_MAX_TIMING_UNIT) {
        HPGV_ERR_MSG_P(0, "Exceed maximum timing unit number.");
        return;
    }

    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    return theTiming->t_proc[unit][proc];
}

/**
 * hpgv_timing_begin
 *
 */
void
hpgv_timing_begin(int unit)
{
#ifdef HPGV_TIMING_DEBUG
    if (unit >= HPGV_MAX_TIMING_UNIT) {
        HPGV_ERR_MSG_P(0, "Exceed maximum timing unit number.");
        return;
    }

    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->timing[unit] -= GETTIME;
}

/**
 * hpgv_timing_end
 *
 */
void
hpgv_timing_end(int unit)
{
#ifdef HPGV_TIMING_DEBUG
    if (unit >= HPGV_MAX_TIMING_UNIT) {
        HPGV_ERR_MSG_P(0, "Exceed maximum timing unit number.");
        return;
    }

    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->timing[unit] += GETTIME;
}

/**
 * hpgv_timing_count
 *
 */
void
hpgv_timing_count(int unit)
{
#ifdef HPGV_TIMING_DEBUG
    if (unit >= HPGV_MAX_TIMING_UNIT) {
        HPGV_ERR_MSG_P(0, "Exceed maximum timing unit number.");
        return;
    }

    if (!theTiming) {
        HPGV_ERR_MSG_P(0, "Timing modelue has not been intialized.");
        return;
    }
#endif

    theTiming->count[unit]++;
}

/**
 * hpgv_timing_gather
 *
 */
void
hpgv_timing_gather()
{
    int mpiid = theTiming->mpiid;
    int groupsize = theTiming->groupsize;
    int root = theTiming->root;
    MPI_Comm mpicomm = theTiming->mpicomm;
    int i;
    MPI_Status status;

    if (mpiid == root) {
        /* new memory */
        theTiming->t_timing
            = (double **)calloc(groupsize, sizeof(double *));
        theTiming->t_count
            = (int **)calloc(groupsize, sizeof(int *));

        D_ASSERT_P(mpiid, theTiming->t_timing && theTiming->t_count,
                    "Out of memory.", DERROR_ERR_MEM);

        for (i = 0; i < groupsize; i++) {
            theTiming->t_timing[i]
                = (double *)calloc(HPGV_MAX_TIMING_UNIT, sizeof(double));

            theTiming->t_count[i]
                = (int *)calloc(HPGV_MAX_TIMING_UNIT, sizeof(int));

            D_ASSERT_P(mpiid, theTiming->t_timing[i] && theTiming->t_count[i],
                        "Out of memory.", DERROR_ERR_MEM);
        }

        for (i = 0; i < HPGV_MAX_TIMING_UNIT; i++ ) {
            theTiming->t_proc[i]
                = (double *)calloc(groupsize, sizeof(double));

            D_ASSERT_P(mpiid, theTiming->t_proc[i],
                        "Out of memory.", DERROR_ERR_MEM);
        }

        for (i = 0; i < groupsize; i++) {
            if (i != root) {
                /* receive from others */
                MPI_Recv(theTiming->t_timing[i],
                         HPGV_MAX_TIMING_UNIT * sizeof(double),
                         MPI_BYTE,
                         i,
                         LOCAL_TIMING_MSG,
                         mpicomm,
                         &status);

                MPI_Recv(theTiming->t_count[i],
                         HPGV_MAX_TIMING_UNIT * sizeof(int),
                         MPI_BYTE,
                         i,
                         LOCAL_COUNT_MSG,
                         mpicomm,
                         &status);
            } else {
                /* copy myself */
                memcpy(theTiming->t_timing[i], theTiming->timing,
                       HPGV_MAX_TIMING_UNIT * sizeof(double));

                memcpy(theTiming->t_count[i], theTiming->count,
                       HPGV_MAX_TIMING_UNIT * sizeof(int));
            }
        }
    } else {
        /* send to root*/
        MPI_Send(theTiming->timing,
                 HPGV_MAX_TIMING_UNIT * sizeof(double),
                 MPI_BYTE,
                 root,
                 LOCAL_TIMING_MSG,
                 mpicomm);

        MPI_Send(theTiming->count,
                 HPGV_MAX_TIMING_UNIT * sizeof(int),
                 MPI_BYTE,
                 root,
                 LOCAL_COUNT_MSG,
                 mpicomm);
    }
}

/**
 * hpgv_timing_save_local
 *
 */
void
hpgv_timing_save_local()
{
    int mpiid = theTiming->mpiid;
    int groupsize = theTiming->groupsize;
    int countroot = theTiming->countroot;
    int root = theTiming->root;
    int totalproc = groupsize;
    int unit, proc;
    char filename[MAXLINE];
    FILE *fp = NULL;

    if (!countroot) {
        totalproc--;
    }

    sprintf(filename, "%s_local.csv", theTiming->contextlocal);
    fp = fopen(filename, "w");
    D_ASSERT_P(mpiid, fp, "Can not open file.", DERROR_ERR_IO);

    fprintf(stderr, "\nproc %d save %s\n", mpiid, filename);

    fprintf(fp, "%s", "PE");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(fp, ",%s", theTiming->name[unit]);
        }
    }
    fprintf(fp, "\n");

    for (proc = 0; proc < groupsize; proc++) {
        if (proc == root && !countroot) {
            continue;
        }

        fprintf(fp, "%d", proc);

        for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
            if (strcmp(theTiming->name[unit], "") != 0) {
                fprintf(fp, ",%f", theTiming->t_proc[unit][proc]);
            }
        }
        fprintf(fp, "\n");
    }

    fclose(fp);
}

/**
 * hpgv_timing_show_local
 *
 */
void
hpgv_timing_show_local()
{
    int groupsize = theTiming->groupsize;
    int countroot = theTiming->countroot;
    int root = theTiming->root;
    int totalproc = groupsize;
    int unit, proc;

    if (!countroot) {
        totalproc--;
    }

    fprintf(stderr, "\nlocal performance data for %s\n",
            theTiming->contextlocal);

    fprintf(stderr, "-------");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(stderr, "-----------");
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "%-6s|", "PE");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(stderr, "%-10s|", theTiming->name[unit]);
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "-------");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(stderr, "-----------");
        }
    }
    fprintf(stderr, "\n");

    for (proc = 0; proc < groupsize; proc++) {
        if (proc == root && !countroot) {
            continue;
        }

        fprintf(stderr, "%-6d|", proc);

        for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
            if (strcmp(theTiming->name[unit], "") != 0) {
                if (strchr(theTiming->name[unit], 'C')){
                    fprintf(stderr, "%-10.0f|", theTiming->t_proc[unit][proc]);
                } else {
                    fprintf(stderr, "%-10.4f|", theTiming->t_proc[unit][proc]);
                }
            }
        }
        fprintf(stderr, "\n");
    }

    fprintf(stderr, "-------");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(stderr, "-----------");
        }
    }
    fprintf(stderr, "\n");
}

/**
 * hpgv_timing_save_global
 *
 */
void
hpgv_timing_save_global()
{
    int mpiid = theTiming->mpiid;
    int groupsize = theTiming->groupsize;
    int countroot = theTiming->countroot;
    int totalproc = groupsize;
    int unit;

    char filename[MAXLINE];
    FILE *fp = NULL;

    if (!countroot) {
        totalproc--;
    }

    sprintf(filename, "%s_global.csv", theTiming->contextglobal);

    fp = fopen(filename, "w");
    D_ASSERT_P(mpiid, fp, "Can not open file.", DERROR_ERR_IO);

    fprintf(stderr, "\nproc %d save %s\n", mpiid, filename);

    fprintf(fp, "%s", " ");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(fp, ",%s", theTiming->name[unit]);
        }
    }
    fprintf(fp, "\n");

    fprintf(fp, "%s", "sum");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(fp, ",%f", theTiming->sum[unit]);
        }
    }
    fprintf(fp, "\n");

    fprintf(fp, "%s", "min");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(fp, ",%f", theTiming->min[unit]);
        }
    }
    fprintf(fp, "\n");

    fprintf(fp, "%s", "max");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(fp, ",%f", theTiming->max[unit]);
        }
    }
    fprintf(fp, "\n");

    fprintf(fp, "%s", "mean");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(fp, ",%f", theTiming->mean[unit]);
        }
    }
    fprintf(fp, "\n");

    fprintf(fp, "%s", "stddev");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(fp, ",%f", theTiming->stddev[unit]);
        }
    }
    fprintf(fp, "\n");

    fclose(fp);
}

/**
 * hpgv_timing_show_global
 *
 */
void
hpgv_timing_show_global()
{
    int groupsize = theTiming->groupsize;
    int countroot = theTiming->countroot;
    int totalproc = groupsize;
    int unit;

    if (!countroot) {
        totalproc--;
    }

    fprintf(stderr, "\nglobal performance data for %s\n",
            theTiming->contextglobal);

    fprintf(stderr, "-------");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(stderr, "-----------");
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "%-6s|", " ");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(stderr, "%-10s|", theTiming->name[unit]);
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "-------");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(stderr, "-----------");
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "%-6s|", "sum");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            if (strchr(theTiming->name[unit], 'C')){
                fprintf(stderr, "%-10.f|", theTiming->sum[unit]);
            } else {
                fprintf(stderr, "%-10.4f|", theTiming->sum[unit]);
            }
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "%-6s|", "min");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            if (strchr(theTiming->name[unit], 'C')){
                fprintf(stderr, "%-10.0f|", theTiming->min[unit]);
            } else {
                fprintf(stderr, "%-10.4f|", theTiming->min[unit]);
            }
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "%-6s|", "max");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            if (strchr(theTiming->name[unit], 'C')){
                fprintf(stderr, "%-10.0f|", theTiming->max[unit]);
            } else {
                fprintf(stderr, "%-10.4f|", theTiming->max[unit]);
            }
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "%-6s|", "mean");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            if (strchr(theTiming->name[unit], 'C')){
                fprintf(stderr, "%-10.0f|", theTiming->mean[unit]);
            } else {
                fprintf(stderr, "%-10.4f|", theTiming->mean[unit]);
            }
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "%-6s|", "stddev");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            if (strchr(theTiming->name[unit], 'C')){
                fprintf(stderr, "%-10.0f|", theTiming->stddev[unit]);
            } else {
                fprintf(stderr, "%-10.4f|", theTiming->stddev[unit]);
            }
        }
    }
    fprintf(stderr, "\n");

    fprintf(stderr, "-------");
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        if (strcmp(theTiming->name[unit], "") != 0) {
            fprintf(stderr, "-----------");
        }
    }
    fprintf(stderr, "\n");
}

/**
 * hpgv_timing_statistics
 *
 */
void
hpgv_timing_statistics()
{
    int mpiid = theTiming->mpiid;
    int groupsize = theTiming->groupsize;
    int root = theTiming->root;
    int unit, proc, t, count;
    int countroot = theTiming->countroot;
    double sumation, max, min, mean, stddev;

    /* gather timing data */
    hpgv_timing_gather();

    /* return if not root */
    if (mpiid != root) {
        return;
    }

    /* calcuate the timing of each unit of each procesor */
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        for (proc = 0; proc < groupsize; proc++) {
            t = theTiming->t_count[proc][unit];
            if (t != 0) {
                theTiming->t_proc[unit][proc]
                    = theTiming->t_timing[proc][unit] / t;
            } else {
                theTiming->t_proc[unit][proc] = 0;
            }
        }
    }

    /* calculate the statistics of each unit */
    for (unit = 0; unit < HPGV_MAX_TIMING_UNIT; unit++) {
        sumation = 0;
        count = 0;
        max = -D_FLOAT_INFINITY;
        min =  D_FLOAT_INFINITY;

        for (proc = 0; proc < groupsize; proc++) {
            if (proc == root && !countroot) {
                continue;
            }

            sumation += theTiming->t_proc[unit][proc];
            count++;

            if (max < theTiming->t_proc[unit][proc]) {
                max = theTiming->t_proc[unit][proc];
            }

            if (min > theTiming->t_proc[unit][proc]) {
                min = theTiming->t_proc[unit][proc];
            }
        }

        if (count != 0) {
            mean = sumation / count;
        } else {
            mean = 0;
        }

        stddev = 0;

        for (proc = 0; proc < groupsize; proc++) {
            if (proc == root && !countroot) {
                continue;
            }

            stddev += SQR(theTiming->t_proc[unit][proc] - mean);
        }

        stddev = sqrt(stddev / count);

        theTiming->sum[unit] = sumation;
        theTiming->min[unit] = min;
        theTiming->max[unit] = max;
        theTiming->mean[unit] = mean;
        theTiming->stddev[unit] = stddev;
    }

    /* save the timing of each proc */
    if (theTiming->savelocal) {
        hpgv_timing_save_local();
    }

    /* print the timing of each proc */
    if (theTiming->showlocal) {
        hpgv_timing_show_local();
    }

    /* save the global timing */
    if (theTiming->saveglobal) {
        hpgv_timing_save_global();
    }

    /* print the global timing */
    if (theTiming->showglobal) {
        hpgv_timing_show_global();
    }
}

/**
 * hpgv_timing_showbreakdown
 *
 */
void
hpgv_timing_showbreakdown(int stage)
{
}
