#include <fstream>
#include <vector>
#include <algorithm>
#include "obj_model_t.h"
#include "vec3d.h"
#include "DError.h"
#include "vertex_triangle_t.h"
#include "constant.h"
#include "mathtool.h"
using namespace lily;

obj_model_t::obj_model_t(void)
    :triangle_tab(NULL), triangle_num(0)
    ,vertex_tab(NULL), vertex_num(0)
    ,vtx_normal_tab(NULL), vtx_normal_num(0)
#ifdef __USE_GL__
    ,m_vbo(NULL)
#endif
    ,m_baseIndex(1)
{
}

obj_model_t::~obj_model_t(void)
{
    obj_free();
}

void obj_model_t::obj_free(void)
{
    if (triangle_tab){
        free(triangle_tab);
        triangle_tab = NULL;
        triangle_num = 0;
    }
    if (vertex_tab){
        free(vertex_tab);
        vertex_tab = NULL;
        vertex_num = 0;
    }
    if (vtx_normal_tab){
        free(vtx_normal_tab);
        vtx_normal_tab = NULL;
        vtx_normal_num = 0;
    }

}

void obj_model_t::obj_model_set_origin( vec3f origin )
{

    //set the origin of our coordinate system to origin,
    //and transform the vertex coords to this coordinate system.
    vec3f* vtx_tab = (vec3f*)(vertex_tab);

    for (int i = 0; i < vertex_num ; i++)
    {
        vtx_tab[i] -= origin;
    }
    m_local_bbox.pMax -= origin;
    m_local_bbox.pMin -= origin;
}

BBox obj_model_t::obj_model_get_local_bbox() const
{
    return m_local_bbox;
}

BBox obj_model_t::obj_model_get_global_bbox() const
{
    return m_local_bbox;
}

vec3f obj_model_t::obj_model_get_center() const
{
    return m_local_bbox.Center();
}

int obj_model_t::obj_model_get_vertex_count() const
{
    return vertex_num;
}

int obj_model_t::obj_model_get_triangle_count() const
{
    return triangle_num;
}

int obj_model_t::obj_model_get_normal_count() const
{
    return vtx_normal_num;
}

void obj_model_t::obj_model_scale_obj( float scale )
{
#ifdef _PRINT_HINT_
    printf("obj_model_scale_obj\n");
#endif
    vec3f* vtx_tab = (vec3f*)(vertex_tab);
    for (int i = 0; i < vertex_num ; i++){
        vtx_tab[i] *= scale;
    }
    m_local_bbox.pMax *= scale;
    m_local_bbox.pMin *= scale;
}

void obj_model_t::obj_model_scale_obj_xyz( float sx, float sy, float sz )
{
    vec3f* vtx_tab = (vec3f*)(vertex_tab);
    vec3f scale_xyz(sx,sy,sz);
    for (int i = 0; i < vertex_num ; i++){
        vtx_tab[i] *= scale_xyz;
    }
    m_local_bbox.pMax *= scale_xyz;
    m_local_bbox.pMin *= scale_xyz;
}


void obj_model_t::obj_model_save_file( const char* file_name )
{
    FILE *fp = NULL;
    fp = fopen(file_name, "wb");
    D_ASSERT(fp, "Can not open file", DERROR_ERR_IO);
    char msg[256];
    fprintf(stderr, "Save obj model to file: %s\n", file_name);
    fflush(stderr);
    //vertex
    fwrite(&(vertex_num), sizeof(int), 1 , fp);
    fwrite(vertex_tab, sizeof(vec3f), vertex_num, fp);
    //normal
    fwrite(&(vtx_normal_num), sizeof(int), 1, fp);
    fwrite(vtx_normal_tab, sizeof(vec3f), vtx_normal_num, fp);
    //triangle faces
    int nElments = vtx_normal_num > 0 ? 6 : 3;

    fwrite(&(triangle_num), sizeof(int), 1, fp);
    fwrite(triangle_tab, sizeof(int)*nElments, triangle_num, fp);

    fclose(fp);
}
void obj_model_t::compute_bbox() 
{
    BBox box;
    int nElments = vtx_normal_num > 0 ? 6 : 3;
    //compute bbox of obj model.
    //triangle_num = triangle_tab_size/6;
    //int nElments = vtx_normal_num > 0 ? 6 : 3;
    for (int i = 0; i < triangle_num ; i++){
        //for each triangle copy 3 vertex
        //indices of triangle i to indices variable
        vec3i indices;
        memcpy( &indices, &(triangle_tab[i*nElments]), sizeof(vec3i) );
        for (int j = 0; j < 3 ; j++){
            //obj file use index that starts with 1
            vec3f vtx;
            memcpy( &vtx, &(vertex_tab[(indices.v(j)-m_baseIndex)*3]), sizeof(vec3f) );
            box = box.Union(vtx);
            //UNION(pt_min, pt_max, vtx);
        }
    }

    m_local_bbox = box;
}

void obj_model_t::obj_model_load_file2(const char* file_name )
{

    char* in_vtx_hdr = "E:/archieve/distance_field/tire1/c_boeing_vertex.hdr";
    char* in_vtx_dat = "E:/archieve/distance_field/tire1/c_boeing_vertex.dat";
    char* in_tria_hdr= "E:/archieve/distance_field/tire1/c_boeing_triangle.hdr";
    char* in_tria_dat= "E:/archieve/distance_field/tire1/c_boeing_triangle.dat";

    cout<<"Load "<<in_vtx_dat<<endl;
    cout<<"Load "<<in_tria_dat<<endl;
    ifstream inf(in_vtx_hdr);
    assert(inf);
    inf >> vertex_num;
    inf.close();
    cout<<"Load "<<vertex_num<<" Vertices.\n";

    vertex_tab = new float[vertex_num*3];
    assert(vertex_tab);

    inf.open(in_vtx_dat, ifstream::binary);
    assert(inf);
    inf.read(reinterpret_cast<char*>(vertex_tab), sizeof(float)*3*vertex_num);
    inf.close();

    inf.open(in_tria_hdr);
    assert(inf);
    inf >> triangle_num;
    inf.close();

    triangle_tab = new int[triangle_num*3];
    assert(triangle_tab);

    inf.open(in_tria_dat, ifstream::binary);
    assert(inf);
    inf.read(reinterpret_cast<char*>(triangle_tab), sizeof(int)*3*triangle_num);
    inf.close();

    compute_bbox();
    fprintf(stderr,"#vtx=%d,\n#face=%d,\n#normals=%d loaded.\n", vertex_num, triangle_num, vtx_normal_num);
    fflush(stderr);
}
#define _PRINT_HINT_
void obj_model_t::obj_model_load_file(const char* file_name )
{
#ifdef _PRINT_HINT_
    fprintf(stderr,"obj_model_load_file\n");
#endif
    FILE *fp = NULL;
    char msg[256];
    int  size = 0;
    fp = fopen(file_name, "rb");
    sprintf(msg, "Can not open file %s", file_name);
    D_ASSERT(fp, msg, DERROR_ERR_IO);
    //vertex count
    fread(&size, sizeof(int), 1, fp);
//    D_ASSERT(size>0, "# of vertex components is not divisable by 3.\n", DERROR_ERR_IO);
    vertex_num = size;
    //vertex buffer
    if (vertex_num > 0){
        vertex_tab = (float*) calloc(vertex_num, sizeof(vec3f));
        D_ASSERT(vertex_tab, "vertex table allocation failed.", DERROR_ERR_MEM);
        fread(vertex_tab, sizeof(vec3f), vertex_num, fp);
    }
    //normal count
    fread(&size, sizeof(int), 1, fp);
//    D_ASSERT(size>=0, "# of vertex normal components is not divisable by 3.\n", DERROR_ERR_IO);
    vtx_normal_num = size;
    //normal buffer
    if (vtx_normal_num > 0){
        vtx_normal_tab = (float*) calloc(vtx_normal_num, sizeof(vec3f));
        D_ASSERT( vtx_normal_tab, "normal table allocation failed.", DERROR_ERR_MEM);
        fread( vtx_normal_tab, sizeof(vec3f), vtx_normal_num, fp);
    }
    //triangle table
    fread(&size, sizeof(int), 1, fp);

//    D_ASSERT(size > 0, "# of vertex normal components is not divisable by 3.\n", DERROR_ERR_IO);
    triangle_num = size;
    int nElments = vtx_normal_num > 0 ? 6 : 3;
    if (triangle_num > 0){
        triangle_tab = (int*) calloc(triangle_num, sizeof(int)*nElments);
        D_ASSERT( triangle_tab, "triangle table allocation failed.", DERROR_ERR_MEM);
        fread( triangle_tab, sizeof(int)*nElments, triangle_num, fp);
    }
    fclose(fp);
    compute_bbox();
    /*
    geometry::BBox box;
    //compute bbox of obj model.
    //triangle_num = triangle_tab_size/6;
    //int nElments = vtx_normal_num > 0 ? 6 : 3;
    for (int i = 0; i < triangle_num ; i++){
        //for each triangle copy 3 vertex
        //indices of triangle i to indices variable
        vec3i indices;
        memcpy( &indices, &(triangle_tab[i*nElments]), sizeof(vec3i) );
        for (int j = 0; j < 3 ; j++){
            //obj file use index that starts with 1
            vec3f vtx;
            memcpy( &vtx, &(vertex_tab[(indices.v(j)-1)*3]), sizeof(float)*3 );
            box = box.Union(vtx);
            //UNION(pt_min, pt_max, vtx);
        }
    }

    m_bbox = box;
    */
#ifdef _PRINT_HINT_
    fprintf(stderr,"#vtx=%d,\n#face=%d,\n#normals=%d loaded.\n", vertex_num, triangle_num, vtx_normal_num);
    fflush(stderr);
#endif
}

void obj_model_t::obj_model_create_from_triangle_soup
    (svector<vertex_triangle_t>& vertex_array,  int myid)
{
    if (vertex_array.size()<=0)
    {
        fprintf(stderr,"PE%d: The # of vertices=%d in triangle soup\n"
                ,myid, vertex_array.size() );
        fflush(stderr);
        return;
    }
    
    //make sure the number of input vertices can be divided by 3.
    if (vertex_array.size()%3 != 0)
    {
        fprintf(stderr,"PE%d: The # of vertices=%d in triangle soup is not exact multiple of 3\n"
            ,myid, vertex_array.size() );
        fflush(stderr);
        exit(1);
    }

    size_t triangle_count = vertex_array.size() / 3;
    if (triangle_count < 0)
    {
        return;
    }

    triangle_tab = (int*)calloc(triangle_count, sizeof(vec3i));//no normal info. Only vertex indices.

    if ( triangle_tab == NULL)
    {
        fprintf(stderr,"PE%d: triangle_tab == NULL\n",myid);
        fflush(stderr);
        exit(1);
    }
    vec3i* triangle_array = (vec3i*)triangle_tab;
    vec3i *cur_triangle = NULL;
    size_t cur_triangle_idx = 0;
    qsort(vertex_array.data_ptr(), vertex_array.size(), sizeof(vertex_triangle_t), __vertex_compare);
    cur_triangle = &(triangle_array[vertex_array[0].m_idx_triangle]);
    size_t last_unique_idx = 0;
    (*cur_triangle)[vertex_array[0].m_local_idx] = last_unique_idx + m_baseIndex;//obj file uses index starting from 1.

    BBox box;
    box.Union(vertex_array[0].m_coord);

    size_t numVtx = vertex_array.size();
    for(size_t i = 1; i < numVtx ; i++)
    {//scan the vertex array that may contains duplicates.
        if ( vertex_array[i].m_coord != vertex_array[last_unique_idx].m_coord )
        {//found a new unique.
            vertex_array[++last_unique_idx] = vertex_array[i];
            box.Union(vertex_array[i].m_coord);
        }
        //Update the vertex index of corresponding triangle that is instant to the vertex_array[i].
        cur_triangle_idx = vertex_array[i].m_idx_triangle;
        if ( cur_triangle_idx >= triangle_count)
        {
            fprintf(stderr,"PE%d: cur_triangle_idx(%d) >= triangle_count(%d)!\n"
                ,myid, cur_triangle_idx, triangle_count);
            fflush(stderr);
            exit(1);
        }

        cur_triangle = &(triangle_array[cur_triangle_idx]);

        if (   vertex_array[i].m_local_idx < 0
            || vertex_array[i].m_local_idx > 2)
        {//m_local_idx={0,1,2}
            fprintf(stderr,"PE%d: vertex_array[i].m_local_idx(%d) out of bound!\n"
                ,myid, vertex_array[i].m_local_idx);
            fflush(stderr);
            exit(1);
        }

        (*cur_triangle)[vertex_array[i].m_local_idx] = last_unique_idx+m_baseIndex;//obj file uses index starting from 1.
    }

    if (vertex_tab)
    {
        free(vertex_tab);
        vertex_num = 0;
    }
    vertex_num = last_unique_idx+1;
    vertex_tab = (float *)calloc(vertex_num, sizeof(vec3f));

    if ( vertex_tab == NULL)
    {
        fprintf(stderr,"PE%d: vertex_tab == NULL!\n"
            ,myid, cur_triangle_idx, triangle_count);
        fflush(stderr);
        exit(1);
    }
    vec3f* vertices = (vec3f*) vertex_tab;
    for (int i = 0 ; i < vertex_num ; i++)
    {
        vertices[i] = vertex_array[i].m_coord;
    }

    triangle_tab = (int*)triangle_array;
    triangle_num = triangle_count ;//* 6;
    if (vtx_normal_tab)
    {
        free(vtx_normal_tab);
        vtx_normal_tab    = NULL;
        vtx_normal_num    = 0;//no normal generated.
    }
    m_local_bbox = box;
    cout<< "obj_model_t: "<< vertex_num <<" unique vertices generated\n";
    cout<< "obj_model_t: "<< triangle_count <<" unique triangle faces generated\n";
}

void obj_model_t::obj_model_remove_duplicated_vertex()
{
#ifdef _PRINT_HINT_
    printf("obj_model_remove_duplicated_vertex\n");
#endif
    if(triangle_num<=0 || vertex_num <=0)
        return;

    vec3i* triangles = (vec3i*) triangle_tab;
    vec3f* coords_array = (vec3f*) vertex_tab;
    svector<vertex_triangle_t> vertices;
    vertices.reserve(triangle_num*3);
    //1. scan each triangle and collect their vertices
    int nElements = vtx_normal_num > 0 ? 2 : 1;
    for (int iTria = 0 ; iTria < triangle_num ; iTria++){
        vec3i coords_index = triangles[nElements*iTria];
        //obj index start at 1.
        vec3f cur_coords1 = coords_array[coords_index.x() - m_baseIndex];
        vec3f cur_coords2 = coords_array[coords_index.y() - m_baseIndex];
        vec3f cur_coords3 = coords_array[coords_index.z() - m_baseIndex];
        vertices.push_back(vertex_triangle_t(cur_coords1, 0, iTria));
        vertices.push_back(vertex_triangle_t(cur_coords2, 1, iTria));
        vertices.push_back(vertex_triangle_t(cur_coords3, 2, iTria));
    }
    //Remove duplicates.
    //1.Sort vertex_array based on vertex's key.
    //sort(vertices.begin(), vertices.end()/*, myfunction*/);
    qsort(vertices.data_ptr(),
          vertices.size(),
          sizeof(vertex_triangle_t),
          __vertex_compare);

    int unique_count = 0;
    //Update vertex index in every triangle,
    // since after sorting all the indices have been changed.
    vec3i& cur_triangle = triangles[(vertices[0].m_idx_triangle)*nElements];
    cur_triangle[vertices[0].m_local_idx] = unique_count+m_baseIndex;//obj file uses index starting from 1.

    int duplicated_vertices_count = vertices.size();
    for (int i = 1; i < duplicated_vertices_count ; ++i){
        //found a different vertex, add it to unique_vertice array.
        if (vertices[i].m_coord != vertices[unique_count].m_coord ){
            /*
            if (vec3f(4.412689,-9.252060,2.884491)
                == vertices[unique_count].m_coord)
            {
                cout << "^^^^\n";
                cout <<"unique_vtx:"<< vertices[unique_count].m_coord
                    <<"!=vertice["<<i<<"]="<<vertices[i].m_coord<<endl;
            }
            */
            vertices[++unique_count] = vertices[i];
        }
        vec3i& cur_triangle = triangles[(vertices[i].m_idx_triangle)*nElements];
        cur_triangle[vertices[i].m_local_idx] = unique_count+m_baseIndex;//obj file uses index starting from 1.
    }
    //Compact array.
#ifdef _PRINT_HINT_
    fprintf(stderr,"********Orignal #vertex=%d********\n",vertex_num);
    fprintf(stderr,"********Unique #vertex=%d********\n",unique_count+1);
    fflush(stderr);
#endif

    vertex_num = unique_count+1;
    vertex_tab = (float*)realloc(vertex_tab, vertex_num*sizeof(vec3f));//(float*)calloc(vertex_num, sizeof(vec3f));//
    vec3f* vertex3f_array = (vec3f*)vertex_tab;
    //update vertex_tab with unique vertices info.
    for (int i = 0 ; i <vertex_num; i++){
        vertex3f_array[i] = vertices[i].m_coord;
    }

    //obj_model_print_vertices();
    //obj_model_print_triangles();
}
void obj_model_t::obj_model_print_vertices(void)
{
    vec3f* coords_array = (vec3f*)vertex_tab;
    for (int i = 0 ; i < vertex_num ; ++i){
        cout<<"vtx"<<i<<" coords="<<coords_array[i]<<endl;
    }
}
void obj_model_t::obj_model_print_triangles(void)
{
    vec3i* triangles = (vec3i*) triangle_tab;
    int nElements = vtx_normal_num > 0 ? 2 : 1;
    for (int i = 0 ; i < triangle_num ; ++i){
        vec3i vtx_indices;
        vtx_indices = triangles[i*nElements];
        printf("*triangle%d(%d,%d,%d)",i,vtx_indices[0],vtx_indices[1],vtx_indices[2]);
        if (vtx_normal_num>0){
            vec3i normals;
            normals = triangles[i*nElements+1];
            printf(", normal_idx(%d,%d,%d)\n", normals[0],normals[1],normals[2]);
        }else{
            printf("\n");
        }
    }
}
void obj_model_t::obj_model_remove_degenerated_triangle( void )
{
#ifdef _PRINT_HINT_
    fprintf(stderr,"obj_model_remove_degenerated_triangle.\n");
#endif
    if(triangle_num<=0 || vertex_num <=0)
        return;

    vec3i* triangles = (vec3i*) triangle_tab;
    vec3f* coords_array = (vec3f*) vertex_tab;
    int nElements = vtx_normal_num > 0 ? 2 : 1;
    //int idxBase = 1;//obj index start at 1.
    int inormal_triangle = 0;
#ifdef _PRINT_HINT_
    fprintf(stderr,"# normals=%d.\n",vtx_normal_num);
#endif
    int countPtCollide=0, countCollinear=0;
    for (int iTria = 0 ; iTria < triangle_num ; iTria++){
        vec3i vtx_indices;
        vtx_indices = triangles[iTria*nElements];
        //Check degenerated triangle and ignore it. Since it is very likely that some large geometry
        //comes with triangle whose two vertices or three vertices shared the same vertex index.
        //in this case vertex to triangle distance computation would fail.(Check the wiki for more details).
        if (  vtx_indices.x()==vtx_indices.y()
            ||vtx_indices.y()==vtx_indices.z()
            ||vtx_indices.x()==vtx_indices.z()){
               //printf("****Warning: degenerated triangle detect at triangle id(%d,%d,%d)=%d, ignored!*****\n",\
                   vtx_indices[0],vtx_indices[1],vtx_indices[2],iTria);
               countPtCollide++;
                continue;
        }else{
            //further check if the coordinates of the triangle collide
            vec3f cur_coords1 = coords_array[vtx_indices.x() - m_baseIndex];
            vec3f cur_coords2 = coords_array[vtx_indices.y() - m_baseIndex];
            vec3f cur_coords3 = coords_array[vtx_indices.z() - m_baseIndex];
           
            vec3f vn1 = (cur_coords2-cur_coords1).cross(cur_coords3-cur_coords1);
            vn1.normalize();
            if (vn1.length() < FLOAT_EPSILON)
            {//three points collinear
                //cout <<"*** Warning: degenrated triangle(coord collapse or 3 points on a line) detected at triangle id"<<iTria
                //    <<":\nv1="<<cur_coords1<<"v2="<<cur_coords2<<"v3="<<cur_coords3;
                cout <<"three points collinear. Skip current triangle!\n";
                countCollinear++;
               continue;
            }else{
                triangles[inormal_triangle*nElements] = triangles[iTria*nElements];
                if (vtx_normal_num>0){
                    triangles[inormal_triangle*nElements+1] = triangles[iTria*nElements+1];
                }
                inormal_triangle++;
            }

        }
    }
    cout<<"++++Total of "<<countCollinear+countPtCollide<<" degenerated triangles detected and skiped!\n";
    cout<<"++++"<<countPtCollide<<" triangles have points collides!\n"
        <<"++++"<<countCollinear<<" triangles have points collinear!\n"
        <<"++++"<<"You really have a nasty geometry model!\n";

#ifdef _PRINT_HINT_
    fprintf(stderr,"*Original #triangle=%d\nNormal #triangle=%d*\n",
            triangle_num, inormal_triangle);
    fflush(stderr);
#endif
    triangle_num = inormal_triangle;
    triangle_tab = (int*)realloc(triangle_tab, triangle_num*sizeof(vec3f)*nElements);
//     this->obj_model_print_vertices();
//     this->obj_model_print_triangles();
}

bool obj_model_t::obj_model_get_triangle_at( int idx, triangle_t& cur_triangle ) const
{
    vec3i*  triangles  = (vec3i*) triangle_tab;
    vec3f*  vtx_tab    = (vec3f*) vertex_tab;
    bool    bHasNormal = vtx_normal_num > 0;
    int     nElements  = bHasNormal  ? 2 : 1;
    int     idxBase = m_baseIndex;//obj index start at 1.
    vec3i   vtx_indices = triangles[idx * nElements];

    //Warning: obj file vertex index starts from 1
    vec3f vtx1 = vtx_tab[(vtx_indices.v(0) - idxBase)];
    vec3f vtx2 = vtx_tab[(vtx_indices.v(1) - idxBase)];
    vec3f vtx3 = vtx_tab[(vtx_indices.v(2) - idxBase)];
    cur_triangle.v[0]   = vtx1;
    cur_triangle.v[1]   = vtx2;
    cur_triangle.v[2]   = vtx3;
    
    //Use existing vertex normals or compute them from scratch based on if obj file provide them or not.
    if (bHasNormal)
    {
        vec3f* vtx_normal = (vec3f*)vtx_normal_tab;
        vec3i normal_indices = triangles[idx*nElements+1];
        vec3f vn1  = vtx_normal[(normal_indices.v(0) - idxBase)];
        vec3f vn2  = vtx_normal[(normal_indices.v(1) - idxBase)];
        vec3f vn3  = vtx_normal[(normal_indices.v(2) - idxBase)];
        vn1.normalize(); vn2.normalize(); vn3.normalize();
        cur_triangle.vn[0] = vn1;
        cur_triangle.vn[1] = vn2;
        cur_triangle.vn[2] = vn3;
        cur_triangle.normal = (vn1+vn2+vn3)/3.0f;
        if (cur_triangle.normal.length()<FLOAT_EPSILON)
        {
            cout <<"^^^^^^^\nNULL original normal("<<vn1.length()<<") at triangle "<<idx<<endl;
            cout <<"Normal index="<<normal_indices.v(0)<<endl;
            cout <<"v1["<<vtx_indices.x()<<"]="<<vtx1
                <<"v2["<<vtx_indices.y()<<"]="<<vtx2
                <<"v3["<<vtx_indices.z()<<"]="<<vtx3
                <<"vn1="<<vn1<<"vn1.length="<<vn1.length()<<endl;
            cout<<"Try to recover the normal using triangle's vertices\n";
            vn1 = (vtx2-vtx1).cross(vtx3-vtx1);
            vn1.normalize();
            cur_triangle.normal = vn1;
            cout<<"Recovered normal="<<vn1;
            if (vn1.length()<FLOAT_EPSILON)
            {
                cout<<"++++++++Recovered normal is still NULL!!++++++++\n";
                cout<<"You're suffering single-floating point numerical error! God bless you!\n";
                cout<<"-----------------------------------------------------------\n";
            }
            cur_triangle.vn[0] = vn1;
            cur_triangle.vn[1] = vn1;
            cur_triangle.vn[2] = vn1;

        }

    }else
    {
        //Compute triangle surface normal using the vertex coords.
        vec3d vn1 = (vec3d(vtx2)-vec3d(vtx1)).cross(vec3d(vtx3)-vec3d(vtx1));
        vn1.normalize();
        cur_triangle.normal[0]= vn1[0];
        cur_triangle.normal[1]= vn1[1];
        cur_triangle.normal[2]= vn1[2];
        if (vn1.length()<FLOAT_EPSILON)
        {
            cout <<"NULL normal("<<vn1.length()<<") generated at triangle "<<idx<<endl;
            cout <<"v1["<<vtx_indices.x()<<"]="<<vtx1
                 <<"v2["<<vtx_indices.y()<<"]="<<vtx2
                 <<"v3["<<vtx_indices.z()<<"]="<<vtx3
                <<"vn1="<<vn1<<"vn1.length="<<vn1.length()<<endl;
            cout <<"God Bless You!\n";
        }
        cur_triangle.vn[0] = cur_triangle.normal;
        cur_triangle.vn[1] = cur_triangle.normal;
        cur_triangle.vn[2] = cur_triangle.normal;
    }

    if (  vtx_indices.x()!=vtx_indices.y()
        &&vtx_indices.y()!=vtx_indices.z()
        &&vtx_indices.x()!=vtx_indices.z())
    {
        return true;
    }else
    {
        printf("****Warning: degenerated triangle detected at triangle%d!*****\n",idx);
        return false;
    }
}