#include <mpi.h>
#include <omp.h>
#include <iostream>
#include "vec3i.h"
#include "vec3f.h"
#include "mathtool.h"
#include "hpgv_utiltiming.h"
#include "DError.h"
#include "parallel_geometry_test.h"
#include "dvolume.h"
#include "distree_t.h"
#include "feature_geometry_t.h"
using namespace lily;

parallel_geometry_test::~parallel_geometry_test(void)
{
}

parallel_geometry_test::parallel_geometry_test(int argc, char** argv)
{
    cout << "Parallel Distance Transform on Geometry Model\n";
    int myid=-1, groupsize=-1;
    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &myid);
    MPI_Comm_size(MPI_COMM_WORLD, &groupsize);
    if (argc != 12) {
        if (myid == 0) {//
            fprintf(stderr, "%s npx npy npz gx gy gz sx leafsize ratio obj_bin thread_num\n", argv[0]);
            fflush(stderr);
        }
        MPI_Finalize();
        return;
    }
    int thread_num = 1;
#ifdef _OPENMP
    thread_num = atoi(argv[11]);
    omp_set_num_threads(thread_num);
#endif
    vec3i np_xyz(atoi(argv[1]),atoi(argv[2]),atoi(argv[3]) );
    vec3i global_domain_size(atoi(argv[4]),atoi(argv[5]),atoi(argv[6]) );
    vec3f sample_size((float)atof(argv[7]));
    vec3f leaf_size((float)atof(argv[8]));
    float ratio = (float)atof(argv[9]);
    char msg[1024];
	const char *output_dir="";
    sprintf(msg, "%sPE_%d_%dx%dx%d_Vol_%dx%dx%d_td_%d"
        ,output_dir
        ,groupsize, np_xyz.x(), np_xyz.y(), np_xyz.z(),
        global_domain_size.x(), global_domain_size.y(), global_domain_size.z()
        ,thread_num);

    bool isInitialized = false;
    isInitialized = init_timing_module(msg, isInitialized);

    vec3i pxyz = vec3i::index1D_to_index3D(myid, np_xyz);
    
    HPGV_TIMING_COUNT(LOAD_GEOMETRY_TIME);
    color_t mycolor=WHITE;
    obj_model_t* obj_model = new obj_model_t();
    obj_model->obj_model_set_baseIndex(0);//!!! choose correct basedindex. obj model file use base-1 index.

    feature_geometry_t local_data_geometry(obj_model);
    local_data_geometry.load_geometry_from_binfile(argv[10]);

    BBox obj_bbox = obj_model->obj_model_get_local_bbox();
    cout <<"##before scaling global bbox:"<<obj_bbox;
    vec3f scale_xyz = (obj_bbox.pMax - obj_bbox.pMin)/(vec3f(global_domain_size)*ratio);
    float scale = 1.0f/scale_xyz.max_extend1f();
    obj_model->obj_model_scale_obj(scale);
    vec3f center = obj_model->obj_model_get_local_bbox().Center();
    obj_model->obj_model_set_origin(center);
    //rescaled and re-centering would create some degenerated triangle due to numerical error
    //clean up the geometry is required!!
    local_data_geometry.remove_degenerated_triangle();
    {
        D_P_MSG(myid,0
            ,"^^^^^^^^^^^^PE%d^^^^pass obj model initialization^^^^^\n");
    }
    obj_bbox = obj_model->obj_model_get_local_bbox();
    vec3i global_feature_size(obj_bbox.pMax - obj_bbox.pMin);
    cout <<"global bbox:"<<obj_bbox;
    cout <<"PE"<<myid<< " global_feature_size="<<global_feature_size<<endl;
    //Determine the geometry that falls into PEi's local block.
    //Check if current PE has local geometry, if it does, its local block is element block.
    mycolor = local_data_geometry.contain_local_geometry(pxyz, np_xyz);

    {
        char msg[128];
        sprintf(msg,"^^^^^mycolor=%s^^^^^\n", mycolor==BLACK? "black" : "white");
        D_P_MSG(myid,0,msg);
    }
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(MY_ALL_SLOVE_TIME);
    HPGV_TIMING_COUNT(MY_ALL_SLOVE_TIME);
    //Generate bitmap.
    int MAX_DEPTH_ID = 9;
    distree_t distance_tree(myid, F_TRIANGLE, MPI_COMM_WORLD, MAX_DEPTH_ID);
    distance_tree.set_sample_size(sample_size);
    distance_tree.set_leaf_node_size(leaf_size);

    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(CREATE_BITMAP_TIME);
    HPGV_TIMING_COUNT(CREATE_BITMAP_TIME);
    bitmap_t *bitmap =
        distance_tree.generate_bitmap(groupsize, mycolor, MPI_COMM_WORLD);
    HPGV_TIMING_END(CREATE_BITMAP_TIME);

    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(CREATE_GLOBAL_TREE_TIME);
    HPGV_TIMING_COUNT(CREATE_GLOBAL_TREE_TIME);
    {
        D_P_MSG(distance_tree.m_myid,0
            ,"^^^^^^^^^^^^PE%d^^^^pass generate_bitmap^^^^^\n");
    }
    //Create block tree for generation of send/recev_table in create_comm_schedule();
    feature_isoblock_t *isoblock_array = new feature_isoblock_t(bitmap);
    distance_tree.set_isoblock_feature(isoblock_array);
    distance_tree.create_global_distree(global_domain_size,
                                        vec3f(global_feature_size),
                                        np_xyz, groupsize);
    {
        D_P_MSG(distance_tree.m_myid,0
            ,"^^^^^^^^^^^^PE%d^^^^pass create_global_distree^^^^^\n");
    }
    HPGV_TIMING_END(CREATE_GLOBAL_TREE_TIME);

    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(CREATE_SCHEDULE_TIME);
    HPGV_TIMING_COUNT(CREATE_SCHEDULE_TIME);
    //Create communication distree
    distance_tree.create_comm_schedule();

    HPGV_TIMING_END(CREATE_SCHEDULE_TIME);
	
    distance_tree.set_feature_array(&local_data_geometry);
    //distance_tree.save_distree_binary("distree");
    distance_tree.verify_schedule(groupsize);

    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(LOCAL_TREE_TIME);
    HPGV_TIMING_COUNT(LOCAL_TREE_TIME);
    {
        D_P_MSG(distance_tree.m_myid,0
            ,"^^^^^^^^^^^^PE%d^^^^pass initial feature array^^^^^\n");
    }
    //Create local Octree using local volume features but with whole domain(volume) BBox.
    distance_tree.refine_global_distree();
    {
        D_P_MSG(distance_tree.m_myid,0
            ,"^^^^^^^^^^^^PE%d^^^^pass refine_global_distree^^^^^\n");
    }

    HPGV_TIMING_END(LOCAL_TREE_TIME);
    //create vertex for each leaf node.
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(CREATE_VTX_ARRAY);
    HPGV_TIMING_COUNT(CREATE_VTX_ARRAY);
    distance_tree.create_local_distree_vertex_array();
    {
        D_P_MSG(distance_tree.m_myid,0
            ,"^^^^^^^^^^^^PE%d^^^^pass create_local_distree_vertex_array^^^^^\n");
    }
    HPGV_TIMING_END(CREATE_VTX_ARRAY);

    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(VERTEX_DISTANCE_TIME);
    HPGV_TIMING_COUNT(VERTEX_DISTANCE_TIME);
    
    distance_tree.compute_vertex_distance_parallel();
    {
        D_P_MSG(distance_tree.m_myid,0
            ,"^^^^^^^^^^^^PE%d^^^^pass compute_vertex_distance_parallel^^^^^\n");
    }
    HPGV_TIMING_END(VERTEX_DISTANCE_TIME);
    HPGV_TIMING_END(MY_ALL_SLOVE_TIME);
    
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    HPGV_TIMING_BEGIN(COMPUTE_DIST_VOL_ALL_TIME);
    HPGV_TIMING_COUNT(COMPUTE_DIST_VOL_ALL_TIME);
    //distance_tree.create_distance_volume();
    distance_tree.create_distance_volume_allreduce();
    //if (distance_tree.m_myid==13)
    {
        D_P_MSG(distance_tree.m_myid,0
            ,"^^^^^^^^^^^^PE%d^^^^pass create_distance_volume^^^^^\n");
    }
	sprintf(msg, "%sdist_vol_%d_%2.2f_%dx%dx%d_p%dx%dx%d_td%d.dat",
			 output_dir, groupsize,	sample_size.x(),
			 global_domain_size.x(),global_domain_size.y(),global_domain_size.z(),
			 np_xyz.x(), np_xyz.y(), np_xyz.z()
             ,thread_num);
	cerr<<"File saved at"<< msg <<endl;
    HPGV_TIMING_BARRIER(MPI_COMM_WORLD);
    distance_tree.save_distance_volume_allreduce(msg);
    
    HPGV_TIMING_END(COMPUTE_DIST_VOL_ALL_TIME);
    HPGV_TIMING_STATISTICS();
    if (bitmap != 0){
        if (bitmap->data != 0){
            free(bitmap->data);
            bitmap->data = NULL;
        }
        free(bitmap);
        bitmap = NULL;
    }
    if(obj_model) delete obj_model;
    MPI_Finalize();
}

